function simulatie20211117_online(Excelfile,sdir,sheet,stat_col)
% Sander Muns, 2021
% disclaimer: Er kunnen op geen enkele wijze rechten aan de output van deze code worden
% ontleend.
% Excelfile: geeft input-Excelfile weer, hierop wordt ook output weggeschreven
% sdir:      optionele input om directory van Excelfile aan te geven.
% sheet:     optionele input om sheet aan te geven, leeg betekent alle sheets afgaan.
% stat_col:  optionele input om aan te geven in welke kolom statistieken worden weggeschreven.

% De belangrijkste statistieken worden weggeschreven in cellen A59:G59 op tabblad stats van inleesbestand.

% voorbeeldaanroep:
% simulatie20211117_online('dkg100E0BSgeenhervVermlftd20211117.xlsx')
if ~exist('Excelfile','var') || isempty(Excelfile)
    Excelfile = 'input.xlsx';
end
if ~exist('sdir','var') || isempty(sdir)
    sdir  = 'Excel';
end
if ~exist('sheet','var') || isempty(sheet)
    sheet  = [];
end
dirfile    = ['./' sdir '/' Excelfile];
if ~contains(dirfile, '.xls')
    dirfile = [dirfile '.xlsx'];
end
disp(['Reading input from ' dirfile])
[parcell,profielcell,sheetcell] = readInput(dirfile,sheet,false);

excelfile_str = strsplit(Excelfile,'.xls');
excelfile_str = excelfile_str{1};

e       = actxserver('Excel.Application'); % # open Activex server
ewb     = e.Workbooks.Open([pwd '/' sdir '/' Excelfile]); % # open file (enter full path!)
N_Sheets= ewb.Sheets.Count;
sheets  = cell(1,N_Sheets);
for k=1:N_Sheets
    sheets{k} = ewb.Sheets.Item(k).Name;
end
disp(sheets)
ewb.Close(false);
e.Quit;
delete(e);
fclose('all');
spath         = ''; % niet relevant als isempty(par.file);  './Scenarios/';
row_sheetnr   = 0;
sheet_compute = 0;
for i=1:length(parcell)
    clearvars  -except  excelfile_str  dirfile  i  meanperjaar  medianperjaar  mCEQperjaar  m0perjaar  parcell  profielcell  row_sheetnr   sdir   sheet  sheet_compute  sheetcell  sheets spath;
    if ~isempty(parcell{i})
    sheet_compute = sheet_compute+1;
    clear x_opt;
    par     = parcell{i};
    profiel = profielcell{i};
    if sheet_compute==1 && (~isnan(par.disp_coh1) && (par.disp_coh1 <= par.disp_coh2))
        mCEQperjaar = NaN( length(parcell) * length(par.disp_coh1 : par.disp_step : par.disp_coh2), par.max_age - par.pension_age + 2 );
        meanperjaar = NaN( size(mCEQperjaar) );
        medianperjaar = NaN( size(mCEQperjaar) );
        m0perjaar   = cell( size(mCEQperjaar,1), 3 );
    end
    if all(~isnan(par.stock_sol)) && ischar(par.stock_sol) && ismember(par.stock_sol,sheets)
        [~,~,Ctmp]  = xlsread(dirfile,'stats');
        Irow0       = strcmp(Ctmp(:,1), 'sheetnaam' );
        Irow_stock  = strcmp(Ctmp(:,1), 'stocks_assets ' );
        Irow2       = strcmp(Ctmp(:,1), 'bonds_assets  ' );
        Irow3       = strcmp(Ctmp(:,1), 'cash_assets' );
        Irow4       = strcmp(Ctmp(:,1), 'hcap_assets' );
        Icol        = strcmp(Ctmp(Irow0 ,:), par.stock_sol );
        par.stock_sol   = cell2mat(Ctmp(Irow_stock,find(Icol,1)));
        par.bond_sol    = cell2mat(Ctmp(Irow2,find(Icol,1)));
        par.cash_sol    = cell2mat(Ctmp(Irow3,find(Icol,1)));
        par.hcap_sol    = cell2mat(Ctmp(Irow4,find(Icol,1)));
        clear Ctmp Irow0 Irow2 Irow3 Irow4 Icol
    end
    if all(~isnan(par.asset_mix)) && ismember(par.asset_mix,sheets)
        [~,~,Ctmp]       = xlsread(dirfile,'stats');
        Irow0            = strcmp(Ctmp(:,1), 'sheetnaam' );
        Irow_stock       = strcmp(Ctmp(:,1), 'stocks_assets ' );
        Irow_bonds       = strcmp(Ctmp(:,1), 'bonds_assets  ' );
        Icol             = strcmp(Ctmp(Irow0 ,:), par.asset_mix );
        par.asset_mix    = 'lineair';

        perc_stocks      = cell2mat(Ctmp(Irow_stock,find(Icol,1)));
        perc_bonds       = cell2mat(Ctmp(Irow_bonds,find(Icol,1)));
        for tmp = {'_first_exposure_age','_last_work_age','_pension_age','_71','_85','_max_age'}
            s = par.(['stock' tmp{1}]);
            if isnan(s) || ~isnumeric(s{1})
                par.(tmp{1})       = perc_stocks;
            end
            s = par.(['bond' tmp{1}]);
            if isnan(s) || ~isnumeric(s{1})
                par.(tmp{1})       = perc_bonds;
            end
        end
        % lineair profiel obv stockfracties op aantal leeftijden
        v_age    = [par.first_exposure_age   par.pension_age-1   ...
            par.pension_age   min([71 85],par.max_age)   par.max_age];
        num_ages = par.max_age + 1 - par.first_exposure_age;
        
        for s = {'bond','stock','toed_bond'}
            v_tmp = [par.([s{1} '_first_exposure_age'])  par.([s{1} '_last_work_age']) ...
                par.([s{1} '_pension_age'])  par.([s{1} '_71']) ...
                par.([s{1} '_85'])  par.([s{1} '_max_age'])];
            I     = ~isnan(v_tmp);
            profiel.(s{1}) = NaN(num_ages, 1);
            profiel.(s{1})(1:(max(v_age(I))+1-v_age(1))) = linprofiles( v_age(1), v_age(I), v_tmp(I), 1);
        end
        if all(isnan(profiel.bond))
            profiel.bond   = repmat(perc_bonds, size(profiel.stock));
        end
        if strcmp(par.toed_renteafdekking,'als obligatielifecycle')
            profiel.toed_bond = profiel.bond;
        end
        profiel.cash    = 1 - profiel.stock - profiel.bond;
        clear Ctmp Icol Irow0 Irow1
    end
    if isfield(par,'KNW') && isfield(par.KNW,'K11')
        par.KNW.K      = [par.KNW.K11  0 ; par.KNW.K21  par.KNW.K22];
    end
    disp(par)
    if strcmp(par.max_comp_age_return,'max age')
        par.max_comp_age_return = par.max_age;
    end
    if strcmp(par.max_comp_age_return,'pension age')
        par.max_comp_age_return = par.pension_age-1;
    end
    if ismember(par.toed,1:3)
        if strcmp(par.toed_renteafdekking,'als optimale beleggingsmix')
            par.toed_renteafdekking = 1-1/par.gamma;
        else
            if isnan(par.toed_renteafdekking)
                par.toed_renteafdekking = 1;
            end
        end
        if strcmp(par.toed_hcapafdekking,'als optimale beleggingsmix')
            par.toed_hcapafdekking = 1-1/par.gamma;
        else
            if isnan(par.toed_hcapafdekking)
                par.toed_hcapafdekking = 1;
            end
        end
    end
    % bepaal enkele karakteristieken
    num_ages            = par.max_age + 1 - par.first_exposure_age;
    num_work_years      = par.pension_age - par.first_exposure_age;
    num_pension_years   = par.max_age + 1 - par.pension_age;
    num_work_periods    = num_work_years;
    num_pension_periods = num_pension_years;
    Iwork               = [true(1,num_work_periods) false(1,num_pension_periods)];
    Ipension            = ~Iwork;
    num_cohorts         = par.nsim_hor + num_pension_periods - 1;
    
    nloops   = ceil((par.nsim*par.nsim_hor*num_ages)/(1000*100*65));
    nsim0    = par.nsim;
    par.nsim = ceil(par.nsim/nloops);
    
    % leeftijd per cohort dat alle pensioenmaanden heeft
    v_birth_year_coh           = (1:par.nsim_hor+num_pension_periods-1) - par.max_age;
    v_first_pension_period_coh = max(2,round(v_birth_year_coh + par.pension_age + 1));
    v_last_pension_period_coh  = min(par.nsim_hor+1, round(v_birth_year_coh + par.max_age+1));
    v_num_pension_periods_coh  = v_last_pension_period_coh - v_first_pension_period_coh + 1;

    v_first_pension_year_coh   = max(2,v_birth_year_coh + par.pension_age + 1);
    v_last_pension_year_coh    = min(par.nsim_hor+1,v_birth_year_coh + par.max_age+1);
    v_first_pension_age_coh    = min(par.max_age, v_first_pension_year_coh - v_birth_year_coh - 1);
    v_last_pension_age_coh     = min(par.max_age, v_last_pension_year_coh - v_birth_year_coh - 1);
    
    % pensioengrondslag-profiel
    if ~isnan(par.franchise_part)
        pension_wage_profile00 = (profiel.wage(Iwork) * (1 - par.franchise_part));
    else
        pension_wage_profile00 = (profiel.wage(Iwork) - profiel.franchise(Iwork));
    end
    if isnumeric(par.wage_inflation) && par.wage_inflation~=0 && ~isnan(par.wage_inflation)
        pension_wage_profile00 = pension_wage_profile00(:) * (1+par.wage_inflation).^(0:par.nsim_hor);
    end
    
    contr_sol = 1 - par.contr_ind;
    if contr_sol < 0
        if contr_sol < -1e-8
            error('negatieve premie solidariteitsvermogen')
        else % afrondingsverschil
            contr_sol =0;
        end
    end
    
    Inonneg_ages = [];
    if ~isnan(par.max_comp_age_return)
        Inonneg_ages = ismember(par.first_exposure_age:par.max_age, par.first_exposure_age:par.max_comp_age_return);
    end
    
    boptimalLC = any(strcmp(par.asset_mix, {'optimal lifecycle','constant obv geaggr. optimal lifecycle'}));
    bcontinue = true;
    while bcontinue
        if ~isempty(par.seed) && ~isnan(par.seed)
            rng(par.seed); % set random number generator
            rng default
        end
        var0.avg_benefitcoh  = [];
        var0.ncw_benefitcoh  = [];
        var0.pensioengroei10 = [];
        var0.pensioengroei20 = [];
        var0.mCRRA           = [];
        var0.mCRRA_report    = [];
        var0.ben1            = [];
        var0.asset_init      = [];
        var0.accrual_init    = [];
        var0.repl_rate_benavg_wageavg = [];
        var0.repl_rate_ben1_wageavg   = [];
        var0.repl_rate_ben1_wagelast  = [];
        var0.repl_rate_benlast_wagelast = [];
        var0.indexation            = [];
        var0.fundratio_beforeindex = [];
        var0.fundratio_afterindex  = [];
        var0.fundratio_0_contr     = [];
        var0.fundratio_1_contr     = [];
        var0.assets_sol = [];
        var0.sum_assets = [];
        var0.assets1    = [];
        var0.percVherv_init   = [];
        var0.assets_end       = [];
        var0.profielstock1    = [];
        var0.stockexposure    = [];
        var0.bondexposure     = [];
        var0.meanfundstockmix = [];
        var0.meanfundbondmix  = [];
        var0.proj_return_eff_opslag1 = [];
        var0.proj_return_eff_opslag_end = [];
        var0.toed_weight1 = [];
        var0.toed_weight  = [];
        var0.toed_exp1 = [];
        var0.toed_explast = [];
        var0.toed_stock_exp1 = [];
        var0.toed_stock_explast = [];

        vsims      = [repmat(par.nsim, nloops-1, 1); nsim0 - (nloops-1)*ceil(nsim0/nloops) ]; 
        vwght_loop = vsims(:)'/sum(vsims);
        
        ncells = length(par.disp_coh1 : par.disp_step : par.disp_coh2) + par.disp_last + (par.disp_dkg_before>0)+ (par.disp_dkg_after>1) + (par.disp_pfreturn>0);
        mean_Uval = cell(ncells,1);
        mean_val  = cell(ncells,1);
        std_val   = cell(ncells,1);
        min_val   = cell(ncells,1);
        val5      = cell(ncells,1);
        median_val= cell(ncells,1);
        val95     = cell(ncells,1);
        max_val   = cell(ncells,1);
        
        profiel0  = profiel;
        
        % loop alleen nodig als anders teveel geheugen noig is
        for loop=1:nloops
        
        profiel = profiel0;
            
        if nloops>1
            if loop==1
                disp([num2str(nloops) ': loops = '])
            end
            fprintf([num2str(loop) '-'])
        end
        par.nsim = vsims(loop);
        if ~(par.smoothper_indiv>1)
            vmatur = [0 1 2 5 20]; % looptijden obligaties
            % voeg evt extra looptijden toe:
            if isnumeric(par.bondmat) && ~isnan(par.bondmat)
                vmatur = unique([vmatur  par.bondmat]); 
            end
            if isnumeric(par.maturity_sol) && ~isnan(par.maturity_sol)
                vmatur = unique([vmatur  par.maturity_sol]); 
            end
            vmatur = vmatur(~isnan(vmatur) & vmatur < par.max_age-par.first_work_age);
        else  % alle looptijden zijn nodig met uitsmeren
            vmatur = 0:(par.max_age-par.first_work_age);
        end
        if strcmp( par.cash_mat,'stepyr' )
             par.cash_mat = 1;
        end
        economy   = read_set(spath, par, vmatur,false); % lees economische variabelen in
        brenteafd = strcmp(par.bondmat,'renteafdekking') || strcmp(par.maturity_sol,'renteafdekking');
         maxRTS = num_ages;
        % langere looptijd (max 100 jr) mogelijk als obl.looptijd wordt gekozen,
        %  zodat obl.port. zelfde looptijd heeft (hogere) volatiliteit van verplichtingen te krijgen:
        if brenteafd
            maxRTS = max(100+1,num_ages);
        end
        % aantal simulatiejaren en aantal simulaties
        [nsim,~,num_sim_prds1] = size(economy.RTS);
        I = (economy.RTS_maturities < maxRTS);
        if isnumeric(par.set_RTS)
            RTSset = 1+repmat(par.set_RTS, [ 1 maxRTS ] );
        else
            RTSset = economy.RTS(:,I,:);
        end
        clear maxRTS
        economy     = rmfield(economy,'RTS'); % bespaar geheugen
        economy.RTS_maturities = economy.RTS_maturities(I);
        IRTSnumages = (economy.RTS_maturities < num_ages);
        if strcmp(par.discount_rate,'RTS') || all(isnan(par.discount_rate))
            mdiscRTS = RTSset(:,IRTSnumages,:);
        else
            mdiscRTS     = (1+par.discount_rate);
        end
        if isnumeric(par.disconto)
            discount_mat_util = par.disconto;
        else
            discount_mat_util = mdiscRTSopslaanspr;
        end
        % waarde annuiteit per opbouwperiode, nodig om opbouw om te zetten naar verplichtingen
        % dimensie 1: simulatiejaren, dimensie 2: leeftijd (ivm sterftekans), dimensie 3: simulatienr
        if strcmp(par.contract,'4A')
            Psurv0                  = ones(size(profiel.Psurv));
            Psurv0(:,2:end,:)       = profiel.Psurv(:,1:end-1,:); % start-of-period overlevingskansen
        else
            [annuitypricePsurv0, annuitypricePsurv1, Psurv0] = getAnnuityPrices(          profiel.Psurv  , mdiscRTS, num_work_periods, 1, true, 1, 0);
        end
        if strcmp(par.bondmat,'matching') || strcmp(par.maturity_sol,'matching') || ( all(~isnan(par.hist_returns)) && ~par.startjong && (strcmp(par.contract(1),'1') || strcmp(par.init_method,'degressieve opbouw') || strcmp(par.init_method,'doorsnee opbouw') ) )
            [annpricePsurv0_b_init, ~, Psurv0_b] = getAnnuityPrices(profiel.Psurv, mdiscRTS(:,:,1), num_work_periods, 1, true, 1, true, [],[]); % discount_mat_mkt
        end
        proj_return_eff_opslag = [];
        if strcmp(par.contract,'4A')
            % aanname: shortgaan altijd tegen cash-rente
            Itmp               = (profiel.bond<0);
            profiel.cash(Itmp) = profiel.cash(Itmp) + profiel.bond(Itmp);
            profiel.bond(Itmp) = 0;
            clear Itmp;

            profiel_stock0 = profiel.stock;
            profiel_bond0 = profiel.bond;
            profiel_cash0 = profiel.cash;
            payout_growth_profile = 0;
            if isnumeric(par.proj_return_opslagfactor * par.proj_return_opslag)
                proj_return_eff_opslag = 1 + par.proj_return_opslagfactor * par.proj_return_opslag;
            end
            if strcmp(par.smoothper_indiv_meth, 'geheugenloos') % geheugenloos uitsmeren met gelijke aanpassingen
                uitsmeersaldo      = NaN(par.nsim, par.nsim_hor+1);
                % gemeten aan eind van periode:
                uitsmeersaldo(:,1) = par.init_uitsmeersaldo;
                [~, vuitsmeer]     = getmuitsmeer(par.smoothper_indiv , par.smoothper_indiv_meth, num_work_periods, num_pension_periods);
                if isempty(proj_return_eff_opslag)
                    % initialiseer a^o/u met s=0:
                    if par.adj_stock_smooth
                        [annpricePsurv0_uitsm_init, ~, ~]    = getAnnuityPrices(profiel.Psurv, mdiscRTS(:,:,1), num_work_periods, 1, true, vuitsmeer, true, [], proj_return_eff_opslag);
                        annuitypricePsurv_ou = annpricePsurv0_uitsm_init./annpricePsurv0_b_init;
                        profiel.stock        = annuitypricePsurv_ou .* profiel_stock0;
                    end
                    [~, ~, ~, proj_return_eff_opslag] = getProjReturnOpslag(par, profiel);
                end
            else
                uitsmeersaldo = []; % zonder uitsmeren
                if isempty(proj_return_eff_opslag) % bepaal evt (mult.plic.) opslag op RTS
                    [proj_return_eff_opslag, ~, ~] = getProjReturnOpslag(par, profiel);
                end
                vuitsmeerqnoemer = 0;
                % bepaal hoogte uitkering uit vermogen:
                if isnumeric(par.proj_return)
                    annuitypricePsurv_payout  = getAnnuityPrices(profiel.Psurv, par.proj_return, num_work_periods, payout_growth_profile, false, 1, vuitsmeerqnoemer, [], proj_return_eff_opslag);
                else
                    annuitypricePsurv_payout  = getAnnuityPrices(profiel.Psurv, RTSset(:,IRTSnumages,min(size(RTSset,3),2)), num_work_periods, payout_growth_profile, false, 1, vuitsmeerqnoemer, [], proj_return_eff_opslag);
                end
            end
        end
        clear discount_mat_proj payout_growth_profile vuitsmeerq;
        
        % initialiseer vermogen en/of aanspraken:
        var.asset_init   = zeros(nsim, num_ages);
        var.accrual_init = zeros(nsim, num_ages);
        if all(~isnan(par.hist_returns)) && ~par.startjong % include historic returns
            if (strcmp(par.contract(1),'1') || strcmp(par.init_method,'degressieve opbouw') || strcmp(par.init_method,'doorsnee opbouw')) % historische opbouw toevoegen
                if all(strcmp(par.init_discount_rate,'mean RTS')) || all(isnan(par.init_discount_rate))
                    parinit_discount_rate = mean(RTSset(:,1:num_ages,1));
                else
                    parinit_discount_rate = repmat(par.init_discount_rate, 1, num_ages);
                end
                if all(strcmp(par.hist_returns,'mean RTS'))
                    parhist_returns = mean(RTSset(:,1:num_ages,1));
                else
                    parhist_returns = repmat(par.hist_returns, 1, num_ages);
                end
                % degressieve opbouwstaffel per periode:
                accrual_period = zeros(nsim, num_ages);
                for t=1:num_work_periods
                    accrual_period(:,t)  = 1./sum( Psurv0_b(Ipension)/Psurv0_b(t) ./ (1 + parinit_discount_rate(:,(num_work_periods:num_ages-1)-t+1)).^((num_work_periods:num_ages-1)-t+1));
                end
                if strcmp(par.init_method,'doorsnee opbouw')
                    mtmp                    = annpricePsurv0_b_init(:,Iwork) .* pension_wage_profile00 .* Psurv0_b(Iwork);
                    accrual_period(:,Iwork) = repmat( sum( accrual_period(:,Iwork) .* mtmp, 2) ./ sum(mtmp, 2) , 1, num_work_periods);
                    clear mtmp
                end
                indexatievoet = cumprod( (1+parhist_returns)./(1+parinit_discount_rate), 2);
                for t=1:num_ages-1
                    var.accrual_init(:, t+1) = sum( fliplr( indexatievoet(:,1:t) ) .* par.contr_rate .* (profiel.wage(1:t) - profiel.franchise(1:t)) .* Psurv0(1:t) .* accrual_period(:,1:t), 2) ;
                end
                if strcmp(par.contract(1),'1')
                    var.accrual_init = contr_sol .* var.accrual_init;
                end
                switch par.invaarq
                    case 'asymptotisch' 
                        q_invaar0 = 1 - (1 - 1/par.init_smoothper) .^ (1:num_ages);
                    case 'lineair'
                        q_invaar0 = min(1, (1:num_ages)/par.init_smoothper);
                end
                if strcmp(par.contract(1),'4')
                    [var.asset_init, var.percVherv_init, ~]     = distribute_initialassets(par, var.accrual_init, squeeze(RTSset(:,IRTSnumages,min(size(RTSset,3),2))), ...
                        mean(proj_return_eff_opslag(:,:),1), q_invaar0, false, false);
                else
                    % ftk-waarde:
                    var.asset_init     = var.accrual_init .* squeeze( annpricePsurv0_b_init );  %  (:,:,1)   annuitypricePsurv0
                    clear accrual_period   Psurv0_init  parhist_returns  parinit_discount_rate
                    var.percVherv_init = sparse([],[],[], size(var.asset_init,1), size(var.asset_init,2));
                    % reken annuiteiten om naar vermogens
                    if strcmp(par.invaarmethode, 'standaardmethode')
                        % vertaal initile aanspraken naar verplichtingen.
                        % aanname: zelfde RTS in alle scenario's
                        v_initRTS        = mean(RTSset(:,1:num_ages,1).^-economy.RTS_maturities(1:num_ages),1)'; % [ 0;
                        % m_init_liab: initile verdisconteerde verplichting, dim1=scenario's, dim2=leeftijden, dim3=horizon h.
                        m_init_liab      = zeros(nsim,num_ages,num_ages);
                        for t=1:num_ages
                            I1        = (num_work_periods+1:num_ages)-t+1;
                            I1(I1<=0) = [];
                            m_init_liab(:,t,I1) = var.accrual_init(:,t)* v_initRTS(I1)';
                        end
                        q_invaar            = permute(q_invaar0, [3 1 2]);
                        % cohort_herstelcap is een naar q gewogen gemiddelde van aanpraken van cohort_herstelcap
                        cohort_herstelcap= sum(q_invaar .* m_init_liab,3) ./ sum( m_init_liab , 3);
                        cohort_herstelcap(isnan(cohort_herstelcap)) = 1;
                        % cohort_herstelcap is een naar q gewogen gemiddelde van cohort_herstelcap
                        herstelcap_fonds = sum( q_invaar .* sum(m_init_liab,2), 3) ./ sum( sum( m_init_liab, 2), 3);
                        cohort_dkg       = 1 +  (par.init_dkg - 1) * cohort_herstelcap./herstelcap_fonds;
                        var.asset_init       = cohort_dkg .* var.asset_init;
                        clear cohort_herstelcap herstelcap_fonds I1  m_init_liab  q  v_initRTS  cohort_dkg q_invaar;
                    end
                end
                clear q_invaar0
            else
                % schaal naar huidige fondsvermogen obv  par.init_dkg
                var.asset_init = par.init_dkg * var.asset_init;
            end
        else
            if ~par.startjong
                for t=1:num_work_periods+1
                    v = mean(sum((par.contr_rate * pension_wage_profile00(1:t-1,1)).*( fliplr( squeeze( RTSset(:,1:t-1,1) ).^(1:t-1) )),2));
                    var.asset_init(:,t) = repmat(v,[nsim 1]);
                    var.asset_init(:,num_work_periods+1:end) = repmat(linspace(v, v/num_pension_periods, num_pension_periods),[nsim 1]);
                    clear v;
                end
            end
        end
        if  strcmp(par.smoothper_indiv_meth, 'geheugenloos') || (~(par.toed>0))
            clear RTSset
        end
        clear asset_init0 annpricePsurv0_b_init;
        if any(strcmp(par.asset_mix, {'optimal lifecycle','constant obv geaggr. optimal lifecycle'})) || ismember(par.toed,1:3)
            % netto waarde toekomstige pensioenpremies,
            % voor betaling van premie in huidige periode.
            I2              = (economy.RTS_maturities(I) >= 0 & economy.RTS_maturities(I) < par.pension_age-par.first_exposure_age);
            if all(all(profiel.Psurv(:,1:end-1,:)==1))
                mat_discount0    = cumsum( ( (1+par.est_wage_inflation) ./ mdiscRTS(:, I2, :) ).^( economy.RTS_maturities(I2) ), 2);
                mat_discount0    = mat_discount0(:,end:-1:1,:);
            else
                mat_discount0 = NaN(nsim, num_work_periods, num_sim_prds1);
                for j=find(I2) % onconditionele sterftekansen, zodat rendement op human capital niet afhangt van overleven
                    mat_discount0(:,j,:)    = sum( Psurv0(:,j:find(I2,1,'last'),:) .* (mdiscRTS(:, 1:find(I2,1,'last')+1-j, :) .* (1+par.est_wage_inflation)).^(0:(find(economy.RTS_maturities(I2))-j)), 2)./Psurv0(:,j); % ./Psurv0(:,j);
                end
            end
        end
        if ~(brenteafd && strcmp(par.contract,'1A'))
            clear discount_mat
        end
        if loop==1
            disp(' ')
            disp(par.sheetname)
            disp(['asset init, gem: ' num2str(mean(sum(var.asset_init,2))) ', std: ' num2str(std(sum(var.asset_init,2)))])
            disp(' ')
        end
        
        v_num_work_periods_coh = max(0,min(num_work_periods, v_last_pension_period_coh - v_num_pension_periods_coh - 1));
        for s = {'bond_sol','hcap_sol','cash_sol'}
            if isempty(par.(s{1})) || isnan(par.(s{1}))
                par.(s{1})    = 0;
            end
        end
        % initialiseer:
        % uitkeringen
        benefit_periods = zeros(nsim,num_pension_periods,num_sim_prds1-1);
        % initialiseer premie-matrix
        contr = zeros( size(economy.wage_index,1), num_ages, num_sim_prds1-1);
        if strcmp(par.contract(1),'1')
            % opbouwprofiel
            accr_profile   = 1 ./ annuitypricePsurv0(:, Iwork, :); % degressieve opbouw
            % verplichtingen
            accrualPVsum    = zeros(nsim, num_sim_prds1);
            % uitbetalingen
            benefit_periods(:,:,1)  = var.accrual_init(:, Ipension);
            benefitsum      = zeros(nsim, num_sim_prds1);
            benefitsum(:,1) = squeeze(sum(benefit_periods(:,:,1),2))';
            % verwaarloos indexatie bij initialisatie
            accrualPVsum(:,1) = squeeze( sum( annuitypricePsurv0(:,:,1) .* var.accrual_init(:,:,1), 2) ) - benefitsum(:,1);
            accrualPVsum_end  = accrualPVsum;
            % dekkingsgraad
            var.fundratio_beforeindex = repmat(par.init_dkg, [nsim num_sim_prds1-1]);
            var.fundratio_afterindex  = repmat(par.init_dkg, [nsim num_sim_prds1-1]);
            var.fundratio_0_contr     = NaN(nsim,num_sim_prds1-1);
            var.fundratio_1_contr     = NaN(nsim,num_sim_prds1-1);
            % indexatiematrix, allemaal 1 betekent geen vaste indexatie maar evt
            % wel obv dekkingsgraad.
            indexation0 = ones(nsim, 1, num_sim_prds1);
            % indexatiematrix
            var.indexation  = ones(nsim, num_sim_prds1);
            if strcmp(par.contract,'1A')
                % tekortjaren
                lowyears = zeros(nsim, num_sim_prds1);
                lowyears(:,1) = par.initial_shortage_yrs;
                % kortingsjaren
                cutyears = zeros(nsim, num_sim_prds1);
                % onconditioneel korten
                unc_cut1A = zeros(nsim, num_sim_prds1);
            end
            % opbouw, start op t=2 met doorschuiven van opbouw
            accrual              = zeros(nsim, num_ages, num_sim_prds1);
            accrual(:,1:end-1,1) = var.accrual_init(:,2:end);
            accrual_pensdate     = NaN(nsim, num_sim_prds1);
            if ~strcmp(par.maturity_sol,'renteafdekking')
                economy = rmfield(economy,'bond_returns');
            end
        else
            accrualPVsum            = [];
            accrualPVsum_end        = [];
            var.fundratio_beforeindex = [];
            var.fundratio_afterindex  = [];
            var.indexation            = [];
            var.fundratio_0_contr     = [];
            var.fundratio_1_contr     = [];
            economy = rmfield(economy,'bond_returns');
        end
        
        if strcmp(par.contract(1),'1')
            assets      = [];   % geen persoonlijk pensioenvermogen
        else
            % PPV
            assets                = zeros(nsim, num_ages, num_sim_prds1-1);
            assets(:, 1:end-1, 1) = var.asset_init(:, 2:end);
        end
        if par.bwriteallstats
            var.meanfundstockmix = NaN(nsim           , num_sim_prds1-1);
            bondbyage            = (par.contract(1)=='4' && par.toed==0) || strcmp(par.bondmat,'matching');
            meanfundbondmix      = NaN(num_sim_prds1-1, (num_ages + strcmp(par.bondmat,'renteafdekking'))*bondbyage + 2*~bondbyage);
        else
            var.meanfundstockmix = [];
            meanfundbondmix      = [];
        end
        if strcmp(par.contract,'4A')
            % geen solidariteitsvermogen
            var.assets_sol = [];
        else
            % initialiseer collectieve vermogen
            var.assets_sol  = zeros(nsim, par.nsim_hor+1);
            if any(~isnan(par.hist_returns)) && strcmp(par.contract(1),'1')
                var.assets_sol(:,1)  = sum(var.asset_init,2);
            end
        end
        if i > 1 || ~exist('stat_col','var') || isempty(stat_col)
            stat_col = sheet_compute+1;
        end
        if ~isnan(par.cut_t0) && par.cut_t0
            if strcmp(par.contract(1),'1') || par.cut_t0 > 1 % par.cut_t0==2: ook in 4-contracten voor invaren korten
                t     = 1;
                tann1 = 1;
                % pas opbouw aan door indexatie en korting:
                [accrualPVsum(:,t), ~, ~, accrual_init0, var.fundratio_beforeindex(:,t), var.fundratio_afterindex(:,t),var.indexation(:,t), cutyears(:,t), lowyears(:,t), unc_cut1A(:,t)] = ...
                    adjust_accrual( squeeze(var.accrual_init), squeeze(annuitypricePsurv0(:,:,tann1)), var.assets_sol(:,t), ...
                    economy.price_inflation(:,1), squeeze(indexation0(:,:,t)), cutyears(:,max(t-1,1)), lowyears(:,max(t-1,1)), unc_cut1A(:,max(t-1,1)), par );
                accrual(:,1:end-1,1) = accrual_init0(:,2:end);
                clear accrual_init0
            end
        end
        
        ret_fund    = NaN(nsim, num_sim_prds1-1);
        hcap        = 0;
        % zelfde stappen als ALM-model CPB, zie Michielsen (2015)
        % https://www.cpb.nl/sites/default/files/publicaties/download/cpb-background-document-september2015-cpb-asset-liability-management-model-pension-analyses.pdf
        % 1 ontvang premies
        % 2 betaal uitkeringen
        % 3 verwerk financiele rendementen
        % 4 bepaal korting en indexatie

        % eerste dimensie = scenario's
        % tweede dimensie = leeftijd
        % derde dimensie  = simulatiejaren
        % diagonalen      = cohorten

        for t=2:par.nsim_hor+1
            if isnumeric(par.set_RTS)
                tann  = 1;
                tann1 = 1;
            else
                tann  = t;
                tann1 = t+1;
            end
            % pensioengrondslag loonprofiel
            if ~strcmp(par.wage_inflation,'scenarioset')
                pension_wage_profile = pension_wage_profile00(min(t,size(pension_wage_profile00,1)),:);
            else
                pension_wage_profile = economy.wage_index(:,t) * pension_wage_profile00(min(t,size(pension_wage_profile00,1)),:);
            end
            % eenjarige overlevingskansen in t-1
            Psurv1 = profiel.Psurv(:,:,min(t-1,end)) ./ [size(profiel.Psurv,1) squeeze(profiel.Psurv(:,1:end-1,min(max(t-2,1),end)))];

            % pensioengrondslag loonprofiel startend met startsalaris voor evt ongeborenen
            pension_wage_profile0   = pension_wage_profile;
            if par.first_work_age > par.first_exposure_age
                min_work_age            = par.first_work_age - par.first_exposure_age + 1; % aanname: evt ongeborenen rekenen met startsalaris
                pension_wage_profile0(:, 1:(min_work_age-1)) = pension_wage_profile(:, min_work_age);
            end
            % bepaal rendement op human capital
            Iworkt = Iwork;
            if par.startjong % nu is toekomstig loon niet meer nodig
                pension_wage_profile(:,t:end) = 0;
                Iworkt(t:end) = false;
            end
            Ipensiont = ~Iworkt;
            % bepaal pensioenpremie
            contr(:,Iwork,t)           = squeeze(par.contr_rate .* pension_wage_profile .* Psurv0(:,Iwork,min(t,size(Psurv0,3))));
            if strcmp(par.bondmat,'matching') || strcmp(par.maturity_sol,'matching')
                % *00: gemeten periode t vlak voor uitkering
                % *01: gemeten periode t vlak na uitkering
                % *10: gemeten periode t+1 vlak voor uitkering
                if t > 2
                    annpricePsurv_b00 = annpricePsurv_b10;
                else
                    [annpricePsurv_b00, ~] = getAnnuityPrices(profiel.Psurv, mdiscRTS(:,:,tann) , num_work_periods, 1, true, 1, true, [],proj_return_eff_opslag);
                end
                [annpricePsurv_b10, ~] = getAnnuityPrices(profiel.Psurv, mdiscRTS(:,:,tann1), num_work_periods, 1, true, 1, true, [],proj_return_eff_opslag);
            end
            if strcmp(par.contract(1),'4')
                if strcmp(par.smoothper_indiv_meth, 'geheugenloos')
                    [annpricePsurv_b01, ~] = getAnnuityPrices(profiel.Psurv, mdiscRTS(:,:,tann) , num_work_periods, 1, true, [0 ones(1,size(mdiscRTS,2)-1)], true, [],proj_return_eff_opslag); % discount_mat_mkt
                    if t > 2
                        annpricePsurv0_uitsm00 = annpricePsurv0_uitsm10;
                    else
                        [annpricePsurv0_uitsm00, ~] = getAnnuityPrices(profiel.Psurv, mdiscRTS(:,:,tann), num_work_periods, 1, true, vuitsmeer        , true, [], proj_return_eff_opslag); % discount_mat_mkt ret10_uitsm
                    end
                    [annpricePsurv0_uitsm01, ~] = getAnnuityPrices(profiel.Psurv, mdiscRTS(:,:,tann), num_work_periods, 1, true, [0 vuitsmeer(1:end-1)], true, [], proj_return_eff_opslag); % discount_mat_mkt ret10_uitsm
                    [annpricePsurv0_uitsm10, ~] = getAnnuityPrices(profiel.Psurv, mdiscRTS(:,:,tann1), num_work_periods, 1, true, vuitsmeer, true, [], proj_return_eff_opslag); % discount_mat_mkt ret10_uitsm
                end
                % begin nieuwe periode met individuele vermogen uit vorige periode:
                assets(:,2:end,t) = assets(:,1:end-1,t-1); % .* repmat( 1-Pdie(1:end-1), [1 1 nsim]); SMU: door biometrisch rendement hier geen aanpassing
                % voeg pensioenpremie toe
                assets(:,Iwork,t) = assets(:,Iwork,t) + par.contr_ind .* contr(:,Iwork,t);
                if par.toed==1  && ~strcmp(par.smoothper_indiv_meth, 'geheugenloos')
                    profiel.stock   = getStockExposureToed(assets(:,:,t), hcap, profiel_stock0, profiel.toed_staffel);
                end
                if strcmp(par.contract,'4A')
                    % bepaal uitkering als WVP met annuiteitsfactor van projectierente
                    if strcmp(par.smoothper_indiv_meth, 'geheugenloos') % geheugenloos met gelijke aanpassingen
                        annpricePsurv_u00 = annpricePsurv_b00 + uitsmeersaldo(:,t-1) .* annpricePsurv0_uitsm00; % a^u(s)
                        % werk a^o/u bij voor geheugenloos uitsmeren:
                       if par.adj_stock_smooth
                           annuitypricePsurv_ou = annpricePsurv0_uitsm00./annpricePsurv_u00;
                           profiel.stock        = [annuitypricePsurv_ou(:,2:end)  zeros(size(annuitypricePsurv_ou,1),1)] .* profiel_stock0;
                           % schaal cash en bonds met verandering in profiel.stock:
                           xb                   = max(profiel_bond0,0) ./ (max(profiel_bond0,0) + max(profiel_cash0,0));
                           xb(isnan(xb))        = 1;
                           % neem aan niet short in bonds:
                           profiel.bond         = xb.*max(1-profiel.stock,0);
                           clear xb
                           profiel.cash         = 1 - profiel.stock - profiel.bond;
                       end
                    else
                        if size(RTSset,3)>1 || par.toed>0 % update projectierendementen
                            payout_growth_profile = 0;
                            if isnumeric(par.proj_return)
                                % discount_mat_proj = par.proj_return; % .^repmat(-economy.RTS_maturities(I),[par.nsim 1 num_sim_prds1]);
                                annpricePsurv_u00  = getAnnuityPrices(profiel.Psurv, par.proj_return, num_work_periods, payout_growth_profile, false, 1, vuitsmeerqnoemer, [], proj_return_eff_opslag);
                            else
                                % discount_mat_proj = economy.RTS(:,I,:); % .^repmat(-economy.RTS_maturities(I),[par.nsim 1 num_sim_prds1]);
                                annpricePsurv_u00  = getAnnuityPrices(profiel.Psurv, RTSset(:,IRTSnumages,tann), num_work_periods, payout_growth_profile, false, 1, vuitsmeerqnoemer, [], proj_return_eff_opslag);
                            end
                        else
                             annpricePsurv_u00 = annuitypricePsurv_payout;
                        end
                    end
                    benefit_t       = assets(:,Ipension,t) ./ annpricePsurv_u00(:,Ipension);
                    clear annpricePsurv_u00;
                    if strcmp(par.smoothper_indiv_meth, 'geheugenloos')
                        % uitsmeerstap 5:
                        benefit_t          = (1 + uitsmeersaldo(:,t-1)/par.smoothper_indiv) .* benefit_t;
                    end
                    if par.smoothper>1 && t > 2
                        benefit_periods(:, 1, t)       = benefit_t(:,1);  % benefit_periods(:, 1, t) + 
                        % uitsmeren als er vorige periode al een uitkering was
                        benefit_periods(:, 2:end-1, t) = benefit_periods(:,1:end-2,t-1) + (benefit_t(:,2:end-1) - benefit_periods(:,1:end-2,t-1)) / par.smoothper;
                        benefit_periods(:, end, t)     = benefit_t(:,end);
                    else
                        benefit_periods(:,:, t)        = benefit_t;
                    end
                    % nooit meer uitbetalen dan aanwezig in persoonlijk pensioenvermogen;
                    benefit_periods(:,:,t)       = min(benefit_periods(:,:, t), assets(:,Ipension,t) );
                    % betaal uitkering uit individuele vermogens
                    assets(:, Ipension,t) = assets(:,Ipension,t) - benefit_periods(:,:,t);
                end
            end
            if any(strcmp(par.asset_mix, {'optimal lifecycle','constant obv geaggr. optimal lifecycle'})) || ismember(par.toed,1:2)
                dim3tmp    = min(t  , size(mat_discount0,3));
                dim3tmp1   = min(t+1, size(mat_discount0,3));
                % bepaal rendement op human capital, -1 aftrekken omdat vorige premieinleg niet meer meetelt. * Psurv1 omdat het rendement anders conditioneel op overleven is:
                hcap_ret_t = ones(nsim, num_ages);
                mtmp       = Psurv1(:,1:num_work_periods-1) .* ( pension_wage_profile0(:,2:end)*(1+par.est_wage_inflation) ) .* squeeze( mat_discount0(:, 2:end  , dim3tmp1 ) ) ...
                                                ./ ( pension_wage_profile0(:,1:end-1)  .* squeeze( mat_discount0(:, 1:end-1, dim3tmp ) - 1 ) );
                hcap_ret_t(:,1:num_work_periods-1) = repmat(mtmp, size(hcap_ret_t,1)/size(mtmp,1), 1);
                hcap_ret_t(isnan(hcap_ret_t) | isinf(hcap_ret_t)) = 1;

                hcap       = Psurv0(:,Iwork) .* par.contr_rate .* ( pension_wage_profile0 ) .* squeeze( mat_discount0(:,:,dim3tmp) - 1 );
                hcap(:,~Iworkt & Iwork) = 0;
                clear mtmp
                % dit klopt: hcap_ret_t == (hcap(:,2:end)+par.contr_rate*pension_wage_profile0(2:end)'.*Psurv0(:,2:num_work_periods))./hcap(:,1:end-1)
            else
                hcap_ret_t = 1;
            end
            clear dim3tmp dim3tmp1 mtmp
            if strcmp(par.contract(1),'1')
                % begin nieuwe periode met sol. vermogen en opbouw uit vorige periode:
                var.assets_sol(:,t)      = var.assets_sol(:,t-1);
                % voeg premie toe aan solidariteitsvermogen
                var.assets_sol(:,t)      = var.assets_sol(:,t) + sum( contr_sol .* contr(:,:,t), 2);
                accrual(:, 2:end, t) = accrual(:,1:end-1,t-1) .* Psurv1(:,1:end-1);
                % voeg nieuwe opbouw toe
                vtmp = 1;
                if ~isnan(par.adjaccrfactor) && par.adjaccrfactor~=0
                    % evt dekkingsgraadneutraal inkopen
                    vtmp = (1 + (1./var.fundratio_afterindex(:,t-1)-1) * par.adjaccrfactor);
                    vtmp(isnan(vtmp)) = 1;
                end
                accrualnew            = vtmp .* contr_sol .* par.contr_rate .* accr_profile(:,Iworkt,tann) .* repmat(pension_wage_profile(Iworkt), [nsim 1]) .* Psurv0(:,Iworkt,min(t,size(Psurv0,3)));
                if strcmp(par.opbouwsystematiek,'doorsnee opbouw')
                    mattmp            = squeeze(annuitypricePsurv0(:,Iworkt,tann)) .* pension_wage_profile00 .* Psurv0(Iworkt);
                    accrualnew        = sum( accrualnew .* mattmp, 2) ./ sum( mattmp, 2);
                end
                accrual(:, Iworkt, t) = accrual(:, Iworkt, t) + accrualnew;
                clear accrualnew mattmp vtmp 

                % bepaal uitkering aan de hand van opbouw en evt AOW
                benefit_t               = accrual(:, Ipension, t);
                benefit_periods(:,:,t)  = benefit_periods(:,:,t) + benefit_t;
                benefitsum(:,t)         = squeeze(sum(benefit_t,2))';
                % betaal uitkering uit solidariteitsvermogen
                var.assets_sol(:,t)         = var.assets_sol(:,t) - benefitsum(:,t);

                % ken beleggingsrendementen toe aan solidariteitsvermogen
                % bepaal eerst het rendement op de obligatieportefeuille
                if strcmp(par.maturity_sol,'matching') || strcmp(par.maturity_sol,'renteafdekking')
                    switch par.maturity_sol
                        case 'matching'
                            bondret = sum( annpricePsurv_b10(:,2:end) .* accrual(:,1:end-1,t) , 2) ./ ...
                                sum( ( annpricePsurv_b00(:,1:end-1) - ... % repmat( (profiel.Psurv(:,1:end-1,t) .* ...
                                Ipension(1:end-1) ) .* accrual(:,1:end-1,t), 2);
                            bondmat_t = 'matching';
                        case 'renteafdekking'
                            % selecteer looptijd die het renterisico het beste afdekt gegeven investering in obligaties:
                            [bondret, bondmat_t] = getRenteafdekkingKX(par.bond_sol, var.fundratio_beforeindex(:,t), economy.nominal_term_structure.b, ...
                                economy.RTS_maturities, squeeze(economy.bond_returns(:,:,t)), ...
                                par.KNW.K * squeeze( economy.state_variables_term_structure(:,:,t) )', ...
                                mdiscRTS(:,:,t).^(-economy.RTS_maturities(IRTSnumages)), par.toed==0, par);
                   end
                   I = isnan(bondret);
                   if any(I)
                       if isnumeric(par.set_RTS)
                           bondret(I) = 1 + par.set_RTS;
                       else
                           if isnumeric(par.V_rInf)
                                bondret(I) = 1 + par.V_rInf;
                           else
                                bondret(I) = 1;
                           end
                       end
                   end
                else
                    bondret = squeeze( economy.bond_returns(:,economy.bond_portf_maturities==par.maturity_sol,t) );
                end
                ret_fund(:,t) = economy.stock_returns(:,t) .* par.stock_sol + ...
                    bondret .* par.bond_sol + economy.cash_return(:,t) .* par.cash_sol    + sum(hcap_ret_t .* par.hcap_sol, 2);
                % ken rendement toe aan solidariteitsvermogen
                var.assets_sol(:,t) = ret_fund(:,t) .* var.assets_sol(:,t);
                if par.bwriteallstats
                    if brenteafd
                        meanfundbondmix(t,:) = [ mean( par.bond_sol, 1)  mean( bondmat_t, 1) ];
                    else
                        meanfundbondmix(t,:) = [ mean( par.bond_sol, 1)  NaN                 ];
                    end
                    var.meanfundstockmix(:,t) = par.stock_sol;
                end
                % pas opbouw aan door indexatie en kortingen:
                [accrualPVsum(:,t), ~, ~, accrual(:,:,t), var.fundratio_beforeindex(:,t), var.fundratio_afterindex(:,t),var.indexation(:,t), cutyears(:,t), lowyears(:,t), unc_cut1A(:,t)] = ...
                        adjust_accrual( squeeze(accrual(:,:,t)), squeeze(annuitypricePsurv1(:,:,tann1)), var.assets_sol(:,t), ...
                         economy.price_inflation(:,min(max(1,t-1),size(economy.price_inflation,2))), squeeze(indexation0(:,:,t)), cutyears(:,max(t-1,1)), lowyears(:,max(t-1,1)), unc_cut1A(:,max(t-1,1)), par );
                if par.age_index
                    daccrual0               = sum( annuitypricePsurv1(:,:,tann1) .* accrual0 .* ( var0.indexation(:,t) - 1), 2);
                    % corrigeer mate van indexatie voor human capital:
                    %  [V(L) + H(L)]/V(L) =approx= 1 + (jaren tot pensioen geteld vanaf volgende periode))/werkjaren
                    dindexation1            = (var.indexation(:,t)-1) * (1 + fliplr(cumsum(fliplr(Iwork(2:end))))./cumsum(Iwork(1:end-1)));
                    daccrual1               = sum( annuitypricePsurv1(:,:,tann1) .* accrual0 .* dindexation1, 2);
                    scalefactor             = daccrual0 ./ daccrual1;
                    accrual(:,1:end-1,t)    = accrual0(:,1:end-1)  + scalefactor .* accrual0(:,1:end-1) .* dindexation1;
                    clear fundratio_afterindex0 scalefactor
                end
                clear accrual0
            end
            var.stockexposure = [];
            var.bondexposure  = [];
            if strcmp(par.contract(1),'4')
                if boptimalLC || ismember(par.toed,1:3)
                    if par.toed~=3 || ~isequal(par.asset_mix,'lineair')
                        % total wealth = financial wealth + human capital
                        total_wealth_t = NaN(nsim, num_ages);
                        % werkenden, aanname is dat beleggingsbeleid zelfde deel van human capital afdekt als de toedeling
                        total_wealth_t(:,Iworkt)    = assets(:,Iworkt,t) + hcap(:,Iworkt);
                        % gepensioneerden
                        total_wealth_t(:,Ipensiont) = assets(:,Ipensiont,t);
                    end
                    var.stockexposure = zeros(nsim, num_ages);
                    var.bondexposure  = zeros(nsim, num_ages);
                    if ~isequal(par.asset_mix,'lineair')
                        % exposures tov total_wealth tellen op tot n
                        c_tmp           = par.BS_risk_par / (par.BS_vol_par * par.gamma);
                        var.stockexposure =     c_tmp   .* total_wealth_t ./ assets(:,:,t);
                        var.bondexposure  = (1 - c_tmp) .* total_wealth_t ./ assets(:,:,t);
                        clear c_tmp
                        if ~isnan(par.max_exposure)
                            var.stockexposure = min(par.max_exposure, var.stockexposure);
                            var.bondexposure  = min(par.max_exposure, var.bondexposure);
                        end
                        var.stockexposure(assets(:,:,t)==0) = 0; % voorkom NaN als geen assets
                        var.bondexposure(assets(:,:,t)==0)  = 0; % voorkom NaN als geen assets

                        if isnumeric(par.asset_mix)
                            profiel.stock = repmat( par.asset_mix   , [num_ages 1]) ;
                            profiel.bond  = repmat( 1-par.asset_mix , [num_ages 1]) ;
                        else
                            if strcmp(par.asset_mix,'constant obv geaggr. optimal lifecycle')
                                profiel.stock = squeeze( repmat( sum( var.stockexposure .* assets(:,:,t) , 2) ./ sum( assets(:,:,t), 2), [1 num_ages]) );
                                profiel.bond  = squeeze( repmat( sum( var.bondexposure  .* assets(:,:,t) , 2) ./ sum( assets(:,:,t), 2), [1 num_ages]) );
                            else
                                profiel.stock = squeeze( var.stockexposure );
                                profiel.bond  = squeeze( var.bondexposure );
                            end
                        end
                    end
%                     if par.toed==1  && ~strcmp(par.smoothper_indiv_meth, 'geheugenloos')
%                         profiel.stock = getStockExposureToed(assets(:,:,t), hcap, 1 .* profiel_stock0, profiel.toed_staffel);
%                     else
                        profiel.stock = rescale(profiel.stock, assets(:,:,t), Inonneg_ages);
%                     end
                    % nooit meer dan 100% stocks:
                    excessf = sum( max(0, profiel.stock - 1) .* assets(:,:,t), 2) ./ sum( max(0, 1 - profiel.stock) .* assets(:,:,t), 2);
                    II      = (excessf>1);
                    profiel.stock(II,:) = min(1,profiel.stock(II,:)) + max(0, profiel.stock(II,:)-1)./excessf(II);
                    clear excessf II;
                    if strcmp(par.bondmat,'cash')
                        profiel.bond = zeros(size(profiel.stock));
                    else
                        profiel.bond = rescale(profiel.bond, assets(:,:,t), Inonneg_ages);
                    end
                    if size(profiel.bond,2)==1
                         profiel.bond =  profiel.bond';
                    end
                end
                % neem aan niet short in bonds:
                profiel.bond    = max(0, profiel.bond);
                profiel.cash    = 0;
                profiel.hcap    = 0;
                if strcmp(par.short_rate,'human capital')
                    profiel.hcap    = 1 - profiel.stock - profiel.bond;
                else
                    profiel.cash    = 1 - profiel.stock - profiel.bond;
                end
                var.stockexposure = repmat( profiel.stock, [ nsim/size(profiel.stock,1) 1]);
                var.bondexposure  = repmat( profiel.bond,  [ nsim/size(profiel.bond,1)  1]);
                if strcmp(par.smoothper_indiv_meth, 'geheugenloos') % geheugenloos uitsmeren met gelijke aanpassingen
                   % uitsmeerstap 1:
                   uitsmeersaldo(:,t)  = (1-1/par.smoothper_indiv) * uitsmeersaldo(:,t-1)./(1 + uitsmeersaldo(:,t-1)/par.smoothper_indiv);
                   annpricePsurv_u10   = annpricePsurv_b10 + uitsmeersaldo(:,t) .* annpricePsurv0_uitsm10; % a^u(s~)
                   if strcmp(par.bondmat,'matching') || strcmp(par.bondmat,'renteafdekking')
                       % verwerk rente
                       if isnumeric(par.set_RTS)
                            matchingret = repmat(1+par.set_RTS, 1, num_ages);
                       else
                           % verwerk uitbetaling:
                            annpricePsurv_u01 = annpricePsurv_b01 + uitsmeersaldo(:,t) .* annpricePsurv0_uitsm01;        % a^u(s~)
                            matchingret = [annpricePsurv_u10(:,2:end)./annpricePsurv_u01(:,1:end-1) / proj_return_eff_opslag   ones(size(annpricePsurv_u10,1),1)];
                       end
                   end
                   clear annpricePsurv_b01  annpricePsurv_u00  annpricePsurv0_uitsm00  annpricePsurv0_uitsm01
                else
                    if strcmp(par.bondmat,'matching') || strcmp(par.bondmat,'renteafdekking')
                        % verwerk rente
                       if isnumeric(par.set_RTS)
                            matchingret = repmat(1+par.set_RTS, 1, num_ages);
                       else
                           matchingret = [annpricePsurv_b10(:,2:end) ./ ...
                              ( annpricePsurv_b00(:,1:end-1) - Ipension(1:end-1) ) / proj_return_eff_opslag    ones(size(annpricePsurv_b00,1),1)];
                       end
                    else
                        matchingret     = economy.cash_return(:,t);
                    end
                end
                % bepaal feitelijk rendement
                if isnumeric(par.bondmat) && ~isnan(par.bondmat)
                    bondret = squeeze( economy.bond_returns(:,economy.bond_portf_maturities==par.bondmat,t) );
                else
                    bondret = matchingret;
                end
                if par.bwriteallstats % schrijf alle statistieken weg
                    if brenteafd
                        if bondbyage
                            meanfundbondmix(t,:) = [ sum( profiel.bond .* assets(:,:,t), 1) ./ sum( assets(:,:,t), 1)           mean(bondmat_t) ];
                        else
                            meanfundbondmix(t,:) = [ sum( sum(profiel.bond .* assets(:,:,t), 1), 2) ./ sum( sum( assets(:,:,t), 1), 2)  mean(bondmat_t) ];
                        end
                    else
                        if bondbyage
                            meanfundbondmix(t,:) = mean( getBondexposures(assets(:,:,t), par, profiel, economy.RTS_maturities, nsim, num_work_periods) );
                        else
                            meanfundbondmix(t,:) = [ sum( sum(profiel.bond .* assets(:,:,t), 1), 2) ./ sum( sum( assets(:,:,t), 1), 2)  par.bondmat ];
                        end
                    end
                    var.meanfundstockmix(:,t) = squeeze( sum(profiel.stock .* squeeze( assets(:,:,t) ), 2) ) ./ squeeze( sum( assets(:,:,t), 2) );
                    if t==2
                        var.profielstock1           = profiel.stock;
                        var.proj_return_eff_opslag1 = proj_return_eff_opslag-(proj_return_eff_opslag>1);
                    end
                end
                
                if boptimalLC
                    ret_nonstock = bondret .* profiel.bond + ... 
                        repmat( economy.cash_return(:,t), [1 num_ages]) .* profiel.cash    + hcap_ret_t .* profiel.hcap;
                    ret_t        = repmat( economy.stock_returns(:,t), [1 num_ages]) .* profiel.stock + ret_nonstock;
                else
                    ret_nonstock = bondret .* profiel.bond + economy.cash_return(:,t) .* profiel.cash + ...
                        sum( hcap_ret_t .* profiel.hcap,2);
                    ret_t        = economy.stock_returns(:,t) .* profiel.stock + ret_nonstock;
                end
                clear ret_nonstock
                ret_t(isnan(ret_t)) = 1;
                ret_t_monet         = (ret_t-1) .* assets(:,:,t);
                ret_fund(:,t)       = 1 + sum(ret_t_monet,2) ./ sum( assets(:,:,t), 2) ;
                if ismember(par.toed,1:3) || strcmp(par.smoothper_indiv_meth, 'geheugenloos') % (leeftijdsafhankelijk) toedelen van rendementen
                    annuityreturn_t = matchingret;
                    if ~isnumeric(par.toed_renteafdekking) && ismember(par.toed_renteafdekking,{'als obligatielifecycle','linear lifecycle'})
                        % zoals in CPB (2020,p.5), https://www.cpb.nl/sites/default/files/omnidownload/CPB-Notitie-juni2020-Nieuwe-pensioenregels.pdf
                        ret_hedgeFH         = (profiel.toed_bond .* (annuityreturn_t - 1) + (1 - profiel.toed_bond) .* (economy.cash_return(:,t) - 1) + 1);
                    else
                        % formule (5) in Muns & Werker (feb2019):
                        ret_hedgeFH         = 1 + par.toed_renteafdekking .* (annuityreturn_t - 1);
                    end
                    if (par.toed==3)
                        hcap       = [profiel_stock0(2:end)./profiel_stock0(end)-1 0] .* assets(:,:,t);
                        hcap_ret_t = ret_hedgeFH;  % per saldo: par.toed_hcapafdekking=0
                    end
                    % formule (8) in 'monetaire vorm'
                    ret_hedgeF_monet           = (ret_hedgeFH-1) .* assets(:,:,t);
                    ret_hedgeF_monet(:,Iworkt) = ret_hedgeF_monet(:,Iworkt) + par.toed_hcapafdekking * (ret_hedgeFH(:,Iworkt) - hcap_ret_t(:,Iworkt)) .* hcap(:,Iworkt);

                    extra_return_monet       = sum( ret_t_monet - ret_hedgeF_monet, 2);
                    if strcmp(par.smoothper_indiv_meth, 'geheugenloos')
                        % werk a^o/u bij:
                        if par.adj_stock_smooth
                            annuitypricePsurv_ou = annpricePsurv0_uitsm10./annpricePsurv_u10;
                        else % als beleggingsexposure niet wordt aangepast bij toedeling
                            annuitypricePsurv_ou = ones(par.nsim,num_ages);
                        end
                        switch par.toed
                            case 1
                                total_wealth_t1 = total_wealth_t+ret_hedgeF_monet;
                            case 2 % geef iedereen zelfde aantal annuteiten obv speculatieve rendement
                                total_wealth_t1 = squeeze( annuitypricePsurv1(:,:,t) - 1 );
                            otherwise
                                if any(profiel.stock(:))
                                    total_wealth_t1 = squeeze( [profiel_stock0(2:end)./profiel_stock0(end) 1] .* assets(:,:,t) ) + ret_hedgeF_monet;
                                else
                                    total_wealth_t1 = squeeze( assets(:,:,t) );
                                end
                        end
                        % uitsmeerstap 4:
                        uitsmeersaldo(:,t) = max(-1, uitsmeersaldo(:,t) + extra_return_monet./sum( total_wealth_t1(:,1:end-1).*annuitypricePsurv_ou(:,2:end), 2)); % teller en noemer gaan *sum(total_wealth_t1(:,1:end-1)) *sum( assets(:,:,t), 2)
                        v                  = total_wealth_t1.*[annuitypricePsurv_ou(:,2:end)  zeros(par.nsim,1)]; % annuitypricePsurv_payout_t;
                        clear total_wealth_t1  annuitypricePsurv_ou  annpricePsurv_u00  annpricePsurv_u10
                    else
                        if par.toed==1
                            v = total_wealth_t + ret_hedgeF_monet; % 20210830S: + ret_hedgeF_monet  toegevoegd.
                        else
                            if par.toed==2 % geef iedereen zelfde aantal annuteiten obv speculatieve rendement
                                v = squeeze( annuitypricePsurv1(:,:,t) - 1 );
                            else
                                % par.toed==3
                                % V_i * R_i^O = O * V_i * y_i / \sum_i {V_i * y_i}
                                if any(profiel.stock(:))
                                    v = squeeze( profiel.stock .* assets(:,:,t) );
                                else
                                    v = squeeze( assets(:,:,t) );
                                end
                            end
                        end
                    end
                    if any(profiel.toed_staffel~=1) && all(~isnan(profiel.toed_staffel))
                        v = v .* profiel.toed_staffel;
                    end
                    var.toed_weight        = v./sum(v,2);
                    ret_extra_monet_by_age = var.toed_weight .* extra_return_monet;
                    ret_t_monet            = ret_hedgeF_monet + ret_extra_monet_by_age;
                    clear ret_hedgeFH  ret_hedgeF_monet  ret_extra_monet_by_age  extra_return_monet
                elseif par.toed==4
                    v           = squeeze( annuitypricePsurv1(:,:,t) - 1);
                    var.toed_weight = v./sum(v,2);
                    ret_t_monet = var.toed_weight .* sum(ret_t_monet,2);
                end
                if t==2
                    var.toed_weight1     = var.toed_weight;
                    % toed_exp_i(%) .* assets_i = toed_weight * assets_fonds
                    % toed_weight is genormaliseerde blootstelling aan opbrengst in euro's
                    var.toed_exp1       = var.toed_weight .*sum( assets(:,:,t), 2) ./ assets(:,:,t);
                    % toed_stockexp_i(%) = toed_exp_i(%) * stockexp_fund(%)
                    %                    = toed_exp_i(%) * sum(stockexp_i *assets_i) / assets_fonds
                    %                    = toed_weight   * sum(stockexp_i *assets_i) / assets_i
                    %  -> assets_fonds valt weg
                    % (noemer voor fund stock exposure en teller (= noemer van noemer) voor toedeling)
                    var.toed_stock_exp1 = var.toed_weight .*sum( profiel.stock .* assets(:,:,t), 2) ./ assets(:,:,t);
                elseif t==1+par.nsim_hor % laatste t
                    % zie var.toed_exp1 voor uitleg
                    var.toed_explast       = var.toed_weight .*sum(                  assets(:,:,t), 2) ./ assets(:,:,t);
                    var.toed_stock_explast = var.toed_weight .*sum( profiel.stock .* assets(:,:,t), 2) ./ assets(:,:,t);
                end
                clear v;
                if strcmp(par.contract,'4A')
                    A1_tmp   = squeeze(assets(:,:,t) + ret_t_monet );
                    assets0  = squeeze(assets(:,:,t));
                    if isnan(par.maxret) && isnan(par.minret)
                        % ken beleggingsrendementen toe aan persoonlijk pensioenvermogen
                        assets(:,:,t)    = A1_tmp;
                    else
                        sumassets0        = sum(assets0, 2);
                        % Om -Inf/Inf-uitkomsten te voorkomen: voorkom neg totale assets
                        sumassets1_norestr = max(0, sum(A1_tmp, 2)); % nodig voor omvang totale assets na return t
                        sumassetsminret   = (1+par.minret) .* sumassets0;
                        sumassetsmaxret   = (1+par.maxret)/(1+par.minret) * sumassetsminret;
                        if ~isnan(par.maxret)% maximale return, om te positieve rendementen op financiele vermogen uit te sluiten
                            % Als sumassets0 > sumassetsmaxret, dan is maximale return sumassetsmaxret voor iedereen behaald.
                            % Schaal in dat geval de maximale return naar boven bij, zodat iedereen dezelfde return krijgt.
                            vmaxret          = max(sumassets1_norestr ./ sumassetsmaxret, 1) * (1+par.maxret);
                            assets1_max      = vmaxret .* assets0;
                            sumassetsmaxret  = vmaxret .* sumassets0;
                            sumassets1       = sum(min(assets1_max, A1_tmp),2);   % kap af op max.return
                            A1_tmp            = assets1_max + min( A1_tmp - assets1_max, 0) ...
                                .* (sumassetsmaxret - sumassets1_norestr)./(sumassetsmaxret - sumassets1);   % schaal totale assets naar boven bij
                            A1_tmp(isnan(A1_tmp)) = 0;
                            % ken beleggingsrendementen toe aan persoonlijk pensioenvermogen
                            assets(:,:,t)    = A1_tmp;
                            clear vmaxret vmatret_monet assets1_max sumassets1 sumassetsmaxret;
                        end
                        if ~isnan(par.minret)% minimale return, om te negatieve rendementen op financiele vermogen uit te sluiten
                            % Als sumassets0 < sumassetsminret, dan is minimale return sumassetsminret niet waar te maken voor iedereen.
                            % Schaal in dat geval de minimale return naar beneden, zodat deze alsnog gehaald kan worden.
                            vminret          = min(sumassets1_norestr ./ sumassetsminret, 1) * (1+par.minret);
                            assets1_min      = vminret .* assets0;
                            sumassetsminret  = vminret .* sumassets0;
                            sumassets1       = sum(max(assets1_min, A1_tmp),2);   % kap af op min.return
                            A1_tmp            = assets1_min + max( A1_tmp - assets1_min,0) ...
                                .* (sumassets1_norestr - sumassetsminret)./(sumassets1 - sumassetsminret);   % schaal totale assets naar beneden bij
                            A1_tmp            = squeeze(A1_tmp);
                            A1_tmp(isnan(A1_tmp)) = 0;
                            % ken beleggingsrendementen toe aan persoonlijk pensioenvermogen
                            assets(:,:,t)    = A1_tmp;
                            clear vminret vminret_monet  assets1_min sumassets1 sumassetsminret;
                        end
                        clear sumassets0
                    end
                    clear assets0 A1_tmp;
                end
            end
            if strcmp(par.contract(1),'1')
                accruals_before_per_contr  = squeeze( accrualPVsum(:,t) ./ sum( contr(:,:,t), 2 ));
                accruals_end_per_contr     = squeeze( accrualPVsum_end(:,t) ./ sum( contr(:,:,t), 2 ));
                accrual_pensdate(:,t)      = accrual(:,num_work_periods,t);
                var.fundratio_0_contr(:,t) = (var.fundratio_beforeindex(:,t)-1) .* accruals_before_per_contr;
                var.fundratio_1_contr(:,t) = (var.fundratio_afterindex(:,t)-1) .* accruals_end_per_contr;
            end
        end
        clear accrual annuityprice  annuitypricePsurv0 annuitypricePsurv0_mkt  annuitypricePsurv1  annuitypricePsurv_payout ...
                     annpricePsurv0_b11 annpricePsurv0_b00 annpricePsurv0_b01 annpricePsurv0_uitsm10  annpricePsurv0_uitsm01 annpricePsurv0_uitsm11  mat_discount0  clear mdiscRTS vuitsmeer
        try
            economy = rmfield(economy,'bond_returns');
        catch 
        end
        if par.AOW > 0 % druk alles uit in termen van gehele loon, dus incl AOW en excl franchise
            benefit_periods      = par.AOW + benefit_periods;
            benefityr            = par.AOW + benefityr;
            pension_wage_profile = pension_wage_profile + profiel.franchise(Iwork);
            profiel.franchise(:) = 0;
        end
        
        % benefit per cohort, schaal door overlevend aantal personen
        benefit_period_coh      = permute(extractDiag(benefit_periods ./ profiel.Psurv(:, find(Iwork,1,'last'):num_ages-1,1:max(1,end-1)), 1:3), [1 3 2]) ;
        benefit_period_coh      = benefit_period_coh(:,:, 2:end);
        pension_wage_period_coh = repmat( permute(extractDiag( permute( pension_wage_profile00, [3 2 1] )  , 1:3), [1 3 2]), [nsim 1 1]) ;
        var.avg_benefitcoh             = squeeze( nanmean(benefit_period_coh,2));
        var.proj_return_eff_opslag_end = proj_return_eff_opslag - (proj_return_eff_opslag>1);
        var.ben1                = NaN(nsim,num_cohorts);
        if size(pension_wage_profile00,1)==1 || all(par.wage_inflation==0) || all(isnan(par.wage_inflation))
            var.repl_rate_benavg_wageavg = squeeze( var.avg_benefitcoh ./ nanmean(profiel.wage(Iwork) - profiel.franchise(Iwork)) );
            % verhouding eerste pensioeninkomen tov gemiddelde pensioengevend inkomen
            var.ben1(:, 1:num_pension_years-1) = squeeze(benefit_periods(:,end:-1:2,2));
            var.ben1(:, num_pension_years:end) = squeeze(benefit_periods(:,1,2:end));
            var.repl_rate_ben1_wageavg         = var.ben1 ./ (mean(profiel.wage(Iwork) - profiel.franchise(Iwork))) ./ profiel.Psurv(:,find(Iwork,1,'last'),:);
        else
            var.repl_rate_benavg_wageavg = squeeze( var.avg_benefitcoh ./ nanmean(pension_wage_period_coh(:,:,num_pension_periods+1:end),2) ); % ./ nanmean(profiel.wage(Iwork) - profiel.franchise(Iwork)) );
            % verhouding eerste pensioeninkomen tov gemiddelde pensioengevend inkomen
            var.ben1(:, 1:num_pension_years-1) = squeeze(benefit_periods(:,end:-1:2,2));
            var.ben1(:, num_pension_years:end) = squeeze(benefit_periods(:,1,2:end));
            var.repl_rate_ben1_wageavg         = var.ben1 ./ nanmean(pension_wage_period_coh(:,:,1:num_sim_prds1-1),2) ./ profiel.Psurv(:,find(Iwork,1,'last'),:);
        end
        % verhouding eerste pensioeninkomen tov laatstverdiende pensioengevende loon
        var.repl_rate_ben1_wagelast                          = NaN(nsim, num_cohorts);
        var.repl_rate_ben1_wagelast(:,1:num_pension_years-1) = squeeze(benefit_periods(:,end:-1:2,2));
        var.repl_rate_ben1_wagelast(:,num_pension_years:end) = squeeze(benefit_periods(:,1,2:end));
        var.repl_rate_ben1_wagelast                          = var.repl_rate_ben1_wagelast ./ pension_wage_profile00(:,num_work_periods) ./ profiel.Psurv(:,find(Iwork,1,'last'),:); % (profiel.wage(num_work_periods) - profiel.franchise(num_work_periods));
        % verhouding laatste pensioeninkomen tov laatstverdiende pensioengevende loon
        var.repl_rate_benlast_wagelast                                = NaN(nsim, num_cohorts);
        var.repl_rate_benlast_wagelast(:,1:end-num_pension_years+1)   = squeeze(benefit_periods(:,end,2:end));
        var.repl_rate_benlast_wagelast(:,end-num_pension_years+2:end) = squeeze(benefit_periods(:,end-1:-1:1,end));
        var.repl_rate_benlast_wagelast                                = var.repl_rate_benlast_wagelast ./ pension_wage_profile00(:,num_work_periods) ./ profiel.Psurv(:,find(Iwork,1,'last'),:);
        
        % uitkeringen per cohort en per leeftijd
        benefitcohageperiod = NaN(nsim, num_pension_years, num_cohorts);
        % netto contante waarde: RTS-verdisconteerde som van uitkeringen per scenario en per cohort
        var.ncw_benefitcoh  = NaN(par.nsim, num_cohorts);
        
        for c=1:num_cohorts % per cohort
            index_pension_ageper   = 1+(round(v_first_pension_age_coh(c)):round(v_last_pension_age_coh(c)))-par.pension_age;
            index_pension_periods  = v_first_pension_period_coh(c):v_last_pension_period_coh(c);
            benefitcohageperiod(:, index_pension_ageper, c) = benefit_period_coh(:,index_pension_periods,c);
            % verdiscontterde benefits per (levend) cohort op pensioendatum
            dim2 = min(1:length(index_pension_periods),    size(discount_mat_util, 2));
            dim3 = min(floor(v_first_pension_year_coh(c)), size(discount_mat_util, 3));
            var.ncw_benefitcoh(:,c) = squeeze( sum( benefitcohageperiod(:, index_pension_ageper, c) .* discount_mat_util(:, dim2, dim3), 2) );
        end
        clear dim2  dim3  benefit_period_coh
        var.pensioengroei10         = NaN(nsim, num_cohorts);
        if num_pension_periods >= 11
            % groei pensioen na 10 jaar
            var.pensioengroei10         = squeeze( benefitcohageperiod(:,11,:) ./ benefitcohageperiod(:,1,:));
        end
        var.pensioengroei20         = NaN(nsim, num_cohorts);
        if num_pension_periods >= 21
            % groei pensioen na 20 jaar
            var.pensioengroei20         = squeeze( benefitcohageperiod(:,21,:) ./ benefitcohageperiod(:,1,:));
        end
        var.mCRRA        = NaN(nsim, num_cohorts);
        var.mCRRA_report = NaN(nsim, num_cohorts);           
        for c = 1:num_cohorts % loop over cohorten
            switch par.utiltype
                case {"NCW_CRRA_ben", "NCW_CRRA_realben"}
                    if c <= num_pension_periods
                        index_pension_ages_c = num_pension_periods+1-v_num_pension_periods_coh(c):num_pension_periods;
                    else
                        index_pension_ages_c = 1:v_num_pension_periods_coh(c);
                    end
                    if all(isnan(par.hist_returns))
                        scale_factor = num_work_periods / v_num_work_periods_coh(c);
                    else
                        scale_factor = 1;
                    end
                    if strcmp(par.utiltype,"NCW_CRRA_ben")
                        m_c = benefitcohageperiod(:, index_pension_ages_c, c)/(mean(pension_wage_profile(pension_wage_profile>0)) ) * scale_factor;
                    end
                    if strcmp(par.utiltype,"NCW_CRRA_realben") % real benefits
                        m_c = (benefitcohageperiod(:, index_pension_ages_c, c) ./ mean( pension_wage_profile00(v_first_pension_period_coh(c):v_last_pension_period_coh(c), :) )) * scale_factor;
                    end
                    dim2 = min(1:v_num_pension_periods_coh(c), size(discount_mat_util, 2));
                    dim3 = min(v_first_pension_period_coh(c) , size(discount_mat_util, 3));
                    
                    var.mCRRA(:,c)        = CRRAutility( m_c, par.gamma,        par.bCEQ, discount_mat_util(:, dim2, dim3), false, true);
                    var.mCRRA_report(:,c) = CRRAutility( m_c, par.gamma_report, par.bCEQ, discount_mat_util(:, dim2, dim3), false, true);
                    clear m_c dim2 dim3
                case "CRRA_NCW_ben"
                    var.mCRRA(:,c)        = CRRAutility( var.ncw_benefitcoh(:,c) , par.gamma       , par.bCEQ, [], true, true );
                    var.mCRRA_report(:,c) = CRRAutility( var.ncw_benefitcoh(:,c) , par.gamma_report, par.bCEQ, [], true, true );
                case "CRRA_avgben"
                    var.mCRRA(:,c)        = CRRAutility( squeeze( nanmean(benefitcohageperiod(:,:,c),2)) , par.gamma       , par.bCEQ, [], true, true );
                    var.mCRRA_report(:,c) = CRRAutility( squeeze( nanmean(benefitcohageperiod(:,:,c),2)) , par.gamma_report, par.bCEQ, [], true, true );
                case "assets/accruals pensioendatum"
                    if strcmp(par.contract(1),'1')
                        % accrualPVsum_pension
                        v = annuitypricePsurv1(:,num_work_periods+1,v_first_pension_period_coh(c)+1) .* accrual_pensdate;
                    else
                        v = assets(:,num_work_periods,v_first_pension_period_coh(c));
                    end
                    var.mCRRA(:,c)              = CRRAutility( v, par.gamma       , par.bCEQ, [], true, true );
                    var.mCRRA_report(:,c)       = CRRAutility( v, par.gamma_report, par.bCEQ, [], true, true );
                    clear v;
            end
        end
        var.assets1                   = [];
        var.assets_end                = [];
        var.sum_assets = 0;
        if ~isempty(assets)
            var.sum_assets = squeeze(sum(assets,2));
        end
        if ~isempty(var.assets_sol)
            var.sum_assets = var.sum_assets + var.assets_sol;
        end
        if var.sum_assets==0
            var.sum_assets=[];
        end
        if strcmp(par.contract(1),'4')
            var.assets1                   = squeeze( assets(:,1,:) );
            var.assets_end                = squeeze( assets(:,:,end) );
        end
        if ~isnan(par.disp_dkg_before) || ~isnan(par.disp_dkg_after) || (~isnan(par.disp_coh1) && (par.disp_coh1 <= par.disp_coh2))
            if ~exist(['.\Excel\' excelfile_str], 'dir')
                mkdir(['.\Excel\' excelfile_str])
            end
            cnt     = 0;
            for s = {'disp_dkg_before','disp_dkg_after','disp_pfreturn'}
                if ~isnan(par.(s{1}))
                    sfile    = ['.\Excel\' excelfile_str '\' s{1}(6:end) '_index.xlsx'];
                    cnt      = cnt+1;
                    loc_cols = 2:par.(s{1})+1;
                    if loop==1
                        firstrow       = 10; 
                        firstrow0      = firstrow;
                        mean_Uval{cnt} = NaN(nloops, length(loc_cols));
                        mean_val{cnt}  = NaN(nloops, length(loc_cols));
                        std_val{cnt}   = NaN(nloops, length(loc_cols));
                        min_val{cnt}   = NaN(nloops, length(loc_cols));
                        val5{cnt}      = NaN(nloops, length(loc_cols));
                        median_val{cnt}= NaN(nloops, length(loc_cols));
                        val95{cnt}     = NaN(nloops, length(loc_cols));
                        max_val{cnt}   = NaN(nloops, length(loc_cols));
                        % veeg sheet leeg
                        xlswrite_wrap(sfile, NaN(firstrow+par.nsim*nloops, par.(s{1}) + 1 ), par.sheetname, 1, 1, [], true);
                    end
                    switch s{1}
                        case 'disp_dkg_before'
                            mtmp_val = var.fundratio_beforeindex(:, loc_cols);
                        case 'disp_dkg_after'
                            mtmp_val = var.fundratio_afterindex(:, loc_cols);
                        case 'disp_pfreturn'
                            mtmp_val = ret_fund(:, loc_cols)-1;
                    end
                    [~, firstrow00]        = xlswrite_wrap(sfile, [firstrow0-firstrow+(1:par.nsim)'  mtmp_val], par.sheetname, 1, firstrow0+1, [], false);
                    mean_Uval{cnt}(loop,:) = mean(mtmp_val.^(1-par.gamma), 1);
                    mean_val{cnt}(loop,:)  = mean(mtmp_val, 1);
                    std_val{cnt}(loop,:)   = std(mtmp_val, [], 1);
                    min_val{cnt}(loop,:)   = min(mtmp_val, [], 1);
                    vtmp                   = prctile(mtmp_val, [5 50 95], 1);
                    val5{cnt}(loop,:)      = vtmp(1,:);
                    median_val{cnt}(loop,:)= vtmp(2,:);
                    val95{cnt}(loop,:)     = vtmp(3,:);
                    clear vtmp;
                    max_val{cnt}(loop,:)   = max(mtmp_val, [], 1);
                    if loop==nloops % laatste loop: totale statistieken wegschrijven
                        vheading          = num2cell( [NaN(firstrow-1,1) [loc_cols-1;
                                                           ( vwght_loop * mean_Uval{cnt} ).^(1/(1-par.gamma));
                                                             vwght_loop * mean_val{cnt} ;
                                                               NaN(6, length(loc_cols))] ]);
                        for j=1:length(loc_cols)
                            scell                     = ['(' char(64+1+j) num2str(firstrow+1) ':' char(64+1+j) num2str(firstrow+nsim0) ')'];
                            vheading(4:firstrow-1,1+j) = {['=STDEV.S' scell],['=MIN' scell],['=PERCENTILE.INC(' scell '; 5%)'],['=MEDIAN' scell],['=PERCENTILE.INC(' scell '; 95%)'],['=MAX' scell]};
                        end
                        vheading(1:firstrow-1,1)  = {'sim.per.(=t-1)', ['CEQ, gamma=' num2str(par.gamma)], 'gemiddelde','st.dev','minimum','5%','mediaan','95%','maximum'}';
                        vheading(firstrow,1)      = {'simulatie'};
                        vheading(firstrow,2:end) = num2cell(loc_cols-1);
                        writecell(vheading, sfile,'Sheet',par.sheetname,'Range','A1'); % ,'AutoFitWidth',false
                        clear vheading scell
                    end
                    clear c col mtmp mtmp_ben loc_ages
                end
            end
            if abs(par.disp_coh2) <= par.max_age || par.disp_last
                if abs(par.disp_coh2) <= par.max_age
                    % schrijf stats voor specifieke cohorten weg:
                    vc1 = par.disp_coh1 : par.disp_step : par.disp_coh2;
                else
                    vc1 = [];
                end
                if par.disp_last
                    vc1 = [vc1  (par.nsim_hor-1) - par.max_age];
                end
                for c1 = vc1
                    if -c1<=par.max_age
                        sfile   = ['.\Excel\' excelfile_str '\' num2str(round(-c1,1)) '.xlsx'];
                        cnt = cnt+1;
                        c   = 1 + (par.max_age+c1);
                        loc_cols = (v_first_pension_age_coh(c):v_last_pension_age_coh(c))-par.pension_age+1;
                        if loop==1
                            firstrow       = 10;
                            firstrow0      = firstrow;                    
                            mean_Uval{cnt} = NaN(nloops, length(loc_cols));
                            mean_val{cnt}  = NaN(nloops, length(loc_cols));
                            std_val{cnt}   = NaN(nloops, length(loc_cols));
                            min_val{cnt}   = NaN(nloops, length(loc_cols));
                            val5{cnt}      = NaN(nloops, length(loc_cols));
                            median_val{cnt}= NaN(nloops, length(loc_cols));
                            val95{cnt}     = NaN(nloops, length(loc_cols));
                            max_val{cnt}   = NaN(nloops, length(loc_cols));
                            % veeg sheet leeg
                            xlswrite_wrap(sfile, NaN(firstrow+par.nsim*nloops, par.max_age-min(abs(par.disp_coh1),abs(par.disp_coh2)) + 4 ), par.sheetname, 1, 1, [], true);
                        end
                        col = 1 - c1 - par.first_exposure_age;
                        if col >= 1
                            mtmp = [var.asset_init(:, col)  var.accrual_init(:, col) ];
                        else
                            mtmp = zeros(nsim,2);
                        end
                        mtmp_val = squeeze(benefitcohageperiod(:, loc_cols, c));
                        [~, firstrow00] = xlswrite_wrap(sfile, [firstrow0-firstrow+(1:par.nsim)'  mtmp  mtmp_val], par.sheetname, 1, firstrow0+1, [], false);
                        clear mtmp
                        
                        mean_Uval{cnt}(loop,:) = mean( mtmp_val.^(1-par.gamma) , 1);
                        mean_val{cnt}(loop,:)  = mean(mtmp_val, 1);
                        std_val{cnt}(loop,:)   = std(mtmp_val, [], 1);
                        min_val{cnt}(loop,:)   = min(mtmp_val, [], 1);
                        vtmp                   = prctile(mtmp_val, [5 50 95], 1);
                        val5{cnt}(loop,:)      = vtmp(1,:);
                        median_val{cnt}(loop,:)= vtmp(2,:);
                        val95{cnt}(loop,:)     = vtmp(3,:);
                        clear vtmp;
                        max_val{cnt}(loop,:)   = max(mtmp_val, [], 1);
                        if loop==nloops
                            vheading    = num2cell([NaN(firstrow-1,3) [v_first_pension_age_coh(c):v_last_pension_age_coh(c);
                                                             ( vwght_loop * mean_Uval{cnt} ).^(1/(1-par.gamma));
                                                               vwght_loop * mean_val{cnt} ;
                                                               NaN(6, length(loc_cols))] ]);
                            for j=1:length(loc_cols)
                                scell                     = ['(' char(64+3+j) num2str(firstrow+1) ':' char(64+3+j) num2str(firstrow+nsim0) ')'];
                                vheading(4:firstrow-1,3+j) = {['=STDEV.S' scell],['=MIN' scell],['=PERCENTILE.INC(' scell '; 5%)'],['=MEDIAN' scell],['=PERCENTILE.INC(' scell '; 95%)'],['=MAX' scell]};
                            end
                            vheading(1:firstrow-1,3)   = {'leeftijd', ['CEQ, gamma=' num2str(par.gamma)], 'gemiddelde','st.dev','minimum','5%','mediaan','95%','maximum'}';
                            vheading(firstrow,1:3)   = {'simulatie', 'init asset', 'init accrual'};
                            vheading(firstrow,4:end) = num2cell(v_first_pension_age_coh(c):v_last_pension_age_coh(c));
                            writecell(vheading, sfile,'Sheet',par.sheetname,'Range','A1'); % ,'AutoFitWidth',false
                            clear vheading scell

                            row_sheetnr = row_sheetnr+1;
                            mCEQperjaar  (row_sheetnr, [false ismember( par.pension_age:par.max_age , v_first_pension_age_coh(c):v_last_pension_age_coh(c) )] ) = ( vwght_loop * mean_Uval{cnt} ).^(1/(1-par.gamma));
                            meanperjaar  (row_sheetnr, [false ismember( par.pension_age:par.max_age , v_first_pension_age_coh(c):v_last_pension_age_coh(c) )] ) =   vwght_loop * mean_val{cnt} ;
                            medianperjaar(row_sheetnr, [false ismember( par.pension_age:par.max_age , v_first_pension_age_coh(c):v_last_pension_age_coh(c) )] ) =   vwght_loop * median_val{cnt} ;

                            if col >= 1
                                initcol =  find( (par.pension_age:par.max_age) == v_first_pension_age_coh(c) ); % er wordt automatisch 1 vanaf getrokken, omdat kolommen starten bij initcol.
                                mCEQperjaar(row_sheetnr,  initcol) = mean( var.accrual_init(:, col) );
                                meanperjaar(row_sheetnr,  initcol) = mCEQperjaar(row_sheetnr,  initcol);
                                medianperjaar(row_sheetnr,initcol) = mCEQperjaar(row_sheetnr,  initcol);
                            end
                            m0perjaar(row_sheetnr, :) = {excelfile_str, par.sheetname, -c1};
                            clear initcol
                        end
                        clear c col mtmp mtmp_ben loc_ages scell
                    end
                end
            end
            firstrow0 = firstrow00;
        end
        var0 = addRows(var0,var);
        [var0.meanfundbondmix] = updateMeans( sum(vwght_loop(1:loop-1)), vwght_loop(loop), {var0.meanfundbondmix}, {meanfundbondmix});
        clear var assets
        end
        % laatste cohort waarvan gehele pensioenperiode bekend is:
        c0      = num_cohorts - num_pension_periods + 1;
        saved   = false;
        % let op: eerste twee periodes wordt niets uitgekeerd omdat eerste periode een
        % 'dummyjaar' is zodat t-1 gebruikt kan worden bij initile
        % periode t=2. Geboortecohort is alsof in jaar 1 wordt gestart (niet 0).
        bcontinue = false; % geen optimalisatie -> stop
    end
    if ~saved
        if isempty(var0.stockexposure) && isfield(profiel,'stock')
            var0.stockexposure = profiel.stock;
            var0.bondexposure  = profiel.bond;
        end
        if isempty(var0.profielstock1) && isfield(profiel,'stock')
            var0.profielstock1 = profiel.stock;
        end
        coh_weight = (min([v_num_work_periods_coh/num_work_periods ; v_num_pension_periods_coh/num_pension_periods]) .* par.disconto_coh.^(1:num_cohorts))'; % weeg cohort obv discontovoet
        writeDescriptives(stat_col,dirfile,par,sheetcell{i},c0,v_first_pension_period_coh,coh_weight, par.bwriteallstats,var0);
        
        saved = true;
    end
    clearvars  -except  par  excelfile_str  dirfile  i  meanperjaar  medianperjaar  mCEQperjaar  m0perjaar  parcell  profielcell  row_sheetnr   sdir   sheet  sheet_compute  sheetcell  sheets spath;
    vutil = [];
    end
end

if (~isnan(par.disp_coh1) && (par.disp_coh1 <= par.disp_coh2))
    meanperjaar(row_sheetnr+1:end,:) = [];
    mCEQperjaar(row_sheetnr+1:end,:) = [];
    m0perjaar(row_sheetnr+1:end,:)   = [];
    vheading = [{'bestandsnaam' , 'contract' , 'startleeftijd' }, num2cell(par.pension_age-1:par.max_age) ];
    
    sfile    = ['./' sdir '/mean' datestr(date,'yyyymmdd') '.xlsx'];    
    fun   = @(x) xlswrite(sfile, x, excelfile_str, 'A1');
    try
        fun([vheading; m0perjaar  num2cell(meanperjaar)]);
    catch
        disp(['Please close file ' sfile])
        disp('Press any key to continue')
        pause;
        fun([vheading; m0perjaar  num2cell(meanperjaar)]);
    end
       
    sfile = ['./' sdir '/CEQ' datestr(date,'yyyymmdd') '.xlsx'];
    fun   = @(x)xlswrite(sfile, x, excelfile_str, 'A1');
    try
        fun([vheading; m0perjaar  num2cell(mCEQperjaar)]);
    catch
        disp(['Please close file ' sfile])
        disp('Press any key to continue')
        pause;
        fun([vheading; m0perjaar  num2cell(mCEQperjaar)]);
    end
end
 
end

function var0 = addRows(var0,var)
% voeg fields in var toe aan fields in var0
    for s = fieldnames(var)'
        var0.(s{1}) = [var0.(s{1}); var.(s{1})];
    end
end

function [muitsmeer, vuitsmeer] = getmuitsmeer(smoothper,smoothper_indiv_meth,num_work_periods,num_pension_periods)
    num_ages  = num_work_periods + num_pension_periods;
    muitsmeer = ones(1,1,num_pension_periods);
    if ~(smoothper>1)
        vuitsmeer = ones(1,num_ages); % geen uitsmeren
    else  % wel uitsmeren
        switch smoothper_indiv_meth % sterftekansen niet in verwerkt
            case {'geheugenloos', 'asymptotisch'}
                vuitsmeer = 1 - (1-1/smoothper).^(1:num_ages);
            case 'asymptotisch gelijke aanpassingen'
                vuitsmeer = 1/smoothper;
            case 'lineair' 
                vuitsmeer = [ (1:smoothper)/smoothper  ones(1, num_ages-smoothper)];
        end
        if ~strcmp(smoothper_indiv_meth,'asymptotisch gelijke aanpassingen')
            vuitsmeer = vuitsmeer(1:min(num_ages, length(vuitsmeer)));
            muitsmeer = getUitsmeerMatrix(vuitsmeer,num_work_periods,num_pension_periods);
        end
    end
end

function varargout = updateMeans(wA, wB, A, B)
    % voeg gemiddeldes in A (gewicht wA/(wA+wB)) samen met B (gewicht wB/(wA+wB))
    varargout = cell(1,length(A));
    for i = 1:length(A)
        if isempty(wA) || wA==0
            varargout{i} = B{i};
        else
            varargout{i} = (wA*A{i} + wB*B{i})/(wA+wB);
        end
    end
end

function [bondret, bondmat_t] = getRenteafdekkingKX(bond_fraction, fundratio_beforeindex_t, b, RTS_maturities, bond_returns_t, KX, wB, bbyage, par)
    if ndims( wB )==3
        wB            = getDiags(wB, ~bbyage); % looptijd verplichtingen per looptijd
        wB(isnan(wB)) = 0;
    end
    wB            = wB./sum(wB, ndims(wB));
    mB            = -b .* RTS_maturities(:);
    % selecteer looptijd die het renterisico het beste afdekt gegeven investering in obligaties:
    if bbyage
        % selecteer looptijd die het renterisico het beste afdekt gegeven investering in obligaties:
        bondmat_t = NaN(size(wB,1), size(wB,2));
        bondret   = NaN(size(wB,1), size(wB,2));
        for age = size(wB,2):-1:1
            if bond_fraction(age)>0 || age==size(wB,2)
                if ndims( wB )==3
                    wBmB   = squeeze( wB(:,age,:) ) * mB(1:size(wB,3),:);
                else % wB is discountmat
                    age1         = age + par.first_work_age;
                    numpensper   = min( par.max_age + 1 - age1, par.max_age + 1 - par.pension_age);
                    firstpensper = max(1, par.pension_age - age1);
                    cols         = firstpensper + (1:numpensper);
                    wBmB         = ( wB(:,cols)./sum(wB(:,cols),2) ) * mB(cols,:);
                end
                diffVar = NaN(size(wBmB,1), size(mB,1));
                for tau=1:size(mB,1)
                    mtmp           = wBmB - bond_fraction(age) .* fundratio_beforeindex_t .* mB(tau,:);
                    diffVar(:,tau) = sum( (mtmp .* KX').^2, 2);
                end
                [~,I]            = min(diffVar,[],2,'linear');
                bondmat_t(:,age) = floor((I-1)/size(diffVar,1));
                bondret(:,age)   = bond_returns_t(I);
            else
                bondmat_t(:,age) = bondmat_t(:,age+1);
                bondret(:,age)   = bondret(:,age+1);
            end
        end
    else
        wBmB    = wB * mB(1:size(wB,2),:);
        diffVar = NaN(size(wB,1), size(mB,1));
        if length(bond_fraction)>1
            bond_fraction = wB * bond_fraction(:);
        end
        for tau=1:size(mB,1)
            mtmp           = wBmB - bond_fraction .* fundratio_beforeindex_t .* mB(tau,:);
            diffVar(:,tau) = sum( (mtmp .* KX').^2 , 2);
        end
        [~,I]     = min(diffVar,[],2,'linear');
        bondmat_t = floor((I-1)/size(diffVar,1));
        bondret   = bond_returns_t(I);
    end
end

function muitsmeerq = getUitsmeerMatrix(vuitsmeerq,num_work_periods,num_pension_periods)
    muitsmeerq = zeros(1,num_work_periods+num_pension_periods,num_pension_periods);
    for j=1:size(muitsmeerq,2)
        periods_in_pension = max( j + 1 - num_work_periods, 1);
        vmatur0            = num_work_periods - j + (periods_in_pension:num_pension_periods);
        muitsmeerq(1,j,periods_in_pension:num_pension_periods) = vuitsmeerq(vmatur0);
    end
end

function [accrualPVsum_t, accrualPVsum_end_t, accrual0, accrual_t, fundratio_beforeindex_t, fundratio_afterindex_t, indexation_t, cutyears_t, lowyears_t, unc_cut1A_t] = ...
         adjust_accrual(accrual_t, annuitypricePsurv1_tann1, assets_sol_t, price_inflation_t1, indexation0_t, cutyears_t1, lowyears_t1, unc_cut1A_t1, par )
    % adjust accrual 
    % als geen biometrisch rendement: pas verplichtingen aan op sterfte
    % pas evt vaste indexatieregel toe op vorige opbouw:
    accrual_t(:, 2:end) = accrual_t(:, 2:end) .* indexation0_t;
    % indexeer/kort alleen aan het einde van de periode:
    % bereken totale verplichtingen per scenario (bij inleg geldt:
    % accrualPVsum(tegen einde periode waarde) = assets_sol*(1 + looptijd gewogen risicovrije rente)
    accrualPVsum_t      = squeeze( sum( annuitypricePsurv1_tann1 .* accrual_t, 2) ); % squeeze( sum( annuitypricePsurv(:,2:end,t+1) .* accrual(:,1:end-1,t), 2) );
    % bepaal dekkingsgraad na rendement en vr indexatie
    if any(accrualPVsum_t)~=0
        fundratio_beforeindex_t  = assets_sol_t ./ accrualPVsum_t;
    else
        fundratio_beforeindex_t  = ones(size(assets_sol_t));
    end
    unc_cut1A_t = zeros(par.nsim, 1);
    lowyears_t  = zeros(par.nsim, 1);
    cutyears_t  = zeros(par.nsim, 1);
    if strcmp(par.contract,'1A')
        % bepaal evt nieuwe uitgesmeerde onvoorwaardelijke korting
        I                = (cutyears_t1==0) & (lowyears_t1>=par.per_before_unc_cut);
        unc_cut1A_t(I)   = min(0, fundratio_beforeindex_t(I)-par.unc_cutratio) / par.smoothper;
        I                = (cutyears_t1>0) & (cutyears_t1<par.smoothper);
        unc_cut1A_t(I)   = unc_cut1A_t1(I);
        I                = (fundratio_beforeindex_t < par.unc_cutratio);
        lowyears_t(I)    =  lowyears_t1( I) + 1;
        I                = (unc_cut1A_t < 0);
        cutyears_t(I)    =  cutyears_t1(I) + 1;
        I                = (cutyears_t(I)>=par.smoothper);
        cutyears_t(I)    = 0;
        % bepaal indexatie/korting obv prijsindexatie
        indexation_t = 1 + max(0, price_inflation_t1-1) ...
            .* max(0, min( par.ftk_index_high, fundratio_beforeindex_t)-par.ftk_index_low)/(par.ftk_index_high - par.ftk_index_low) ...
            + min(0, fundratio_beforeindex_t-par.ftk_cut) ./ fundratio_beforeindex_t / par.smoothper + unc_cut1A_t;
    end
    Imin                = (fundratio_beforeindex_t < par.min_covratio);
    % bepaal indexatie/korting
    indexation_t(Imin)  = min( indexation_t(Imin), 1 + (fundratio_beforeindex_t(Imin)-par.min_covratio) / par.min_covratio );
    Imax                = (fundratio_beforeindex_t > par.max_covratio + (price_inflation_t1-1));
    % bepaal indexatie/korting
    if par.max_covratio_smoothper==1
        indexation_t(Imax)  = max( indexation_t(Imax), 1 + (fundratio_beforeindex_t(Imax)-par.max_covratio) / par.max_covratio );
    else    % CPB-interpretatie = spreding van 5 jaar zodat indexatie prijsinflatie + (dkg-150%)/5
            % Dus geen schaling met 150%, zie mail Luuk Metselaar 15-1-2021 11:36.
        indexation_t(Imax)  = max( indexation_t(Imax), 1 + (fundratio_beforeindex_t(Imax)-par.max_covratio) / par.max_covratio_smoothper ); % par.max_covratio
    end
    indexation_t( isnan(indexation_t)) = 1;
    % bepaal opbouw na indexatie
    accrual0            = squeeze( accrual_t );
    accrual_t           = accrual0 .* indexation_t;
    
    accrualPVsum_end_t       = squeeze(sum( annuitypricePsurv1_tann1 .* accrual_t,2));
    fundratio_afterindex_t   = assets_sol_t ./ accrualPVsum_end_t;
    fundratio_afterindex_t( isnan(fundratio_afterindex_t) ) = 1;
end


function [V1herverdeel, percVherverdeel, V1] = distribute_initialassets(par, yFTK, r, proj_ret_opslfacteq_prem, q_invaar0, bdisp, bfigure)
% gebaseerd op distribute_initialassets.m en formules_projectierendement_invaren_20201228.docx
    if ~exist('bdisp','var') || isempty(bdisp)
        bdisp = false;
    end
    if ~exist('bfigure','var') || isempty(bfigure)
        bfigure = false;
    end
    if ~exist('r','var') || isempty(r)
        r           = 1.01; % risicovrije rente
    end
    v_payout   = [zeros(1, par.pension_age-par.first_exposure_age)  ones(1, par.max_age-par.pension_age+1)];
    nsim       = size(r,1);
    
    if length(proj_ret_opslfacteq_prem)==1
        proj_ret_opslfacteq_prem = repmat(proj_ret_opslfacteq_prem,1,length(v_payout));
    end
    
    Afactorqr       = NaN(nsim, par.max_age-par.first_exposure_age+1);
    Afactor1rho     = NaN(nsim, par.max_age-par.first_exposure_age+1);
    Afactor1rhofict = NaN(nsim, par.max_age-par.first_exposure_age+1);
    Afactor1r       = NaN(nsim, par.max_age-par.first_exposure_age+1);
    for a = 1:size(Afactor1rho,2)
        a1 = length(v_payout) + 1 - a;  % a:end-1 -> 1:a1 ??
        if length(r)==1
            rA = r;
        else
            rA = r(:,1:a1);
        end
        Afactor1r(:,a)       = Afactor( v_payout(a:end), r(:,1:a1), 1, 1);
        Afactorqr(:,a)       = Afactor( v_payout(a:end) .* q_invaar0(1:end+1-a), rA, 1, 1); % (1+r0)*(1+wlamr*eq_premium)-1
        if par.proj_return_opslagfactor > 0
            Afactor1rho(:,a) = Afactor( v_payout(a:end), r(:,1:a1), [1  proj_ret_opslfacteq_prem(a:end-1)], 1);
        else
            Afactor1rho(:,a) = Afactor1r(:,a);
        end
        Afactor1rhofict(:,a) = Afactor1rho(:,a);
       
    end
    VFTK   = yFTK .* Afactor1r;
   
    % verwerk initile dkg in initile inkomen (dus na stap 1 invaren)
    vLambda= yFTK .* Afactorqr;
    vqa    = Afactorqr ./ Afactor1r;   % (5a)
    if all(VFTK(:)==0)
        q0 = 1;
    else
        q0 = sum(vLambda,2) ./ sum(VFTK,2); % (5b)
    end
    dV01 = (vqa ./ q0) .* (par.init_dkg - 1); % (4)
    V1   = VFTK .* (1 + dV01);
    % y1   = V1 ./ Afactor1r; %(6)
    dy01 = dV01; % (7)
    dy12 = Afactor1r ./ Afactor1rho - 1; % (8)
    dy02 = (1+dy01) .* (1+dy12) - 1;
    dy02fict = (1 + dy02) .* Afactor1rho ./ Afactor1rhofict - 1; % (8)
    
    % kijk welke compensatie nodig is om inkomen voor jongste gepensioneerden gelijk te houden
    if isnumeric(par.herverdeelijkpunt)
        if par.herverdeelijkpunt>=par.first_exposure_age 
            xi02 = max( dy02fict(:,par.herverdeelijkpunt-par.first_exposure_age+1:end), [], 2);
        else % ijkpunt is al een percentage
            xi02 = par.herverdeelijkpunt;
        end
    else
        if strcmp(par.herverdeelijkpunt,'maximum aanpassing gepensioneerden')
            xi02 = max( dy02fict(:,par.pension_age-par.first_exposure_age+1:end), [], 2);
        end
    end
    excessypos02  = max(dy02-xi02,0)./(1+dy02); % vertaal overschot terug 
    excessyneg02  = -min(dy02-xi02,0)./(1+dy02);
    deltaVneg02xi = excessyneg02 .* V1;
    Ipens         = (par.first_exposure_age:par.max_age)>=par.pension_age;
    dV02xipens    = abs(sum( deltaVneg02xi(:, Ipens), 2));
    if strcmp(par.herverdeelfinanciering,'% vermogen allen') || bdisp || bfigure
        dpercV02xi_all      = dV02xipens ./ sum(V1,2);
    end
    if strcmp(par.herverdeelfinanciering,'% vermogen werkenden') || bdisp || bfigure
        dpercV02xi_werk     = min( dV02xipens ./ sum(V1(:, ~Ipens), 2), 1);
        vermsaldo           = min((sum( V1(:,~Ipens), 2) ./ dV02xipens),1);
        if any(vermsaldo<1)
            disp(['onvoldoende vermogen werkenden in ' num2str(sum(vermsaldo<1)) ' van de ' num2str(length(vermsaldo)) ' scenario`s'])
            deltaVneg02xi   = vermsaldo.*deltaVneg02xi;
        end
    end
    if strcmp(par.herverdeelfinanciering,'% overschot vermogen werkenden') || bdisp || bfigure
        dpercV02xi_prop     = zeros( size(VFTK,1), sum(~Ipens));
        I                   = (sum( excessypos02(:,~Ipens) .* V1(:,~Ipens), 2)>0);
        dpercV02xi_prop(I,:)= dV02xipens(I,:) ./ sum( excessypos02(I,~Ipens) .* V1(I,~Ipens), 2) .* excessypos02(I,~Ipens);
        dy02_xi_prop        = dy02;
        % reken voor dy02_xi_* terug naar y0 door te delen door (1+dy01)
        dy02_xi_prop(I,:)   = [dy02(I,~Ipens)-dpercV02xi_prop(I,:)./(1+dy01(I, ~Ipens))   max( dy02(I, Ipens), xi02(I))];
        if any(~I)
            disp(['geen beschikbaar overschot vermogen werkenden in ' num2str(sum(~I)) ' van de ' num2str(length(I)) ' scenario`s'])
        end
    end
    if bdisp
        Ipos  = (dy02 > 0);
        disp(' --- Ten opzichte van FTK zonder verwerking dkg-verschil ---')
        disp(['oppervlak y positief: ' num2str(    sum(dy02( Ipos)) , 2)])
        disp(['oppervlak y negatief: ' num2str(abs(sum(dy02(~Ipos))), 2)])
        disp(['ratio y pos./neg.:    ' num2str(abs(sum(dy02( Ipos))./sum(dy02(~Ipos))),2)])
        disp(['oppervlak V positief: ' num2str(    sum(dy02( Ipos) .* V1( Ipos)), 2)])
        disp(['oppervlak V negatief: ' num2str(abs(sum(dy02(~Ipos) .* V1(~Ipos))),2)])
        disp(['ratio V pos./neg.:    ' num2str(abs(sum(dy02( Ipos) .* V1( Ipos))./sum(dy02(~Ipos).*V1(~Ipos))),2)])
        disp(' ')
        disp('aanpassing pensioenuitkering fictief rendement jongste gepensioneerde:')
        disp(['xi = ' num2str(100*nanmean(xi02),2) '%'])
        disp(['oppervlak onder xi: ' num2str( abs( sum( deltaVneg02xi) ), 2)])
        disp(['oppervlak gepensioneerden onder xi: ' num2str( abs( sum( deltaVneg02xi(:, Ipens)) ), 2)])
        disp(['oppervlak gepensioneerden onder xi tov totaal vermogen: '               num2str(100*mean(dpercV02xi_all), 2) '%'])
        disp(['oppervlak gepensioneerden onder xi tov vermogen niet-gepensioneerden: ' num2str(100*mean(dpercV02xi_werk), 2) '%'])
        disp(['oppervlak gepensioneerden onder xi tov vermogen niet-gepensioneerden boven xi: ' num2str(100*mean(dy02_xi_prop), 2) '%'])
    end
    V1herverdeel = V1;
    switch par.herverdeelfinanciering
        case '% vermogen allen'
            V1herverdeel           = V1herverdeel .* (1 - dpercV02xi_all );
         case '% vermogen werkenden'
            V1herverdeel(:,~Ipens) = V1herverdeel(:,~Ipens) .* (1 - dpercV02xi_werk);
        case '% overschot vermogen werkenden'
            V1herverdeel(:,~Ipens) = V1herverdeel(:,~Ipens) .* (1 - dpercV02xi_prop);
    end
    V1herverdeel(:, Ipens) = V1herverdeel(:, Ipens) + deltaVneg02xi(:, Ipens);
    percVherverdeel        = V1herverdeel./V1 - 1;
    percVherverdeel(isnan(percVherverdeel)) = 0;
end

function writeDescriptives(stat_col,dirfile, par, sheet, c0, v_first_pension_period_coh, coh_weight, bwriteallstats, var0)

    colLetters = xlsColNum2Str(1:500); % converteer kolomnummer naar kolomletter
    
    var0.asset_init     = fliplr(var0.asset_init(:,1:end-1));
    var0.accrual_init   = fliplr(var0.accrual_init(:,1:end-1));
    var0.percVherv_init   = fliplr(var0.percVherv_init(:,1:end-1));
    if size(var0.asset_init,2) < size(var0.ncw_benefitcoh,2)
        extracol         = size(var0.ncw_benefitcoh,2) - size(var0.asset_init,2);
        var0.asset_init     = [var0.asset_init     zeros( size(var0.asset_init,1)  , extracol)];
        var0.percVherv_init = [var0.percVherv_init zeros( size(var0.asset_init,1)  , extracol)];
        var0.accrual_init   = [var0.accrual_init   zeros( size(var0.accrual_init,1), extracol)];
        var0.ben1           = [var0.ben1           zeros( size(var0.ben1,1)        , extracol)];
    end
    mM = [var0.ncw_benefitcoh(:,c0)          var0.pensioengroei10(:,c0) var0.pensioengroei20(:,c0) ...
            var0.mCRRA(:,c0)                 var0.mCRRA_report(:,c0)    var0.accrual_init(:,c0)  ...
            var0.percVherv_init(:,c0)        var0.asset_init(:,c0)      var0.ben1(:,c0)  ...
            var0.repl_rate_benavg_wageavg(:,c0)  var0.repl_rate_ben1_wageavg(:,c0) ...
            var0.repl_rate_ben1_wagelast(:,c0)   var0.repl_rate_benlast_wagelast(:,c0)];
    cvars = {'ncw_benefit','pensgroei10','pensgroei20', ... 
        ['CRRA_' num2str(par.gamma) '_benefits'], ['CRRA_' num2str(par.gamma_report) '_benefits'], ... 
        'init.aanspraak', 'init.herv(%verm)', 'init.verm.na.herv', 'init.uitk.na.herv.', ...
        'vvr_gempensgemloon', 'vvr1_gemloon','vvr1_laatsteloon','vvrlaatste_laatsteloon'};
    [Mstat,rowns] = displayDescriptives(mM,cvars, [],par.nsim,[],[false true true false false true true true true true true true true], ...
        [false true true false false false true false false false false false false]);
    startcol  = 15;
    vheading  = ['geboortecohort' cvars NaN cvars];
    cMstat    = [NaN vheading(1+(1:length(cvars))); rowns(:) num2cell(Mstat) ];
    [col2,row2a] = xlswrite_wrap(dirfile, cMstat, sheet, startcol, 1, colLetters, true);
    
    % CRRA-resultaten wegschrijven naar tabblad stats:
    xlswrite_wrap(dirfile, {sheet}, 'stats',stat_col,2,colLetters, true);
    I     = contains(cvars,'CRRA_');
    Mstat = Mstat(:, I);
    vgam0 = cvars(I);
    vgam  = cell(size(vgam0));
    vrowns = [];
    for i=1:length(vgam0)
        stmp    = vgam0{i};
        vgam{i} = ['gamma = ' stmp(6)];
        rowns(end-1:end) = {vgam{i}; NaN}; % geen pos en neg rij nodig, wel lege rij
        vrowns  = [vrowns; rowns(:)];
    end
    Mstat(end-1:end,:) = NaN;  % geen pos en neg rij nodig, wel lege rijen
    I          = ~any(isnan(var0.mCRRA));
    weightCRRA = mean(var0.mCRRA(:,I) * coh_weight(I)/sum(coh_weight(I)));
    
    if size(var0.stockexposure,2)==1
        var0.stockexposure = var0.stockexposure';
    end
    if size(var0.bondexposure,2)==1
        var0.bondexposure  = var0.bondexposure';
    end
    if (par.contract(1)=='1')
        w_assets = [par.stock_sol; par.bond_sol; par.cash_sol   ];
    else
        % gewichten in termen van assets
        w_assets      = NaN(size(var0.assets_end,1),3);
        % stocks:
        w_assets(:,1) = sum( var0.stockexposure .* var0.assets_end , 2) ./ sum( var0.assets_end, 2);
        % bonds:
        w_assets(:,2) = sum( var0.bondexposure .* var0.assets_end , 2) ./ sum( var0.assets_end, 2);
        % cash:
        w_assets(:,3) = 1 - sum(w_assets(:,1:2),2);
        w_assets      = mean(w_assets)';
    end
    if size(var0.stockexposure,1)==1 % terug transformeren
        var0.stockexposure = var0.stockexposure';
        var0.bondexposure  = var0.bondexposure';
    end
        
    if par.contract(1)=='1'
        bond_maturity = par.maturity_sol;
    else
        bond_maturity = par.bondmat;
    end
    if ~isnumeric(bond_maturity)
        bond_maturity = NaN;
    end
    xlswrite_wrap(dirfile, [Mstat(:); w_assets; par.stock_sol; bond_maturity; weightCRRA], 'stats', stat_col, 3, colLetters, true);
    % voeg definitie van rijen toe
    xlswrite_wrap(dirfile, {['geboortecohort ' num2str(c0-2-par.max_age)]}, 'stats',1,1,colLetters, true);
    xlswrite_wrap(dirfile, [vrowns; 'stocks_assets '; 'bonds_assets  '; 'cash_assets'; 'sol_stocks    '; 'bond_maturity '; 'wghtCRRA      '], 'stats', 1, 3, colLetters, true);
    
    cvars_per = {};
    if ~isempty(var0.fundratio_1_contr)
        mMbuffer_per = var0.fundratio_1_contr(:,end);
        cvars_per = [cvars_per {'excess funding ratio % contribution'}];
    else
        mMbuffer_per = zeros(size(var0.fundratio_1_contr, 1),0);
    end
    if ~isempty(mMbuffer_per)
        [Mstat_per,rowns] = displayDescriptives(mMbuffer_per,cvars_per, [],par.nsim,[],true(1,size(mMbuffer_per,2)), ...
            false(1,size(mMbuffer_per,2)));
        vheadingper = [NaN cvars_per];
        cMstat      = [NaN vheadingper(1+(1:length(cvars_per))); rowns(:) num2cell(Mstat_per) ];
        [~,row2b]   = xlswrite_wrap(dirfile, cMstat, sheet, col2+2, 1, colLetters, true);
        row2        = max([row2a row2b]) + 2;
    else
        row2 = row2a + 2;
    end
    
    mM = NaN(length(v_first_pension_period_coh), length(vheading));
    for c = 1:size(var0.mCRRA,2)
        mMc = [var0.ncw_benefitcoh(:,c)          var0.pensioengroei10(:,c) var0.pensioengroei20(:,c) ...
                var0.mCRRA(:,c)         var0.mCRRA_report(:,c) ...
                var0.accrual_init(:,c)  var0.percVherv_init(:,c)  var0.asset_init(:,c)  ...
                var0.ben1(:,c) ...
                var0.repl_rate_benavg_wageavg(:,c) ...
                var0.repl_rate_ben1_wageavg(:,c)  ...
                var0.repl_rate_ben1_wagelast(:, c)...
                var0.repl_rate_benlast_wagelast(:, c)];
        mM(c,:) = [c-par.max_age  mean(mMc) NaN median(mMc)];
    end
    if all(isnan(par.hist_returns))
        startcohnr    = par.max_age+1-par.pension_age;
        startcohbirth = 1-par.pension_age;
    else
        startcohnr    = 1;
        startcohbirth = 1-par.max_age;
    end
    cMstat   =[cell(1) 'mean'          cell(size(cvars)) 'median' cell(size(cvars)-[0 1]) ;
               vheading; num2cell(mM( startcohnr:end+1-startcohnr,:)) ];
    currcol = xlswrite_wrap(dirfile, cMstat, sheet, startcol, row2, colLetters, true);
    if bwriteallstats
        currcol = xlswrite_series(var0.percVherv_init(:, startcohnr:end), true, par.nsim , 'herv.perc.',...
                'geboortecohort', startcohbirth, dirfile,sheet,currcol+2, row2, colLetters,1,{'mean','stdev','median'});
        currcol = xlswrite_series(var0.mCRRA(:, startcohnr:end),true, par.nsim , ['CRRA' num2str(par.gamma)],...
                'geboortecohort', startcohbirth, dirfile,sheet,currcol+2, row2, colLetters,1,{'mean','stdev','median'});
        if any(var0.toed_weight1(:))
            currcol = xlswrite_series(var0.toed_weight1, true, par.nsim , 'verdeling excess return NPC (%,1e sim.per)',...
                    'leeftijd', par.first_exposure_age, dirfile,sheet,currcol+2, row2, colLetters,1,{'mean','median','stdev'});
        end
        if any(var0.toed_weight(:))
            currcol = xlswrite_series(var0.toed_weight, true, par.nsim , 'verdeling excess return NPC (%,laatste sim.per)',...
                    'leeftijd', par.first_exposure_age, dirfile,sheet,currcol+2, row2, colLetters,1,{'mean','median','stdev'});
        end
        if any(var0.toed_exp1(:))
            currcol = xlswrite_series(var0.toed_exp1, true, par.nsim , 'exposure excess return NPC (%,1e sim.per)',...
                    'leeftijd', par.first_exposure_age, dirfile,sheet,currcol+2, row2, colLetters,1,{'mean','median','stdev'});
        end
        if any(var0.toed_explast(:))
            currcol = xlswrite_series(var0.toed_explast, true, par.nsim , 'exposure excess return NPC (%,laatste sim.per)',...
                    'leeftijd', par.first_exposure_age, dirfile,sheet,currcol+2, row2, colLetters,1,{'mean','median','stdev'});
        end
        if any(var0.toed_stock_exp1(:))
            currcol = xlswrite_series(var0.toed_stock_exp1, true, par.nsim , 'stock exposure excess return NPC (%,1e sim.per)',...
                    'leeftijd', par.first_exposure_age, dirfile,sheet,currcol+2, row2, colLetters,1,{'mean','median','stdev'});
        end
        if any(var0.toed_stock_explast(:))
            currcol = xlswrite_series(var0.toed_stock_explast, true, par.nsim , 'stock exposure excess return NPC (%,laatste sim.per)',...
                    'leeftijd', par.first_exposure_age, dirfile,sheet,currcol+2, row2, colLetters,1,{'mean','median','stdev'});
        end
        if any(var0.profielstock1(:))
            currcol = xlswrite_series(var0.profielstock1, true, par.nsim , 'stock exposure lifecycle WVP (%,1e sim.per)',...
                    'leeftijd', par.first_exposure_age, dirfile,sheet,currcol+2, row2, colLetters,1,{'mean','median','stdev'});
        end
        if any(var0.stockexposure(:))
            currcol = xlswrite_series(var0.stockexposure, true, par.nsim , 'stock exposure lifecycle WVP (%,laatste sim.per)',...
                    'leeftijd', par.first_exposure_age, dirfile,sheet,currcol+2, row2, colLetters,1,{'mean','median','stdev'});
        end
        if any(var0.bondexposure(:))
            currcol = xlswrite_series(var0.bondexposure, true, par.nsim , 'bond exposure (%,laatste sim.per)',...
                'leeftijd', par.first_exposure_age, dirfile,sheet,currcol+2, row2, colLetters,1,{'mean','median','stdev'});
        end
        if any(var0.proj_return_eff_opslag1)
            if length(var0.proj_return_eff_opslag1)==1
                slftd    = NaN;
                firstexp = NaN;
            else
                slftd    = 'leeftijd';
                firstexp = par.first_exposure_age;
            end
            currcol = xlswrite_series(var0.proj_return_eff_opslag1,    true, par.nsim , 'opslag proj.rend.(%,1e sim.per)',...
                    slftd, firstexp, dirfile,sheet,currcol+2, row2, colLetters,1,{'mean','stdev'});
        end
        if any(var0.proj_return_eff_opslag_end)
            if length(var0.proj_return_eff_opslag_end)==1
                slftd    = NaN;
                firstexp = NaN;
            else
                slftd    = 'leeftijd';
                firstexp = par.first_exposure_age;
            end
            currcol = xlswrite_series(var0.proj_return_eff_opslag_end, true, par.nsim , 'opslag proj.rend.(%,laatste sim.per)',...
                    slftd, firstexp, dirfile, sheet,currcol+2, row2, colLetters,1,{'mean','stdev'});
        end
        clear slftd firstexp;
        if any(var0.repl_rate_ben1_wagelast(:))
            currcol  = xlswrite_series(var0.repl_rate_ben1_wagelast(:, startcohnr:end),true, par.nsim,'vvr eerste pensioenjaar',...
                'geboortecohort', startcohbirth, dirfile,sheet, currcol+2, row2, colLetters, 1);
        end
        if any(var0.repl_rate_benlast_wagelast(:))
            currcol  = xlswrite_series(var0.repl_rate_benlast_wagelast(:,startcohnr:end),true, par.nsim,'vvr laatste pensioenjaar',...
                'geboortecohort', startcohbirth, dirfile,sheet, currcol+2, row2, colLetters, 1);
        end
        if ~isempty(var0.indexation)
            currcol  = xlswrite_series(var0.indexation(:,2:end), true,par.nsim,'indexatie',...
                'periode',0,dirfile,sheet,currcol+2, row2, colLetters, 1);
            currcol  = xlswrite_series(cumprod(var0.indexation(:,2:end),2),true, par.nsim,'cumulatieve indexatie',...
                'periode',0,dirfile,sheet,currcol+2, row2, colLetters, 1);
        end
        if ~isempty(var0.fundratio_beforeindex)
            currcol  = xlswrite_series(var0.fundratio_beforeindex(:,2:end),true, par.nsim,'dekkingsgraad voor indexatie en korting',...
                'periode',0,dirfile,sheet,currcol+2, row2, colLetters, 1);
        end
        if ~isempty(var0.fundratio_afterindex)
            currcol  = xlswrite_series(var0.fundratio_afterindex(:,2:end),true, par.nsim,'dekkingsgraad na indexatie en korting',...
                'periode',0,dirfile,sheet,currcol+2, row2, colLetters, 1);
        end
        if any(var0.fundratio_0_contr(:))
            currcol  = xlswrite_series(var0.fundratio_0_contr(:,2:end),true, par.nsim,'excess funding before indexation % premiesom',...
                'periode',0,dirfile,sheet, currcol+2, row2, colLetters, 1);
        end
        if any(var0.fundratio_1_contr(:))
            currcol  = xlswrite_series(var0.fundratio_1_contr(:,2:end),true, par.nsim,'excess funding after indexation % premiesom',...
                'periode',0,dirfile,sheet, currcol+2, row2, colLetters, 1);
        end
        if any(var0.sum_assets(:))
            currcol  = xlswrite_series(var0.sum_assets(:,2:end),false, par.nsim,'som assets',...
                'periode',0,dirfile,sheet, currcol+2, row2, colLetters, 1);
        end
        if any(var0.assets1(:))
            currcol  = xlswrite_series(var0.assets1(:,2:end),false, par.nsim,'assets na eerste werkzame periode',...
                'periode',0,dirfile,sheet, currcol+2, row2, colLetters, 1);
        end
        if any(var0.assets_end(:))
            currcol  = xlswrite_series(var0.assets_end,false, par.nsim,'assets in laatste simulatieperiode',...
                'na leeftijd',par.first_exposure_age,dirfile,sheet, currcol+2, row2, colLetters,1);
        end
        if any(var0.meanfundstockmix(:))
            currcol  = xlswrite_series(var0.meanfundstockmix, false, par.nsim,'stock exposure van pensioenfonds',...
                'periode',0,dirfile,sheet, currcol+2, row2, colLetters,1);
        end
        if any(var0.meanfundbondmix(:))
            num_ages = length(par.first_exposure_age:par.max_age);
            if size(var0.meanfundbondmix,2)>2
                if par.contract(1)=='4' && strcmp(par.bondmat,'renteafdekking') && par.toed==0
                    cMstat1   =['looptijd obligatie, gemiddelde per leeftijd'                   cell(1,1+num_ages) ;
                               'periode'                                                        {'fractie'}      num2cell(par.first_exposure_age:par.max_age);
                               num2cell((0 : size(var0.meanfundbondmix,1)-1))'   num2cell(var0.meanfundbondmix)];
                else
                    cMstat1   =['bond exposures van pensioenfonds, gemiddelde per looptijd'   cell(1,num_ages) ;
                               'periode'                                         num2cell(0:size(var0.meanfundbondmix,2)-1);
                               num2cell((0 : size(var0.meanfundbondmix,1)-1))'  num2cell(var0.meanfundbondmix)];
                end
            else
                cMstat1   =['bond exposures van pensioenfonds'   cell(1,2) ;
                           'periode'                             {'fractie','looptijd (jr)'};
                           num2cell((0 : size(var0.meanfundbondmix,1)-1))'  num2cell(var0.meanfundbondmix)];
            end
            currcol  = xlswrite_wrap(dirfile, cMstat1, sheet, currcol+2, row2, colLetters, true);
        end
    end
end

function y = getDiags( mat, bsum )
    % get diagonals to obtain liabilities by maturity
    % rows of y are simulations
    % mat: (nsim, num ages, pension period)
    % bsum: sum over all ages or not
    mat  = permute( mat, [3 2 1]);
    svtmp = size( mat );
    if bsum
        y     = NaN(svtmp(3), svtmp(2));
    else
        y     = NaN(svtmp(3), svtmp(2), svtmp(2));
    end
    for i = 0:svtmp(2)-1
        ind        = ((svtmp(1)*(svtmp(2)-svtmp(1))+1):(svtmp(1)+1):svtmp(1)*svtmp(2)) - i*svtmp(1) + (svtmp(1)*svtmp(2))*(0:svtmp(3)-1)';
        ind(:,ind(1,:)<1) = []; % remove pre-entry ages
        if bsum
            y(:,i+1)   = sum( mat( ind ) , 2);
        else
            startage = max(1, svtmp(2) - svtmp(1) + 1 - i);
            endage   = startage - 1 + size(ind,2);
            y(:, startage:endage, i+1)   = mat( ind );
        end
        mat( ind ) = 0;
    end
end

function currcol  = xlswrite_series(data,bperc,numobs,sname,stype,startrow,dirfile,sheet,currcol, rownr, colLetters,stepyr,I)
    [nobs,nsim_hor] = size(data);
    if nobs==1 % enlarge data artificially such that all operations work properly:
        data = [data; data];
    end
    [Mstat,rownms] = displayDescriptives(data, [],[], numobs, [], ...
        repmat(bperc,1,nsim_hor),false(1,nsim_hor),false);
    if ~exist('I','var') || isempty(I)
        I = 1:length(rownms)-2;
    else
        I = ismember(rownms,I);
    end
    cMstat   =[sname      cell(1,length(rownms(I))) ;
           stype      rownms(I);
           num2cell([startrow+1/stepyr*(0:size(Mstat,2)-1)'  Mstat(I,:)']) ];
    currcol  = xlswrite_wrap(dirfile, cMstat, sheet, currcol, rownr, colLetters, true);        
end

function U = CRRAutility(v, gamma, bCEQ, weight, bCRRAscen, skip0)
%CRRAutility bepaald CRRA utility voor de uitkomsten in vector v 
% v: elementen gebruikt om CEQ te bepalen bij nut in gamma
% gamma: risicoaversie
% bCEQ = 0: utility, 1: CEQ per scenario, 2: CEQ over alle scenario's
% weight bevat evt periodegewichten
% bCRRAscen = true als CRRA over scenario's als n obs per scenario
% skip0 = true als 0 niet wordt genomen, zodat NaN's worden vermeden
    if ~exist('bCRRAscen','var')
        bCRRAscen = true;
    end
    if size(v,2)==1 && isnumeric(gamma) && bCRRAscen
        v = v'; % CRRA toegepast over scenario's
    end
    if iscell(v)
        v = [v{:}];
    end
    if ~exist('weight','var') || isempty(weight)
        weight = ones(size(v));
    else
        if sum(size(weight)==1)>0 && bCEQ~=2
            weight = repmat(weight(:)', [size(v,1) 1] );
        end
    end
    if exist('skip0','var') && skip0
        weight      = weight .* ((v>0) & ~isnan(v));
        weight_corr = mean((v>0) | isnan(v)); % neglect nonpositive assets
    else
        weight_corr = 1;
    end
    if length(gamma)>1 || ~exist('bCEQ','var') || isempty(bCEQ)
        bCEQ = 1;
    end
    if (gamma~=1)
        U = nansum( weight .* weight_corr .* (repmat(v,[1 size(weight,2)/size(v,2)]).^(1-gamma) - 1), 2) / (1-gamma);
    else
        U = nansum( weight .* weight_corr .*repmat(log(v),[1 size(weight,2)/size(v,2)]), 2);
    end
    if bCEQ % reken nut in y om naar certainty equivalents
        if bCEQ==2 % CEQ obv gemiddelde over alle scenario's
            if (gamma~=1)  
                % mean(U) = mean(sum( weight .* (CEQ.^(1-gamma) - 1), 2) / (1-gamma))
                % mean(U)*(1-gamma) = mean(sum( weight, 2)) .* (CEQ.^(1-gamma) - 1)
                % mean(U)*(1-gamma) ./ mean(sum( weight, 2)) = (CEQ.^(1-gamma) - 1)
                % (mean(U)*(1-gamma) ./ mean(sum( weight, 2)) + 1).^(1/(1-gamma)) = CEQ
                % U = repmat( ( mean(U)*(1-gamma) ./ mean(sum( weight, 2)) + 1).^(1/(1-gamma)), [size(v,1) 1]);
                U = ( mean(U)*(1-gamma) ./ mean(sum( weight, 2)) + 1).^(1/(1-gamma));
            else
                % mean(U) = mean(sum( weight .* ln(CEQ), 2)
                % mean(U) = mean(sum( weight, 2)) * ln(CEQ)
                % exp(mean(U) ./ mean(sum( weight, 2)) ) = CEQ
                U = exp(mean(U) ./ mean( sum( weight, 2)) );
            end            
        else  % CEQ per scenario
            if (gamma~=1)
                % U = sum( weight .* (CEQ.^(1-gamma) - 1), 2) / (1-gamma)
                % U*(1-gamma) = sum( weight, 2) .* (CEQ.^(1-gamma) - 1)
                % U*(1-gamma) ./ sum( weight, 2) = (CEQ.^(1-gamma) - 1)
                % (U*(1-gamma) ./ sum( weight, 2) + 1).^(1/(1-gamma)) = CEQ
                U = (U*(1-gamma) ./ sum( weight, 2) + 1).^(1/(1-gamma));
            else
                % U = sum( weight .* ln(CEQ), 2)
                % U = sum( weight, 2) * ln(CEQ)
                % exp(U ./ sum( weight, 2) ) = CEQ
                U = exp(U ./ sum( weight, 2) );
            end
        end
    end
end

function [annuityprice0, annuityprice1, Psurv0] = getAnnuityPrices(Psurv, discount_grossrate, num_working_periods, growthrate, bannuity1, vuitsmeerqteller, vuitsmeerqnoemer, vhorizon, vopslag, badditive)
    % Psurv                  = end-of-period overlevingskansen per afzonderlijke periode
    % discount_grossrate_cum = disconteringsmatrix
    % num_working_periods    = aantal werkzame jaren
    % growthrate             = groeivoet van uitbetalingen tijdens pensioenperiode, 1 = constante uitbetalingen
    % bannuity1              = annuiteit in periode 1
    % vuitsmeerqteller       = uitsmeervector van uitkeringen (0: geen uitsmeren)
    % vuitsmeerqnoemer       = uitsmeervector in disconteringsvoet (0: geen uitsmeren)
    if ~exist('growthrate','var')
        growthrate = 0;
    end
    [~, num_ages, ~]  = size(discount_grossrate+Psurv);
    
    if ~exist('vuitsmeerqteller','var') || all(vuitsmeerqteller(:)==0) || all(vuitsmeerqteller(:)==1)
        vuitsmeerqteller = ones(1,num_ages); % geen gesloten uitsmeren in benefits
    end
    if ~exist('vuitsmeerqnoemer','var') || all(vuitsmeerqnoemer(:)==0)
        vuitsmeerqnoemer = ones(1,num_ages); % geen gesloten uitsmeren in discountrate
    end 
    if ~exist('vhorizon','var') || isempty(vhorizon)
        vhorizon = 0:size(discount_grossrate,2)-1;
    else
        vhorizon = vhorizon(:)';
    end
    if ~exist('badditive','var') || isempty(badditive)
        badditive = false;
    end
    if ~exist('vopslag','var') || isempty(vopslag)
        if badditive
            vopslag = zeros(1, size(discount_grossrate,2));
        else
            vopslag = ones(1, size(discount_grossrate,2));
        end
    else
        if size(vopslag,2)==1
            vopslag = repmat(vopslag,1, size(discount_grossrate,2));
        end
    end
    [nsim,  num_ages,  num_sim_years]  = size(discount_grossrate+Psurv+vopslag);
    
    % aantal pensioenjaren
    annuityprice0   = NaN(nsim, num_ages, num_sim_years);
    if bannuity1
        annuityprice1 = zeros(nsim, num_ages, num_sim_years);
    else
        annuityprice1 = [];
    end
    Psurv0            = ones(size(Psurv));
    Psurv0(:,2:end,:) = Psurv(:,1:end-1,:); % start-of-period overlevingskansen
    if length(growthrate)==1 % geen periode- en leeftijdgevoeligheid
        growthrate = zeros(1,num_ages);
    end
    if ~badditive
        discount_grossrate = discount_grossrate.^-vhorizon;
    end

    % prijs pensioenannuteit per leeftijdsperiode
    for age=1:num_ages
        clear vbenefit num_pension_periods_to_pension_age;
        if length(growthrate)>1 && any(growthrate)
            hist_pension_periods   = age - num_working_periods-1;
            vbenefit               = zeros(size(Psurv0));
            Ibenefit               = num_working_periods+1+max(hist_pension_periods,0):size(vbenefit,2);
            benefit_level          = cumprod( [1 1+vuitsmeerqnoemer(1:end-age).*growthrate(age:end-1) ] );
            % neem inverse zodat hoge return voor goedkope annuiteit zorgt:
            vbenefit(:,Ibenefit,:) = 1./benefit_level(end-length(Ibenefit)+1:end);
        else
            vbenefit = ones(size(Psurv0));
            vbenefit(:,1:num_working_periods,:) = 0;
        end
        mat_pq = zeros(size(Psurv0));
        mat_pq(:, age:end, :) = Psurv0(:,age:end,:) .* vbenefit(:,age:end,:); % benefits worden aan het begin van de periode betaald.
        if badditive
            % let op dit klopt niet, want vopslag gaat over jaarlijkse (forward)rendementen, terwijl discount_grossrate een meetkundig gemiddelde is.
            % in zelfde periode volgt nog evt indexatie over huidige periode
            matM = mat_pq(:,age:end, :) .* vuitsmeerqteller(1:end+1-age) ...
                .* ( discount_grossrate(:, 1:num_ages+1-age, :) .* vopslag(:,age:end) ).^repmat( -vhorizon(1:num_ages+1-age),[nsim  1  num_sim_years])./ Psurv0(:,age,:);
            annuityprice0(:,age,:)                = sum( matM, 2);
            clear matM
            % aan het eind van de periode, voor indexatie en korting:
            if age < num_ages && bannuity1
                annuityprice1(:,age,:) = sum( mat_pq(:, age+1:end,:) .* vuitsmeerqteller(1:end-age) ...
                    .* ( discount_grossrate(:, 1:num_ages-age, :) .* vopslag(:,age+1:end) ).^repmat( -vhorizon(1:num_ages-age),[nsim  1  num_sim_years]), 2) ./ Psurv0(:,age,:);
            end
        else
            % in zelfde periode volgt nog evt indexatie over huidige periode
            matM = ( mat_pq(:,age:end, :) .* vuitsmeerqteller(1:end+1-age) ...
                .* discount_grossrate(:, 1:num_ages+1-age, :) ./ cumprod([ones(size(vopslag,1),1) vopslag(:,age:end-1,:)], 2) ) ./ Psurv0(:,age,:);
            annuityprice0(:,age,:) = sum( matM, 2);
            % verplichtingen0(:,1:num_ages+1-age,:) = verplichtingen0(:,1:num_ages+1-age,:) + matM;
            % aan het eind van de periode, voor indexatie en korting:
            if age < num_ages && bannuity1
                matM2 = mat_pq(:, age+1:end,:) .* vuitsmeerqteller(1:end-age) ...
                    .* discount_grossrate(:, 1:num_ages-age, :) ./ cumprod([ vopslag(:,age:end-1)] );  % ones(size(vopslag,1),1)
                annuityprice1(:,age,:) = sum( matM2, 2) ./ Psurv0(:,age,:);
            end
            clear matM matM2
        end
    end
end

function [Mstat,rowns] = displayDescriptives(mM,colns,rowns,numsim,vperc0,bperc,bgross,bdisplay)
    % mM     contains the data
    % colns  contains column names
    % rowns  contains row names
    % numsim contains number of observations/simulations
    % vperc  vector that contains percentiles
    % bperc  boolean vector that indicates if a variable is in percentages or not
    % bgross equals one if 1 + rate
    % bdisplay is true if statistics are displayed on screen (default is false)
    if ~exist('bdisplay','var') || isempty(bdisplay)
        bdisplay = false;
    else
        if (~exist('colns','var') || isempty(colns))
            colns = cellstr([repmat('t',[size(mM,2) 1]) num2str((1:size(mM,2))')])';
            colns = strrep(colns,' ','');
        end
    end
    bshort = (max(abs(std(mM)./mean(mM)),[],'omitnan') < 1e-10) && isempty(vperc0);
    if bshort
            rowns = {'mean','stdev','median'};
            vperc0 = [];
    else
        if ~exist('rowns','var') || isempty(rowns)
            rowns = {'geomean','mean','stdev','2.5% mean low','97.5% mean high','min','0.5%','1%','2.5%','5%','25%', ...
                     'median','75%','95%','97.5%','99%','99.5%','max','CRRA2,5','CRRA5','P(<0)','P(>0)'};
        end
        if ~exist('vperc0','var') || isempty(vperc0)
            vperc0 = [0.5 1 2.5 5 25 50 75 95 97.5 99 99.5];
        end
    end
    if ~exist('bperc','var') || isempty(bperc)
        bperc = true(1,size(mM,2));
    end
    if ~exist('bgross','var') || isempty(bgross)
        bgross = false(1,size(mM,2));
    else
        if length(bgross)==1
            bgross = repmat(bgross,1,size(mM,2));
        end
    end
    if bshort
        mM = mM - repmat(bgross, size(mM,1), 1);
    else
        gm     = NaN(1,size(mM,2));
        CRRA25 = NaN(1,size(mM,2));
        CRRA5  = NaN(1,size(mM,2));
        for i=1:size(mM,2)
            v     = mM(:,i);
            try
                gm(i)     = geomean(v(~isnan(v)))-bgross(i);
            catch
                gm(i)    = NaN; % geen meetkundig gemiddelde als bijvoorbeeld elementen negatief zijn.
            end
            CRRA25(i) = CRRAutility({v}, 2.5, 2, [], false, true);
            CRRA5(i)  = CRRAutility({v}, 5  , 2, [], false, true);
        end
        mM = mM - repmat(bgross, size(mM,1), 1);
        if size(mM,2)==1
            vperctile = prctile(mM, vperc0(:)')';
        else
            vperctile = prctile(mM, vperc0(:)');
        end
    end
    try % nieuwere Matlab-versies met omitnan-optie
        med0 = median(mM,1,'omitnan');
        m0 = mean(mM,1,'omitnan');
        s0 = std(mM,[],1,'omitnan');
    catch % oudere Matlab-versies met NaN-functie
        med0 = nanmedian(mM,1);
        m0 = nanmean(mM,1);
        s0 = nanstd(mM,[],1);
    end
    if bshort
        Mstat = [med0; m0; s0];
    else
        Mstat = [gm; m0; s0; m0+norminv(.025)*s0/numsim^.5; m0+norminv(.975)*s0/numsim^.5; min(mM); ...
                vperctile; max(mM,[],1); CRRA25; CRRA5; mean(mM<0,1); mean(mM>0,1)];
    end
    perctable = repmat(100.^(bperc(:)'),[size(Mstat,1) 1]);
    Mstat     = perctable .* Mstat;
    if bdisplay
        disp(array2table(Mstat,'VariableNames',matlab.lang.makeUniqueStrings( matlab.lang.makeValidName(colns) ),'RowNames', rowns) );
    end
end

function Acoh = extractDiag(A,dim)
% deze functie kan cohorteffecten bepalen
    % A:      three dimensional matrix
    % dim(1): scenario
    % dim(2): age
    % dim(3): time
    if ~exist('dim','var') || isempty(dim)
        dim = 1:3; % default: dim=1:3
    end
    [nsim,numages,nhor] = size(A);
    A           = permute(A, dim(3:-1:1));
    firstelem   = [nhor:-1:1 nhor*(1:numages-1)+1];
    firstperiod = [nhor:-1:1 ones(1,numages-1)];
    lastperiod  = min(firstperiod+numages-1, ...
                    [nhor*ones(1,length(firstperiod)-min(nhor,numages)) min(nhor,numages):-1:1] );
    Acoh        = NaN(nhor, length(firstelem), nsim);
    coh         = 0;
    stepsize    = nhor+1;
    % Acoh
    % first dim: time
    % second dim: cohort
    % third dim: scenario
    for i = 1:length(firstelem)
        coh = coh+1;
        numperiods = lastperiod(i)-firstperiod(i)+1;
        Acoh(firstperiod(i):lastperiod(i),coh,:) = ...
            A( repmat((firstelem(i)+stepsize*(0:numperiods-1))',[1 nsim] ) ...
                + nhor*numages*repmat(0:(nsim-1),[numperiods 1]) );
    end
    Acoh = Acoh(:,end:-1:1,:); % rank cohorts from oldest to youngest
    [~,dimloc] = sort(dim(3:-1:1));
    Acoh       = permute(Acoh, dimloc);
end

function bondexposure_maturity_t = getBondexposures(assets_t, par, profiel, RTSmaturities, nsim, num_work_periods)
    % vind blootstelling per obligatielooptijd
    cashcol     = (RTSmaturities==par.cash_mat);
    Psurv0      = [1  profiel.Psurv(1:end-1)];
    if strcmp('matching', par.bondmat)
        % geen payout_growth_profile
        vbondexposure_maturity = zeros(nsim, length(RTSmaturities));
        for j=1:size(assets_t,2)
            % als long-positie in bonds: long  pensioenlooptijden, want bedoeld als pensioeninkomen
            %                            short werklooptijden, want toekomstig loon is onderpand van lening
            pensAgesj    = max(j+1,num_work_periods+1):length(Psurv0);
            Psurv0Pensj  = Psurv0( pensAgesj ) / Psurv0(j);
            maturitiesPens = ismember( RTSmaturities, ( pensAgesj - j) );
            if strcmp(par.bondmat,'matching')
                vbondexposure_maturity( :,maturitiesPens) = vbondexposure_maturity( :,maturitiesPens) + ...
                    (assets_t( :,j) .* profiel.bond( :,j)) .* Psurv0Pensj / sum(Psurv0Pensj);
            end
        end
    end
    if strcmp(par.bondmat,'cash')
        vbondexposure_maturity(:, cashcol)     = vbondexposure_maturity(:, cashcol) + sum( assets_t .* profiel.bond, 2);
    elseif isnumeric(par.bondmat)
        vbondexposure_maturity(:, RTSmaturities==par.bondmat) = vbondexposure_maturity(:, RTSmaturities==par.bondmat) + sum( assets_t .* profiel.bond, 2);
    end
    vbondexposure_maturity(:, cashcol)     = vbondexposure_maturity(:, cashcol) + sum( assets_t .* profiel.cash, 2);
    bondexposure_maturity_t = vbondexposure_maturity ./ sum( assets_t, 2);
end

function [proj_return_opslag, EexcessR, stock_exposure, fundproj_return_opslag] = getProjReturnOpslag(par, profiel)
    proj_return_opslag = 1 + par.proj_return_opslag;
    stock_exposure     = profiel.stock;
    EexcessR           = [];
    fundproj_return_opslag = proj_return_opslag;
end

function stock_exposure       = getStockExposureToed(assets_t, hcap_t, profiel_stock, profiel_toed_staffel)
    wealth_t                     = assets_t;
    wealth_t(:,1:size(hcap_t,2)) = wealth_t(:,1:size(hcap_t,2)) + hcap_t;
    fund_stock_exposure  = nansum(profiel_stock .* assets_t, 2) ./ nansum(assets_t, 2);
    % zorg dat gewogen gemiddelde gelijk is aan 1:
    mtmp                 = profiel_toed_staffel .* wealth_t ./ assets_t;
    toed_exposure        = mtmp ./ (nansum(mtmp .* assets_t, 2) ./ nansum( assets_t, 2) );
    toed_exposure(isnan(toed_exposure)) = 0;
    stock_exposure       = toed_exposure .* fund_stock_exposure;
end

function economy0 = read_set(spath,par,vmatur,bdisplay)
    % leest scenario's in die met Matlab zijn gegenereerd
    if ~isempty(par.sfile)
        load([spath par.sfile]);
    end
    
    par.nsim_hor = par.nsim_hor + (par.burnin-1);
    
    nsim_hor2   = par.nsim_hor+2;
    maxmaturity = 100;
    all_ones1 = ones(par.nsim, nsim_hor2);
    if strcmp(par.scenarioset,'all 1') % scenario with all ones
        % vlakke RTS die over de tijd varieert. Hier: rente gelijk aan 10-jaarsrente
        economy0.stock_returns    = all_ones1;
        economy0.Estockreturn     = 1;
        economy0.Estockretgeom    = 1;
        economy0.bond_portf_maturities   = vmatur;
        economy0.bond_returns     = ones(par.nsim, length(economy0.bond_portf_maturities), nsim_hor2);
        economy0.cash_return   = all_ones1;
        economy0.price_inflation  = all_ones1;
        economy0.wage_inflation   = all_ones1;
        economy0.RTS              = ones(par.nsim, 1+maxmaturity, nsim_hor2);
        economy0.RTS_maturities   = linspace(0,maxmaturity,1+maxmaturity);
        return
    end
    if strcmp(par.scenarioset,'KNW') || strcmp(par.set_stocks,'KNW') || strcmp(par.set_RTS,'KNW')
        [economy0,~,muY,SigmaY,muYgeomP] = generate_KNW_set(true,[],par,vmatur,false,false);
        economy0.VarX          = SigmaY(1:size( economy0.nominal_term_structure.b, 2), 1:size( economy0.nominal_term_structure.b, 2));
        economy0.cash_return   = squeeze( economy0.bond_returns(:,economy0.RTS_maturities==par.cash_mat,:) ); % ones(par.nsim, nsim_hor1);
        economy0.Estockreturn  = 1+muY(4);
        economy0.Estockretgeom = 1+muYgeomP(4);
        clear muY muYgeomP SigmaY;
    end
    if ~strcmp(par.scenarioset,'KNW') && (~strcmp(par.set_stocks,'KNW') || ~strcmp(par.set_RTS,'KNW'))
        if strcmp(par.set_RTS,'scenarioset') || ~isempty(par.sfile)
            [economy0.RTS,economy0.RTS_maturities] = constructRTS(economy0, 1);
            economy0.RTS             = 1+economy0.RTS(1:par.nsim,:,1:nsim_hor2);
            economy0.bond_portf_maturities   = economy.bond_portf_maturities;
            economy0.bond_returns    = NaN(par.nsim, length(economy0.bond_portf_maturities), nsim_hor2);
            % overschrijf evt cash return
            economy0.cash_return = squeeze( economy0.RTS(:,economy0.RTS_maturities==par.cash_mat,:) );
            RTSlocs                          = find(ismember(economy0.RTS_maturities,economy0.bond_portf_maturities(2:end)));
            economy0.bond_returns            = NaN(par.nsim, length(economy0.bond_portf_maturities), nsim_hor2-1);
            % looptijd nul is altijd continu tegen cashtarief:
            economy0.bond_returns(:, economy0.bond_portf_maturities==0, :)     = economy0.RTS(:, 1, 1:nsim_hor2-1);
            % langere looptijden worden iedere periode geherbalanceerd
            economy0.bond_returns(:,2:end,:) = economy0.RTS(:,RTSlocs,1:end-1) ...
                                                    .^repmat(economy0.bond_portf_maturities(2:end),[par.nsim 1 nsim_hor2-1]) ...
                                                   ./ economy0.RTS(:,RTSlocs-1,2:end) ...
                                                    .^repmat(economy0.bond_portf_maturities(2:end)-1,[par.nsim 1 nsim_hor2-1]);
        else
            if(isnumeric(par.set_RTS))
                if par.set_RTS > 1
                    par.set_RTS = par.set_RTS - 1;
                end
                economy0.RTS                     = repmat(1+par.set_RTS,[par.nsim 1+maxmaturity nsim_hor2]);       
                economy0.RTS_maturities          = linspace(0,maxmaturity, 1 + maxmaturity);
                economy0.bond_portf_maturities   = vmatur;
                economy0.bond_returns            = NaN(par.nsim, length(economy0.bond_portf_maturities), nsim_hor2-1);
                economy0.bond_returns            = repmat( 1+par.set_RTS, [par.nsim length(economy0.bond_portf_maturities) nsim_hor2-1]);
                economy0.cash_return             = (1+par.set_RTS) * all_ones1;
            end
        end
        if strcmp(par.set_stocks,'Black Scholes')
            % simulate gross return using Black-Scholes model
            % log( S_t / S_{t-1} ) =  N( r + L*sig - sig^2/2, sig^2 );
            %      S_t / S_{t-1}   = LN( r + L*sig - sig^2/2, sig^2 );
            %   E[ S_t / S_{t-1} ] = exp(r + L*sig );
            % Var[ S_t / S_{t-1} ] = [ exp( sig^2 ) - 1 ] * exp( 2*(r + L*sig) );
            % drift = r + L*sig
            % economy0.stock_returns = ones(par.nsim, size(economy0.cash_return,2));
            timestep               = 1;
            if length(unique(economy0.cash_return))==1
                mu = log(economy0.cash_return(1,2)) + par.BS_risk_par * par.BS_vol_par * timestep;
            else
                mu = log(economy0.cash_return) + par.BS_risk_par * par.BS_vol_par * timestep;
            end
            mZ2                    = randn(par.nsim,size(economy0.cash_return,2));
            economy0.stock_returns = exp( (mu - .5 * timestep * par.BS_vol_par^2) + par.BS_vol_par *sqrt(timestep) * mZ2 );
            economy0.stock_returns = economy0.stock_returns(:, 1:nsim_hor2-1);
            economy0.Estockreturn  = exp( mu - .5 * timestep * par.BS_vol_par^2 + par.BS_vol_par^2*timestep/2); % mean of lognormal distribution
            economy0.Estockretgeom = exp( mu - .5 * timestep * par.BS_vol_par^2 );
        elseif isnumeric(par.set_stocks)
            if par.set_stocks > 1
                par.set_stocks = par.set_stocks - 1;
            end
            economy0.stock_returns = repmat( (1 + par.set_stocks), par.nsim, nsim_hor2-1);
            economy0.Estockreturn  = 1 + par.set_stocks;
            economy0.Estockretgeom = economy0.Estockreturn;
        else
            economy0.stock_returns = 1+economy.stock_return(1:par.nsim, 1:nsim_hor2-1);
        end
    end
    if ~strcmp(par.price_inflation,'scenarioset')
        if isnumeric(par.price_inflation)
            economy0.price_inflation = 1+par.price_inflation;
        end
        economy0.price_index = economy0.price_inflation.^(-1:par.nsim-1);
    else
        economy0.price_index = [ones(par.nsim,1) cumprod(economy0.price_inflation,2)];
        economy0.price_index = economy0.price_index ./ economy0.price_index(:,2);
    end
    if ~strcmp(par.wage_inflation,'scenarioset')
        if isnumeric(par.wage_inflation)
            if par.wage_inflation > 1
                par.wage_inflation = par.wage_inflation - 1;
            end
            economy0.wage_inflation = (1+par.wage_inflation); % * ones(par.nsim, par.nsim_hor);
            economy0.wage_index     = economy0.wage_inflation.^(-1:par.nsim-1);
        end
    else
        economy0.wage_index      = [ones(par.nsim,1) cumprod(economy0.wage_inflation,2)];
        economy0.wage_index      = economy0.wage_index ./ economy0.wage_index(:,2);
        economy0.franchise_index = [ones(par.nsim,1) cumprod(economy0.wage_inflation,2)];
        economy0.franchise_index = economy0.franchise_index ./ economy0.franchise_index(:,2);
    end
    par.nsim_hor = par.nsim_hor - (par.burnin-1);
    if par.burnin < 1
        dupl = 1-par.burnin;
        for s = {'RTS','bond_returns','price_inflation','state_variables_term_structure','stock_returns','wage_inflation','cash_return'}
            if isfield(economy0, s{1})
                switch length( size(economy0.(s{1})) )
                    case 2
                        if min(size(economy0.(s{1})))>1
                            economy0.(s{1}) = [repmat( economy0.(s{1})(:,1), 1, dupl)      economy0.(s{1})  ];
                        end
                    case 3
                        mtmp                            = economy0.(s{1});
                        economy0.(s{1})                 = NaN(size(economy0.(s{1})) + [0 0 dupl]);
                        economy0.(s{1})(:,:,1:dupl)     = repmat( mtmp(:,:,1), 1, 1, dupl);
                        economy0.(s{1})(:,:,dupl+1:end) = mtmp;
                        clear mtmp
                end
            end
        end
    else
        if par.burnin > 1
            for s = {'RTS','bond_returns','price_inflation','state_variables_term_structure','stock_returns','wage_inflation','cash_return'}
                switch length( size(economy0.(s{1})) )
                    case 2
                        economy0.(s{1}) = economy0.(s{1})(:,par.burnin+1:end);
                    case 3
                        economy0.(s{1}) = economy0.(s{1})(:,:,par.burnin+1:end);
                end
            end
        end
    end
    
    if bdisplay
        bondcols = ismember( economy0.RTS_maturities, economy0.bond_portf_maturities);
        nfactor  = length( economy0.stock_returns(:,end) ) / length( economy0.price_inflation(:,end) );
        if size(economy0.bond_returns,2)==size(economy0.RTS,2)
            Mm       = [ economy0.stock_returns(:,end)  squeeze( economy0.bond_returns(:,bondcols,end) )  squeeze( economy0.RTS(:,bondcols,end) )  repmat( economy0.price_inflation(:,end), nfactor, 1) ];
        else
            Mm       = [ economy0.stock_returns(:,end)  squeeze( economy0.bond_returns(:,:,end) )         squeeze( economy0.RTS(:,bondcols,end) )  repmat( economy0.price_inflation(:,end), nfactor, 1) ];
        end
        k        = length(economy0.bond_portf_maturities);
        vv       = join([["stock_return"; repmat("bond_return",k,1); repmat("RTS",k,1); "price_inflation"] [""; repmat( num2str(economy0.bond_portf_maturities(:)), 2,1); ""] ],"");
        disp(' ')
        disp([  "" "mean (%)"   "st.dev (%)";
                vv 100*mean(Mm-1)' 100*std(Mm)'])
        disp(' ')
        disp('correlation matrix')
        disp([  "" vv';
        vv corr(Mm)])
    end
end

function [RTS, maturities] = constructRTS(economy, stepyr)
    nsim    = size(economy.stock_returns, 1);
    sim_hor = size(economy.stock_returns, 2);
    if stepyr>1
        economy.nominal_term_structure.a = from_yearly_to_periodically(economy.nominal_term_structure.a, stepyr,1);
        economy.nominal_term_structure.b = from_yearly_to_periodically(economy.nominal_term_structure.b, stepyr,1);
    end
    % construeer RTS met gesimuleeerde states
    RTS = NaN(nsim, length(economy.nominal_term_structure.maturities), sim_hor+1);
    for s=1:nsim
        RTS(s,:,:) = exp( repmat( economy.nominal_term_structure.a(:), [1 sim_hor+1]) ...
                    + economy.nominal_term_structure.b*squeeze(economy.state_variables_term_structure(s,:,:))); %  - 1
    end
    maturities = economy.nominal_term_structure.maturities;
end

function M = from_yearly_to_periodically(v,n,dim)
    if ~exist('dim','var')
        dim = 1;
    end
    if dim==2
        v = v';
    end
    M = NaN(size(v).*[n 1]);
    stepsizes  = diff(v,1,dim)/n;
    for i=1:size(v,2)
        vM     = v(:,i)' + (-(n-1):0)'*[stepsizes(1) stepsizes(:,i)'];
        M(:,i) = vM(:);
    end
    if dim==2
        M = M';
    end
end

function [par, profiel,sheets_cell] = readInput(inputfile,sheet,bclearsheet)
    % bclearsheet:   true:  kolommen moeten hoe dan ook schoon
    %                false: kolommen hoeven niet per se schoon 
    % lees parameters in
    if isempty(sheet)
        [~,sheets_cell] = xlsfinfo(inputfile);
    else
        sheets_cell = {sheet};
    end
    % n sheet met leeftijdsprofielen, andere bevatten contracten
    numcells = length(sheets_cell);
    par      = cell(1, numcells);
    profiel  = cell(1,numcells);
    % lees leeftijdsprofielen in
    try
        [~,~,C]  = xlsread(inputfile,'profiel');
        [row,colvar] = find(strcmp(C,'age'));
        for i = colvar:size(C,2)
            s    = C{row,i};
            dat1 = cell2mat(C(row+1:size(C,1),i));
            try
                profiel0.(s) = dat1;
            catch % stop bij lege cel
                break
            end
        end
    catch
    end
    
    try
        [~,~,C]       = xlsread(inputfile,'Psurv');
        [row, colvar] = find(strcmp(C,'age'));
        if isfield(profiel0,'age')
            I = [false; ismember(cell2mat(C(row+1:end,colvar)) , profiel0.age)];
        else
            I = [false; true(size(C,1), 1)];
        end
        profiel0.Psurv  = cell2mat(C(I,colvar+1:size(C,2)));
    catch
    end
    
    colLetters = xlsColNum2Str(1:500); % converteer kolomnummer naar kolomletter  
    for i=1:numcells
        if ~ismember(sheets_cell{i},{'main','profiel','stats','Psurv'})
            [~,~,C]       = xlsread(inputfile, sheets_cell{i});
            [~,colremove] = find(strcmp(C,'geomean'));
            if ~isempty(colremove)
                if colremove(1) >= 5
                    [C{:,colremove(1):end}] = deal(NaN);
                    xlswrite_wrap(inputfile, C(:,colremove(1):end), sheets_cell{i},colremove(1),1,colLetters,bclearsheet);
                end
            end
            [row,colvar]    = find(strcmp(C,'variable name'));
            [~,colmod]      = find(strcmp(C,'in model'));
            for j = row+1:size(C,1)
                s = C{j,colvar(1)};
                dat1 = C{j,colmod(1)};
                try
                    tmp    = str2double(dat1);
                    if (~isempty(tmp) && ~isnan(tmp)) || strcmp(dat1,'NaN')
                        dat1 = tmp;
                    end
                    tmploc = strfind(s,'.');
                    if isempty(tmploc)
                        par{i}.(s) = dat1;
                    else
                        par{i}.(s(1:tmploc-1)).(s(tmploc+1:end)) = dat1;
                    end
                catch % sla lege rijen over
                end
            end
            switch par{i}.scenarioset
                case {'componenten','all 1','KNW'}
                    par{i}.sfile       = [];
                otherwise
                    par{i}.sfile       = [par{i}.scenarioset '.mat'];
            end
            if exist('profiel0','var')
                profiel{i}     = profiel0;
                if isfield(profiel0,'age')
                    par{i}.max_age = min(par{i}.max_age, max(profiel0.age));
                else
                    par{i}.max_age = min(par{i}.max_age, profiel{i}.age(1+find(profiel{i}.Psurv(:,1)>0,1,'last')));
                end
                if size(profiel{i}.Psurv, 2) > 1 % vul evt ontbrekende overlevingskansen aan
                    profiel{i}.Psurv = [profiel{i}.Psurv(:,1)  profiel{i}.Psurv];
                    profiel{i}.Psurv(:,end+(1:par{i}.nsim_hor+2-size(profiel{i}.Psurv,2))) ...
                        = repmat(profiel{i}.Psurv(:,end),1,par{i}.nsim_hor+2-size(profiel{i}.Psurv,2));
                end
            else
                if strcmp(par{i}.asset_mix,'specifiek leeftijdsprofiel')
                    error('ontbrekend leeftijdsprofiel')
                end
            end
            if ~isfield(profiel{i},'Psurv')
                profiel{i}.Psurv     = ones(1,par{i}.max_age+1-par{i}.first_exposure_age);
                profiel{i}.Psurv(end)= 0;
            end
            num_ages = par{i}.max_age + 1 - par{i}.first_exposure_age;
            if strcmp(par{i}.asset_mix, 'lineair')
                % lineair profiel obv exposures op aantal leeftijden
                v_age    = [par{i}.first_exposure_age   par{i}.pension_age-1   ...
                    par{i}.pension_age   min([71 85],par{i}.max_age)   par{i}.max_age];
                for s = {'bond','stock'}
                    v_tmp = [par{i}.([s{1} '_first_exposure_age'])  par{i}.([s{1} '_last_work_age']) ...
                        par{i}.([s{1} '_pension_age'])  par{i}.([s{1} '_71']) ...
                        par{i}.([s{1} '_85'])  par{i}.([s{1} '_max_age'])];
                    I     = ~isnan(v_tmp);
                    profiel{i}.(s{1}) = NaN(1,num_ages);
                    profiel{i}.(s{1})(1:(max(v_age(I))+1-v_age(1))) = linprofiles( v_age(1), v_age(I), v_tmp(I), 1);
                end
                profiel{i}.cash    = 1 - profiel{i}.stock - profiel{i}.bond;
            end
            if ~isnan(par{i}.toed) && par{i}.toed > 0
                if strcmp(par{i}.toed_renteafdekking,'als obligatielifecycle')
                    profiel{i}.toed_bond = profiel{i}.bond;
                end
                % lineair toedelingsprofiel
                vtmp             = [par{i}.pension_age-1             par{i}.pension_age                    71];
                vtmp_toed_return = [par{i}.toed_return_last_workage  par{i}.toed_return_pension_age  par{i}.toed_return_71];
                if par{i}.max_age<=71
                    vtmp_toed_return(vtmp==71) = [];
                    vtmp(vtmp==71)             = [];
                end
                [~,loc]  = sort(vtmp);

                v_age    = [par{i}.first_exposure_age                 vtmp(loc)                 par{i}.max_age];
                v_toed   = [par{i}.toed_return_first_exposure_age     vtmp_toed_return(loc)      par{i}.toed_return_max_age];
                I        = ~isnan(v_toed);
                profiel{i}.toed_staffel    = NaN(1,num_ages);
                profiel{i}.toed_staffel(1:(max(v_age(I))+1-v_age(1))) = linprofiles( v_age(1), v_age(I), v_toed(I), 1);
                clear vtmp vtmp_toed_return v_toed
            end
            if ~isempty(profiel{i})
                if isfield(profiel{i},'age')
                    I = profiel{i}.age < par{i}.first_exposure_age | profiel{i}.age > par{i}.max_age;
                    profiel{i}.age(I) = [];
                end
                for s=setdiff(fieldnames(profiel{i}),'age')'
                    v = profiel{i}.(s{1});
                    if isfield(profiel{i},'age')
                        try
                            v(I,:) = [];
                        catch
                        end
                    else
                        v(par{i}.max_age+1:end,:) = [];
                    end
                    profiel{i}.(s{1}) = v;
                end
                if ~isfield(profiel{i},'wage')
                    profiel{i}.wage      = ones(1,par{i}.max_age+1-par{i}.first_exposure_age);
                    profiel{i}.wage(1:par{i}.first_work_age-par{i}.first_exposure_age) = 0;
                    profiel{i}.wage((par{i}.pension_age:par{i}.max_age)-par{i}.first_exposure_age+1) = 0;
                end
                if ~isfield(profiel{i},'franchise')
                    profiel{i}.franchise = zeros(1,par{i}.max_age+1-par{i}.first_exposure_age);
                end
                if isfield(profiel{i},'age')
                    par{i}.first_exposure_age = max(par{i}.first_exposure_age, ...
                        profiel{i}.age(find(profiel{i}.wage>0,1)));
                end
            else
                profiel{i}.wage      = ones(1,par{i}.max_age+1-par{i}.first_exposure_age);
                profiel{i}.wage(1:par{i}.first_work_age-par{i}.first_exposure_age) = 0;
                profiel{i}.wage((par{i}.pension_age:par{i}.max_age)-par{i}.first_exposure_age+1) = 0;
                profiel{i}.franchise = zeros(1,par{i}.max_age+1-par{i}.first_exposure_age);
            end
        end
    end
end

function vv = rescale(vv, assets, Inonneg_ages)
    % herschaal zodanig dat negatieve totaalrendementen niet mogelijk zijn.
    % met andere woorden, de totale exposure is maximaal 1;
    if ~isempty(Inonneg_ages)
        v_tmp   = sum( vv(:,Inonneg_ages) .* assets(:,Inonneg_ages), 2) ./ sum(assets(:,Inonneg_ages), 2);
        vv(:,Inonneg_ages)  = vv(:,Inonneg_ages) ./ max(v_tmp,1);
    end
end

function A = Afactor(v_payout, vrbyperiod, brforward, opslagfractie)
% opslagfractie heeft betrekking op evt opslag op cumulatieve rendement.
% opslag op jaarlijkse rendement moet verwerkt zijn in brforward.
    brforward = opslagfractie * cumprod(brforward);
    if length(vrbyperiod)==1
        A = sum( v_payout ./ cumprod( [ 1  repmat(vrbyperiod,1,length(v_payout)-1) ] ) ./ brforward ) ;
    else
        A = sum( v_payout ./  vrbyperiod(:,1:size(v_payout,2)).^(0:size(v_payout,2)-1) ./ brforward , 2);
%        A = sum( v_payout./  cumprod( [ ones(size(rho,1),1)  rho(:,1:size(v_payout,2)-1) ] , 2) , 2);
    end
end

function v_intp = linprofiles(age0, v_age, p, stepyr, breciproke)
%LINPROFILES maak lineair profiel van interpolatieprofielen
% age0  is de eerste leeftijd in return vector v
% v_age is een vector met leeftijden
% p     is een vector met fracties
% breciproke is als reciproke lineair is: 1/n, 1/(n-1), 1/(n-2), ..., 1/1
% v_intp is de outputvector met lineair genterpoleerde fracties
    if ~exist('stepyr','var')
        stepyr = 1;
    end
    if ~exist('breciproke','var')
        breciproke = false;
    end
    if length(p)==1
        p = repmat(p,1,length(v_age));
    end
    v_age0 = v_age - age0 + 1;
    v_intp = zeros( v_age0(end)*stepyr, 1);
    for i=2:length(v_age0)
        if ~breciproke
            v_intp( v_age0(i-1)*stepyr:v_age0(i)*stepyr ) = linspace(p(i-1), p(i), stepyr*(v_age0(i)-v_age0(i-1))+1);
        else
            v_intp( v_age0(i-1)*stepyr:v_age0(i)*stepyr ) = 1./linspace(1/p(i-1), 1/p(i), stepyr*(v_age0(i)-v_age0(i-1))+1);
        end
    end
end

function [lastcol,lastrow] = xlswrite_wrap(file, data, sheet, firstcol, firstrow, colLetters, bclearsheet)
% write data at range of sheet on file
% wrapper function of xlswrite
% firstcol   is first column number
% firstrow   is first row number
% colLetters is an optional cell array with column letters in each cell
% bclearsheet:   true:  kolommen moeten hoe dan ook schoon
%                false: kolommen hoeven niet per se schoon 
% lastcol    is rightmost column number
% lastrow    is bottom row number
    if ~exist('bclearsheet','var')
        bclearsheet = true;
    end
    if ~exist('colLetters','var') || isempty(colLetters)
        colLetters = xlsColNum2Str(1:500); % converteer kolomnummer naar kolomletter  
    end
    range = [colLetters{firstcol} num2str(firstrow)];
    warning('off', 'MATLAB:xlswrite:AddSheet')
    fun   = @(x)xlswrite(file, data, num2str(sheet), range);
% deze functies blijken slecht/langzaam/niet te werken:
%         fun   = @(x)writecell  (data, file, 'Sheet', num2str(sheet), 'Range', range);
%         fun   = @(x)writematrix(data, file, 'Sheet', num2str(sheet), 'Range', range);
    try
        fun();
    catch
        if bclearsheet % sheet moet hoe dan ook schoon
            pause(1); % soms is wachten op voorgaand schrijven nodig
            try
                fun();
            catch
                pause(1); % soms is wachten op voorgaand schrijven nodig
                try
                    fun();
                catch
                    disp(['Please close file ' file])
                    disp('Press any key to continue')
                    pause;
                    fun();
                end
            end
        end
    end
    lastcol = firstcol + size(data,2) - 1;
    lastrow = firstrow + size(data,1) - 1;
end

function [economyP, economyQ, muY, SigmaP, muYgeomP] = generate_KNW_set(bsimulate,init_date,par,vmatur,bsave,bQ,bscaleQ)
    % bsimulate = simuleer nieuwe KNW-set
    % initdate  = maand (ultimo) waarop initialisatie wordt bepaald.
    %             keuzes [201702 201703 201912 202001:202012];
    %             initdate niet relevant als stype = 'HBT2017Q1'
    % par       = parameterset, opties:zie functie getPars(sset) voor de mogelijkheden
    % vmatur    = bond maturities
    % bsave     = 1 if sets are saved
    % bQ        = 0 if Q-set is obtained
    % bscaleQ   = 1 if Q-set is scaled
    
    % voorbeeldaanroep:
    % vStats = generate_KNW_set(true,true,201702,'low',true,true)
    vTau      = (0: par.max_age-par.first_work_age) ; % meegenomen jaren voor RTS
    if strcmp(par.bondmat,'renteafdekking') || strcmp(par.maturity_sol,'renteafdekking')
        vTau = [vTau  par.max_age-par.first_work_age+1:100];
    end
    
    vTauBond1 = 1 + vmatur;
    currpath  = pwd;
    num       = strfind(currpath,'v2017\') + length('v2017\') - 1;
    datapath  = [currpath(1:num) '.\data\Matlab_files\']; %ONE OR TWO DOTS
    if ~exist('bQ','var')
        bQ = false;
    end
    if ~exist('bscaleQ','var')
        bscaleQ = true;
    end
    if bsimulate % simulate new set
        % Step 0: initialize the parameters
        KNW      = par.KNW;
        KNW.vlam  = [KNW.vLam1   KNW.vLam2];
        KNW.del1pi= [KNW.del1pi1 KNW.del1pi2];
        KNW.del1R = [KNW.del1R1  KNW.del1R2];
        KNW.mLam  = [KNW.mLam11  KNW.mLam12; KNW.mLam21 KNW.mLam22];
        % KNW.K     = [KNW.K11     0         ; KNW.K21    KNW.K22];
        KNW.sigPi = [KNW.sigPi1  KNW.sigPi2  KNW.sigPi3];
        KNW.sigS  = [KNW.sigS1   KNW.sigS2   KNW.sigS3   KNW.sigS4];
        X0        = [KNW.X0_1; KNW.X0_2];
        KNW.X0_1  = [];
        KNW.X0_2  = [];
        clear sset
        % dit is bij P- en Q-set hetzelfde:
        [a,b]                            = getRTScoeffs(KNW.del0R, KNW.del1R, KNW.vlam, KNW.mLam, KNW.K, vTau);
        economy.nominal_term_structure.a = a;
        economy.nominal_term_structure.b = b;
        % rele cofficinten volgt uit (14), (23)/(24) en (31)/(32):
        % KNW.del0r                       = KNW.del0R-(KNW.del0pi-KNW.sigPi(1:2)*KNW.vlam(:));
        % KNW.del1r                       = KNW.del1R-(KNW.del1pi-KNW.sigPi(1:2)*KNW.mLam);
        % [a,b]                           = getRTScoeffs(KNW.del0r, KNW.del1r, KNW.vlam-KNW.sigPi(1:2), KNW.mLam, KNW.K, vTau);
        % economy.real_term_structure.a   = a;
        % economy.real_term_structure.b   = b;
        
        economy.bscale                            = bscaleQ;
        economy.nominal_term_structure.maturities = vTau;
        economy.bond_portf_maturities             = vTauBond1-1;
        economy.scenario                          = 'KNW';
        economy.state_init_date                   = init_date;
        
        if isempty(X0)
            % X0 obv waargenomen looptijden en RTS van ultimo juni 2020 in CEP2017:
            X0 = getX0(economy.nominal_term_structure.a, economy.nominal_term_structure.b,[1:10 12 15 20 25 30 40 50 60 100], economy.state_init_date);
            disp('X0 is set using RTS CEP2017!');
        end
        economy.init_state = X0;  % initile state
        
        %% Step 1: simulate the P & Q set
        % P-set: implementeer simulatie X (state variables), inflatie, aandelen Draper (42)
        [economyP, mZ, ~, SigmaP, muY, ~, muYgeomP] = simulate(economy, KNW.del0pi, KNW.del1pi, KNW.del0R, KNW.del1R,zeros(1,length(KNW.K)), ...
            KNW.mLam,     KNW.K, KNW.sigPi, KNW.sigS, KNW.vlam, KNW.etaS, vTau, vTauBond1, [], par.nsim, par.nsim_hor, X0, datapath, 'P',bsave,false);
        if bQ
            % Q-set: implementeer Q-simulatie X (state variables), inflatie, aandelen Draper (58)
            [economyQ, ~, ~] = simulate(economy, KNW.del0pi, KNW.del1pi, KNW.del0R, KNW.del1R,            -KNW.vlam, ...
                zeros(size(KNW.mLam)), KNW.K+KNW.mLam, KNW.sigPi, KNW.sigS, KNW.vlam,   0, vTau, vTauBond1, mZ, Nscen, simhor, X0, datapath, 'Q',bsave,bscaleQ);
        else
            economyQ = [];
        end
        clear mZ        
        
        %% Step 2: generate statistics
        RTS_P = economyP.RTS; % constructRTS(economyP);
        if bQ
            RTS_Q = economyQ.RTS; % constructRTS(economyQ);
        end
        %% figure statistics
        close all; figure;
        vyears = 2020:2100;
        vyrRTS = [2020 2060];
        for i=1:2
            subplot(2,2,i)
            plot(vTau, 100*(squeeze(geomean(1 + RTS_P(:,:, 1)) ) -1) ), hold on;
            plot(vTau, 100*(squeeze(geomean(1 + RTS_Q(:,:, 1)) ) -1) )
            xlabel('looptijd (jaar)'), ylabel('% per jaar'), title(['RTS ' num2str(vyrRTS(i))]), legend({'P','Q'},'Location','SouthEast')
        end

        subplot(2,2,3)
        plot(vyears, 100*(squeeze(geomean(1 + economyP.price_inflation(:, vyears-year0+1)) - 1) ) ), hold on;
        if bQ
            plot(vyears, 100*(squeeze(geomean(1 + economyQ.price_inflation(:, vyears-year0+1)) - 1) ) )
        end
        xlabel('jaar'), ylabel('% per jaar'), title('prijsinflatie'), legend({'P','Q'},'Location','SouthEast')

        subplot(2,2,4)
        plot(vyears, 100*(squeeze(geomean(1 + economyP.stock_return(:, vyears-year0+1)) - 1) ) ), hold on;
        if bQ
            plot(vyears, 100*(squeeze(geomean(1 + economyQ.stock_return(:, vyears-year0+1) ) -1 ) ) )
        end
        xlabel('jaar'), ylabel('% per jaar'), title('aandeelrendement'), legend({'P','Q'},'Location','SouthEast')
        %% display the number of statistics
        priceP = economyP.price_inflation(:,vyears-year0+1);
        stockP = economyP.stock_returns(:,vyears-year0+1);
        if bQ
            priceQ = economyQ.price_inflation(:,vyears-year0+1);
            stockQ = economyQ.stock_returns(:,vyears-year0+1);

            descr_sim_pars = 100*([geomean(1 + priceP(:))-1 std(priceP(:))  geomean(1 + priceQ(:))-1 std(priceQ(:)) ...
                geomean(1 + stockP(:))-1 std(stockP(:))  geomean(1 + stockQ(:))-1 std(stockQ(:))]);
            disp('meetkundige gemiddeldes en standaard deviaties')
            disp('    prijsinflP          prijsinflQ          aandeelP            aandeelQ')
            disp('    mu        std       mu         std      mu        std       mu        std')
        else
            descr_sim_pars = 100*([geomean(1 + priceP(:))-1 std(priceP(:)) geomean(1 + stockP(:))-1 std(stockP(:)) ]);
            disp('meetkundige gemiddeldes en standaard deviaties')
            disp('    prijsinflP          aandeelP            ')
            disp('    mu        std       mu        std       ')
        end
        disp(descr_sim_pars)
    end
end


%% The default values of the parameters
% gets used in function getDefaultPars (which gets used in generate_KNW_set)
function [del0pi,del1pi,del0R,del1R,K,vlam,mlam,sigPi,sigS,etaS,X0] = getDefaultPars()
    % default parameters in TFT: afgeronde gecalibreerde waardes zijn gerapporteerd in Draper (2014)
    % https://www.cpb.nl/sites/default/files/publicaties/download/cpb-background-document-financial-market-model-netherlands.pdf
    %NOTE: X0 is Geometric Brownian Motion -> long run expectation = 0
    %Expected inflation = del0pi + del1pi'*X0
    del0pi    =  1.9827657/100; %TABLE 4 long run inflation expectation is 2% (increased)
    del1pi(1) = -0.6289262/100;
    del1pi(2) =  0.1407911/100;

    %Nominal interest rate = del0R + del1R'*X0
    del0R     =  2.3994103/100;  %long run nominal interest is 2.4%
    del1R(1)  = -1.4835086/100
    del1R(2)  =  0.529618/100;

    %Process real interest rate and expected inflation: dXt = -K*Xdt + sig*dZt wit X(0)=X0 and dZt standard Brownian motion
    K(1,1)    =   7.6278548/100;
    K(2,1)    = -18.9995715/100;
    K(2,2)    =  35.2497604/100;

    %Prices of risk: Lt = vlam = vlam + mlam X0
    %long run parameters for price of risk for bonds (verschil lange <-> korte rente)
    vlam(1)   = .241685893; %TABLE 4: long run bond equals UFR of 3.9%(lowered), governs the risk premium on bonds
    vlam(2)   = .039258592;
    % Let op! (2,1) en (1,2) zijn verwisseld in TFT (mlam getransponeerd). Wel correct in Draper (2014) en Muns (2015)
    %state dependent parameters for price of risk
    mlam(1,1) =  .149281628;
    mlam(2,1) =  .089171663;
    mlam(1,2) = -.38139472;
    mlam(2,2) = -.083099766;

    %Realized inflation process: pi dt + sigPi'*dZt
    sigPi(1)  =  2.31129E-4;
    sigPi(2)  = -5.67828E-5;
    sigPi(3)  =  .6086625/100;
    %Stock return process: (rt+etaS) dt + sigS dZt
    sigS(1)   = -5.284295/100;
    sigS(2)   = -0.7552064/100;
    sigS(3)   = -2.1067319/100;
    sigS(4)   = 17.6849817/100; %TABLE 4 volatility of stocks equal to 20% (increased)
    
    etaS      = 6.5662395/100;  %TABLE 4 Equity risk premium increased, equals to 7%
    
    X0        = [];
end

%% Generate the parameters of the RTS given initial parameters
% gets used in function generate_KNW_set (as well in 'if bimulate' as in 'else')
function [a,b] = getRTScoeffs(del0R,del1R,vlam,mLam,K,vTau0)
    % bepaal a- en b-vectoren aan de hand van Muns(2015)
    % https://www.cpb.nl/sites/default/files/publicaties/download/cpb-background-document-august2015-financial-market-model-netherlands.pdf
    vTau   = vTau0(vTau0>0);
    k      = size(K,1);
    T      = length(vTau);
    
    % onder vgl.(16)
    M      = transpose(K + mLam);
    b0     = -M\del1R(:);

    % (32)
    detM = det(M);
    if detM <=0
        disp('determinant of M is negative: RTS may not stabilize')
    end
    
    % (40) E[RTS] (in logs)
    [MV,MD] = eig(M);
    lambda  = diag(MD);
    MVinv   = inv(MV);
    b       = zeros(k);
    B       = repmat(b0,1,T);
    expminlamtau = exp( -lambda(:)*vTau );
    for i=1:k
        b(:,i) = (MV(:,i)*MVinv(i,:))*del1R(:)./lambda(i);
        B      = B + b(:,i)*expminlamtau(i,:); % (37)
    end
    B       = checkImag(B,'B');
    
%     % check using (14)
%     B0   = zeros(k,length(vTau));
%     for t =1:length(vTau)
%         B0(:,t) = M\((expm(-M*t) - eye(k))*del1R(:));
%     end
    
    limRTS  = del0R + (vlam(:) - 0.5*b0(:))'*b0(:);
    a01     = -limRTS;
    ai1     = transpose(b0(:)-vlam(:))*b;
    aij1    = .5*transpose(b)*b;
  
    ai  = -ai1(:)./lambda(:);
    aij = -aij1./(repmat(lambda(:),[1 k]) + repmat(transpose(lambda(:)),[k 1]) );
    a0  = - sum(ai(:)) - sum(aij(:));
    
%     % (40) as a check
%     R = -a01 - (a0 + sum(ai(:).*expminlamtau))./vTau;
%     for i=1:k
%         for j=1:k
%             R = R - aij(i,j)*exp( -(lambda(i)+lambda(j))*vTau )./vTau;
%         end
%     end
%     R       = checkImag(R,'R');
    % (39)
    A       = a0 + a01*vTau + sum(repmat(ai(:),[1 T]).*expminlamtau);
    for i=1:k
       for j=1:k
            A = A + aij(i,j)*exp( -(lambda(i)+lambda(j))*vTau );
       end
    end
    A       = checkImag(A,'A');
    a       = -(A./vTau)';
    b       = -(B./repmat(vTau,[k 1]))';
    
    if any(vTau0==0)
        a = [del0R; a];     % (15), l'Hopital
        b = [del1R(:)'; b]; % (16), l'Hopital
    end
end

% gets used in function getRTScoeffs (which gets used in generate_KNW_set)
function y = checkImag(y,s)
    if max(abs(imag(y)))<1e-15
        y      = real(y); % verwaarloosbare imaginaire getallen
    else
        disp([s ' contains imaginary numbers'])
    end
end

% Generates the parameters which best fits the initial term structure
% uses function getRTS0
% gets used in function generate_KNW_set (in step 0) in case no X0 is given
function [X0opt, log_y0opt] = getX0(a,b,vTau0,date0)
    % bepaal initile vector X0 door gekwadrateerde afstand tot bepaalde RTS te minimaliseren
    log_y0        = getRTS0(date0);
    X0opt         = fminunc(@(x) getRTSy(x, a(1+vTau0), b(1+vTau0,:), log_y0(vTau0)),[0; 0]);
    [~, log_y0opt] = getRTSy(X0opt,a,b,0);
    figure;
    plot(1:length(log_y0), 100*(exp(log_y0)-1), '-', 0:length(log_y0), 100*(exp(log_y0opt)-1),'--')
    legend({'observed','estimated'},'Location','SouthEast')
end

% looks up the initial RTS
% gets used in function getX0 (which gets used in generate_KNW_set)
function RTS0 = getRTS0(date0)
    vdates = [201702 201703 201912 202001:202012];
    % euroswap RTS excl UFR van 28 Feb 2017, 31 Mrt 2017
    % en in de MLT-raming van dec 2019 t/m dec 2020.
    RTS0 = [-0.00218	-0.001800342	-0.001050921	-0.00024045	0.000762671	...
        0.00190094	0.003142135	0.004423089	0.005658017	0.006804954	...
        0.007881917	0.008780267	0.009601798	0.010306501	0.010917641	...
        0.011365519	0.011760871	0.012112425	0.012427076	0.012710346	...
        0.012856166	0.012988748	0.013109815	0.013220807	0.01332293	...
        0.013363211	0.01340051	0.013435146	0.013467395	0.013497494	...
        0.013464423	0.01343342	0.013404297	0.013376888	0.013351045	...
        0.013326639	0.013303553	0.013281682	0.013260934	0.013241223	...
        0.013147374	0.013058003	0.012972796	0.012891468	0.012813762	...
        0.012739439	0.012668284	0.012600099	0.012534701	0.012471923	...
        0.01241161	0.012353621	0.012297823	0.012244094	0.012192322	...
        0.012142402	0.012094235	0.012047731	0.012002806	0.011959381	...
        0.01191738	0.011876737	0.011837385	0.011799264	0.011762318	...
        0.011726493	0.011691738	0.011658006	0.011625254	0.011593438	...
        0.011562519	0.01153246	0.011503226	0.011474782	0.011447098	...
        0.011420143	0.011393889	0.011368308	0.011343376	0.011319068	...
        0.01129536	0.011272232	0.011249661	0.011227628	0.011206113	...
        0.0111851	0.01116457	0.011144507	0.011124896	0.01110572	...
        0.011086967	0.011068621	0.01105067	0.011033102	0.011015903	...
        0.010999063	0.010982571	0.010966416	0.010950587	0.010935075 ; 
        -0.002211	-0.001390571	-0.000540608	0.000656732	0.001676954	....
        0.003030687	0.004137879	0.005376164	0.006573035	0.007792079	...
        0.008767203	0.009580526	0.010379076	0.011064049	0.011658068	...
        0.012094486	0.012479717	0.012822267	0.013128858	0.013404869	...
        0.013553973	0.013689541	0.013813337	0.013926829	0.014031253	...
        0.014079992	0.014125122	0.01416703	0.01420605	0.01424247	...
        0.014244866	0.014247113	0.014249224	0.01425121	0.014253083	...
        0.014254852	0.014256525	0.014258111	0.014259615	0.014261043	...
        0.014184421	0.014111453	0.014041884	0.013975482	0.013912034	...
        0.013851349	0.01379325	0.013737575	0.013684175	0.013632914	...
        0.013583665	0.013536313	0.01349075	0.013446876	0.013404599	...
        0.013363834	0.013324501	0.013286526	0.013249839	0.013214377	...
        0.013180078	0.013146887	0.013114751	0.013083619	0.013053447	...
        0.01302419	0.012995807	0.012968259	0.012941511	0.012915528	...
        0.012890277	0.012865728	0.012841852	0.012818623	0.012796013	...
        0.012773998	0.012752556	0.012731664	0.012711302	0.012691449	...
        0.012672086	0.012653196	0.012634762	0.012616767	0.012599195	...
        0.012582033	0.012565265	0.012548879	0.012532861	0.012517199	...
        0.012501882	0.012486898	0.012472236	0.012457887	0.01244384	...
        0.012430086	0.012416615	0.01240342	0.012390491	0.012377821;
        0.001951282	0.002591364	0.003421741	0.004639597	0.005781467	0.00719762	0.008490963	0.009815428	0.011126546	0.012277019	0.013327931	0.014204528	0.015005186	0.01569197	0.01628756	0.01671875	0.017099365	0.017437809	0.017740723	0.018013423	0.018144717	0.018264089	0.018373094	0.018473026	0.018564971	0.01860457	0.018641237	0.018675286	0.018706987	0.018736577	0.018731865	0.018727449	0.018723299	0.018719394	0.018715712	0.018712235	0.018708946	0.01870583	0.018702873	0.018700065	0.018614277	0.018532582	0.018454693	0.018380349	0.018309315	0.018241374	0.018176328	0.018113997	0.018054213	0.017996824	0.017941688	0.017888676	0.017837667	0.01778855	0.017741221	0.017695584	0.017651551	0.017609038	0.017567967	0.017528268	0.017489871	0.017452714	0.017416738	0.017381888	0.017348111	0.017315359	0.017283585	0.017252747	0.017222804	0.017193717	0.01716545	0.017137969	0.017111242	0.017085238	0.017059928	0.017035284	0.017011281	0.016987895	0.0169651	0.016942876	0.016921202	0.016900056	0.016879421	0.016859277	0.016839607	0.016820396	0.016801626	0.016783283	0.016765353	0.016747821	0.016730675	0.016713902	0.016697491	0.016681428	0.016665704	0.016650308	0.016635229	0.016620459	0.016605987	0.016591804;
        0.00216594	0.002818159	0.003660673	0.004890665	0.006044672	0.007472962	0.008778442	0.010115044	0.011438298	0.012600908	0.01365182	0.014528416	0.015329075	0.016015859	0.016611449	0.017042639	0.017423254	0.017761698	0.018064612	0.018337312	0.018468606	0.018587978	0.018696983	0.018796915	0.01888886	0.018928459	0.018965126	0.018999175	0.019030876	0.019060466	0.019055754	0.019051337	0.019047188	0.019043283	0.019039601	0.019036124	0.019032835	0.019029718	0.019026762	0.019023953	0.018938166	0.018856471	0.018778582	0.018704238	0.018633204	0.018565263	0.018500217	0.018437886	0.018378102	0.018320713	0.018265577	0.018212565	0.018161556	0.018112439	0.01806511	0.018019473	0.01797544	0.017932927	0.017891856	0.017852156	0.01781376	0.017776603	0.017740627	0.017705777	0.017672	0.017639248	0.017607474	0.017576636	0.017546693	0.017517606	0.017489339	0.017461858	0.017435131	0.017409127	0.017383817	0.017359173	0.01733517	0.017311783	0.017288989	0.017266765	0.017245091	0.017223945	0.01720331	0.017183166	0.017163496	0.017144284	0.017125515	0.017107172	0.017089242	0.01707171	0.017054564	0.017037791	0.017021379	0.017005317	0.016989593	0.016974197	0.016959118	0.016944347	0.016929875	0.016915693;
        0.002380598	0.003044954	0.003899604	0.005141733	0.006307877	0.007748304	0.009065921	0.010414659	0.01175005	0.012924797	0.013975709	0.014852305	0.015652964	0.016339748	0.016935338	0.017366528	0.017747143	0.018085586	0.018388501	0.018661201	0.018792494	0.018911867	0.019020872	0.019120803	0.019212749	0.019252348	0.019289015	0.019323063	0.019354765	0.019384354	0.019379643	0.019375226	0.019371077	0.019367172	0.01936349	0.019360013	0.019356724	0.019353607	0.019350651	0.019347842	0.019262055	0.01918036	0.01910247	0.019028127	0.018957093	0.018889152	0.018824106	0.018761775	0.018701991	0.018644602	0.018589466	0.018536454	0.018485445	0.018436328	0.018388999	0.018343362	0.018299329	0.018256815	0.018215745	0.018176045	0.018137649	0.018100492	0.018064516	0.018029666	0.017995889	0.017963137	0.017931363	0.017900525	0.017870582	0.017841495	0.017813228	0.017785747	0.01775902	0.017733016	0.017707705	0.017683062	0.017659059	0.017635672	0.017612878	0.017590654	0.01756898	0.017547834	0.017527198	0.017507055	0.017487385	0.017468173	0.017449404	0.017431061	0.017413131	0.017395599	0.017378453	0.01736168	0.017345268	0.017329206	0.017313482	0.017298086	0.017283007	0.017268236	0.017253764	0.017239582;
        0.002595257	0.003271749	0.004138536	0.005392802	0.006571083	0.008023646	0.009353399	0.010714275	0.012061802	0.013248686	0.014299598	0.015176194	0.015976853	0.016663637	0.017259227	0.017690417	0.018071031	0.018409475	0.018712389	0.018985089	0.019116383	0.019235756	0.019344761	0.019444692	0.019536638	0.019576237	0.019612903	0.019646952	0.019678654	0.019708243	0.019703532	0.019699115	0.019694966	0.019691061	0.019687379	0.019683902	0.019680612	0.019677496	0.01967454	0.019671731	0.019585944	0.019504249	0.019426359	0.019352016	0.019280982	0.019213041	0.019147995	0.019085663	0.01902588	0.018968491	0.018913355	0.018860343	0.018809334	0.018760217	0.018712887	0.018667251	0.018623217	0.018580704	0.018539634	0.018499934	0.018461538	0.018424381	0.018388405	0.018353555	0.018319778	0.018287026	0.018255252	0.018224414	0.018194471	0.018165384	0.018137117	0.018109636	0.018082909	0.018056904	0.018031594	0.018006951	0.017982948	0.017959561	0.017936767	0.017914543	0.017892868	0.017871723	0.017851087	0.017830943	0.017811274	0.017792062	0.017773293	0.01775495	0.017737019	0.017719488	0.017702342	0.017685569	0.017669157	0.017653095	0.017637371	0.017621974	0.017606896	0.017592125	0.017577653	0.017563471;
        0.002825342	0.0035138	0.004392553	0.005658785	0.006849031	0.00831356	0.00965528	0.011028121	0.012387614	0.013586463	0.014637375	0.015513972	0.016314631	0.017001415	0.017597005	0.018028195	0.018408809	0.018747253	0.019050167	0.019322867	0.019454161	0.019573534	0.019682539	0.01978247	0.019874416	0.019914014	0.019950681	0.01998473	0.020016432	0.020046021	0.02004131	0.020036893	0.020032744	0.020028839	0.020025157	0.02002168	0.02001839	0.020015274	0.020012318	0.020009509	0.019923722	0.019842027	0.019764137	0.019689794	0.01961876	0.019550818	0.019485773	0.019423441	0.019363657	0.019306268	0.019251133	0.019198121	0.019147112	0.019097994	0.019050665	0.019005029	0.018960995	0.018918482	0.018877412	0.018837712	0.018799315	0.018762159	0.018726183	0.018691332	0.018657556	0.018624803	0.01859303	0.018562192	0.018532248	0.018503161	0.018474895	0.018447414	0.018420686	0.018394682	0.018369372	0.018344729	0.018320726	0.018297339	0.018274545	0.018252321	0.018230646	0.018209501	0.018188865	0.018168721	0.018149052	0.01812984	0.01811107	0.018092727	0.018074797	0.018057266	0.01804012	0.018023347	0.018006935	0.017990873	0.017975148	0.017959752	0.017944674	0.017929903	0.017915431	0.017901249;
        0.003055427	0.003755851	0.00464657	0.005924768	0.00712698	0.008603475	0.00995716	0.011341967	0.012713426	0.013924241	0.014975153	0.01585175	0.016652409	0.017339193	0.017934783	0.018365973	0.018746587	0.019085031	0.019387945	0.019660645	0.019791939	0.019911312	0.020020317	0.020120248	0.020212194	0.020251792	0.020288459	0.020322508	0.02035421	0.020383799	0.020379088	0.020374671	0.020370522	0.020366617	0.020362935	0.020359457	0.020356168	0.020353052	0.020350095	0.020347287	0.0202615	0.020179804	0.020101915	0.020027572	0.019956537	0.019888596	0.019823551	0.019761219	0.019701435	0.019644046	0.019588911	0.019535898	0.019484889	0.019435772	0.019388443	0.019342806	0.019298773	0.01925626	0.019215189	0.01917549	0.019137093	0.019099936	0.019063961	0.01902911	0.018995333	0.018962581	0.018930808	0.018899969	0.018870026	0.018840939	0.018812672	0.018785191	0.018758464	0.01873246	0.01870715	0.018682506	0.018658504	0.018635117	0.018612323	0.018590099	0.018568424	0.018547278	0.018526643	0.018506499	0.018486829	0.018467618	0.018448848	0.018430505	0.018412575	0.018395043	0.018377898	0.018361125	0.018344713	0.01832865	0.018312926	0.01829753	0.018282451	0.018267681	0.018253209	0.018239026;
        0.003285513	0.003997903	0.004900587	0.006190751	0.007404929	0.008893389	0.01025904	0.011655813	0.013039238	0.014262019	0.015312931	0.016189528	0.016990186	0.01767697	0.01827256	0.01870375	0.019084365	0.019422809	0.019725723	0.019998423	0.020129717	0.020249089	0.020358094	0.020458026	0.020549971	0.02058957	0.020626237	0.020660286	0.020691987	0.020721577	0.020716865	0.020712449	0.020708299	0.020704394	0.020700712	0.020697235	0.020693946	0.02069083	0.020687873	0.020685065	0.020599277	0.020517582	0.020439693	0.020365349	0.020294315	0.020226374	0.020161328	0.020098997	0.020039213	0.019981824	0.019926688	0.019873676	0.019822667	0.01977355	0.019726221	0.019680584	0.019636551	0.019594038	0.019552967	0.019513268	0.019474871	0.019437714	0.019401738	0.019366888	0.019333111	0.019300359	0.019268585	0.019237747	0.019207804	0.019178717	0.01915045	0.019122969	0.019096242	0.019070238	0.019044928	0.019020284	0.018996281	0.018972895	0.0189501	0.018927876	0.018906202	0.018885056	0.018864421	0.018844277	0.018824607	0.018805396	0.018786626	0.018768283	0.018750353	0.018732821	0.018715675	0.018698902	0.018682491	0.018666428	0.018650704	0.018635308	0.018620229	0.018605459	0.018590987	0.018576804;
        0.003538504	0.004263031	0.005177852	0.006480152	0.007706467	0.009207065	0.010584852	0.011993762	0.013389324	0.014624241	0.015675153	0.01655175	0.017352409	0.018039193	0.018634783	0.019065973	0.019446587	0.019785031	0.020087945	0.020360645	0.020491939	0.020611312	0.020720317	0.020820248	0.020912194	0.020951792	0.020988459	0.021022508	0.02105421	0.021083799	0.021079088	0.021074671	0.021070522	0.021066617	0.021062935	0.021059457	0.021056168	0.021053052	0.021050095	0.021047287	0.0209615	0.020879804	0.020801915	0.020727572	0.020656537	0.020588596	0.020523551	0.020461219	0.020401435	0.020344046	0.020288911	0.020235898	0.020184889	0.020135772	0.020088443	0.020042806	0.019998773	0.01995626	0.019915189	0.01987549	0.019837093	0.019799936	0.019763961	0.01972911	0.019695333	0.019662581	0.019630808	0.019599969	0.019570026	0.019540939	0.019512672	0.019485191	0.019458464	0.01943246	0.01940715	0.019382506	0.019358504	0.019335117	0.019312323	0.019290099	0.019268424	0.019247278	0.019226643	0.019206499	0.019186829	0.019167618	0.019148848	0.019130505	0.019112575	0.019095043	0.019077898	0.019061125	0.019044713	0.01902865	0.019012926	0.01899753	0.018982451	0.018967681	0.018953209	0.018939026;
        0.003791496	0.004528159	0.005455117	0.006769554	0.008008006	0.00952074	0.010910664	0.01233171	0.013739409	0.014986463	0.016037375	0.016913972	0.017714631	0.018401415	0.018997005	0.019428195	0.019808809	0.020147253	0.020450167	0.020722867	0.020854161	0.020973534	0.021082539	0.02118247	0.021274416	0.021314014	0.021350681	0.02138473	0.021416432	0.021446021	0.02144131	0.021436893	0.021432744	0.021428839	0.021425157	0.02142168	0.02141839	0.021415274	0.021412318	0.021409509	0.021323722	0.021242027	0.021164137	0.021089794	0.02101876	0.020950818	0.020885773	0.020823441	0.020763657	0.020706268	0.020651133	0.020598121	0.020547112	0.020497994	0.020450665	0.020405029	0.020360995	0.020318482	0.020277412	0.020237712	0.020199315	0.020162159	0.020126183	0.020091332	0.020057556	0.020024803	0.01999303	0.019962192	0.019932248	0.019903161	0.019874895	0.019847414	0.019820686	0.019794682	0.019769372	0.019744729	0.019720726	0.019697339	0.019674545	0.019652321	0.019630646	0.019609501	0.019588865	0.019568721	0.019549052	0.01952984	0.01951107	0.019492727	0.019474797	0.019457266	0.01944012	0.019423347	0.019406935	0.019390873	0.019375148	0.019359752	0.019344674	0.019329903	0.019315431	0.019301249;
        0.004044487	0.004793287	0.005732382	0.007058956	0.008309544	0.009834415	0.011236476	0.012669659	0.014089495	0.015348686	0.016399598	0.017276194	0.018076853	0.018763637	0.019359227	0.019790417	0.020171031	0.020509475	0.020812389	0.021085089	0.021216383	0.021335756	0.021444761	0.021544692	0.021636638	0.021676237	0.021712903	0.021746952	0.021778654	0.021808243	0.021803532	0.021799115	0.021794966	0.021791061	0.021787379	0.021783902	0.021780612	0.021777496	0.02177454	0.021771731	0.021685944	0.021604249	0.021526359	0.021452016	0.021380982	0.021313041	0.021247995	0.021185663	0.02112588	0.021068491	0.021013355	0.020960343	0.020909334	0.020860217	0.020812887	0.020767251	0.020723217	0.020680704	0.020639634	0.020599934	0.020561538	0.020524381	0.020488405	0.020453555	0.020419778	0.020387026	0.020355252	0.020324414	0.020294471	0.020265384	0.020237117	0.020209636	0.020182909	0.020156904	0.020131594	0.020106951	0.020082948	0.020059561	0.020036767	0.020014543	0.019992868	0.019971723	0.019951087	0.019930943	0.019911274	0.019892062	0.019873293	0.01985495	0.019837019	0.019819488	0.019802342	0.019785569	0.019769157	0.019753095	0.019737371	0.019721974	0.019706896	0.019692125	0.019677653	0.019663471;
        0.004316368	0.005077304	0.006028536	0.007367246	0.008629971	0.010166979	0.011581177	0.013026497	0.014458469	0.015729797	0.016780709	0.017657305	0.018457964	0.019144748	0.019740338	0.020171528	0.020552143	0.020890586	0.021193501	0.021466201	0.021597494	0.021716867	0.021825872	0.021925803	0.022017749	0.022057348	0.022094015	0.022128063	0.022159765	0.022189354	0.022184643	0.022180226	0.022176077	0.022172172	0.02216849	0.022165013	0.022161724	0.022158607	0.022155651	0.022152842	0.022067055	0.02198536	0.02190747	0.021833127	0.021762093	0.021694152	0.021629106	0.021566775	0.021506991	0.021449602	0.021394466	0.021341454	0.021290445	0.021241328	0.021193999	0.021148362	0.021104329	0.021061815	0.021020745	0.020981045	0.020942649	0.020905492	0.020869516	0.020834666	0.020800889	0.020768137	0.020736363	0.020705525	0.020675582	0.020646495	0.020618228	0.020590747	0.02056402	0.020538016	0.020512705	0.020488062	0.020464059	0.020440672	0.020417878	0.020395654	0.02037398	0.020352834	0.020332198	0.020312055	0.020292385	0.020273173	0.020254404	0.020236061	0.020218131	0.020200599	0.020183453	0.02016668	0.020150268	0.020134206	0.020118482	0.020103086	0.020088007	0.020073236	0.020058764	0.020044582;
        0.004588248	0.005361321	0.00632469	0.007675537	0.008950399	0.010499543	0.011925878	0.013383334	0.014827443	0.016110908	0.01716182	0.018038416	0.018839075	0.019525859	0.020121449	0.020552639	0.020933254	0.021271698	0.021574612	0.021847312	0.021978606	0.022097978	0.022206983	0.022306915	0.02239886	0.022438459	0.022475126	0.022509175	0.022540876	0.022570466	0.022565754	0.022561337	0.022557188	0.022553283	0.022549601	0.022546124	0.022542835	0.022539718	0.022536762	0.022533953	0.022448166	0.022366471	0.022288582	0.022214238	0.022143204	0.022075263	0.022010217	0.021947886	0.021888102	0.021830713	0.021775577	0.021722565	0.021671556	0.021622439	0.02157511	0.021529473	0.02148544	0.021442927	0.021401856	0.021362156	0.02132376	0.021286603	0.021250627	0.021215777	0.021182	0.021149248	0.021117474	0.021086636	0.021056693	0.021027606	0.020999339	0.020971858	0.020945131	0.020919127	0.020893817	0.020869173	0.02084517	0.020821783	0.020798989	0.020776765	0.020755091	0.020733945	0.02071331	0.020693166	0.020673496	0.020654284	0.020635515	0.020617172	0.020599242	0.02058171	0.020564564	0.020547791	0.020531379	0.020515317	0.020499593	0.020484197	0.020469118	0.020454347	0.020439875	0.020425693;
        0.004860128	0.005645339	0.006620844	0.007983827	0.009270826	0.010832107	0.012270579	0.013740172	0.015196418	0.016492019	0.017542931	0.018419528	0.019220186	0.01990697	0.02050256	0.02093375	0.021314365	0.021652809	0.021955723	0.022228423	0.022359717	0.022479089	0.022588094	0.022688026	0.022779971	0.02281957	0.022856237	0.022890286	0.022921987	0.022951577	0.022946865	0.022942449	0.022938299	0.022934394	0.022930712	0.022927235	0.022923946	0.02292083	0.022917873	0.022915065	0.022829277	0.022747582	0.022669693	0.022595349	0.022524315	0.022456374	0.022391328	0.022328997	0.022269213	0.022211824	0.022156688	0.022103676	0.022052667	0.02200355	0.021956221	0.021910584	0.021866551	0.021824038	0.021782967	0.021743268	0.021704871	0.021667714	0.021631738	0.021596888	0.021563111	0.021530359	0.021498585	0.021467747	0.021437804	0.021408717	0.02138045	0.021352969	0.021326242	0.021300238	0.021274928	0.021250284	0.021226281	0.021202895	0.0211801	0.021157876	0.021136202	0.021115056	0.021094421	0.021074277	0.021054607	0.021035396	0.021016626	0.020998283	0.020980353	0.020962821	0.020945675	0.020928902	0.020912491	0.020896428	0.020880704	0.020865308	0.020850229	0.020835459	0.020820987	0.020806804];
        RTS0 = log(1+RTS0(ismember(vdates,date0),:));
end

%%
% gets used in function getX0 (which gets used in generate_KNW_set)
function [log_y_diff, log_y] = getRTSy(X,a,b,log_y0)
    X          = X(:);
    log_y      = a + b*X;
    log_y_diff = sum((log_y(:) - log_y0(:)).^2);
end

%% simulates either the P or the Q set
% uses function getVarPars & getVarStats
% gets used in function generate_KNW_set
function [economy,mZ,mu,Sigma,muY,SigmaY,vgeomeanexp,vdifStats] = simulate(economy,del0pi,del1pi,del0R,del1R,driftX,mLam,K,sigPi,sigS,vlam,etaS,vTau,vTaubond1,mZ,Nscen,simhor,X0,savepath0,stype,bsave,bscaleQ)
    % simuleer nieuwe set
    [vgam, mGam, V, cholV] = getVarPars(del0pi, del1pi, del0R, del1R, driftX, K, sigPi, sigS, vlam, mLam, etaS, -economy.nominal_term_structure.b'.*vTau, vTau, vTaubond1);
    [mu,Sigma,muY,SigmaY,vgeomeanexp,vdifStats] = getVarStats(mGam,vgam,V);
    Y                 = zeros(Nscen,size(mGam,1), simhor+2);
    Y(:,1:length(X0),1) = repmat(X0(:)', [Nscen 1 1]);
    
    if isempty(mZ)
        mZ      = randn(Nscen, size(cholV,2), simhor+1);
    else
        dif = size(cholV,2) - size(mZ,2);
        if dif < 0
            mZ      = mZ(:, 1:size(cholV,2), :);
        elseif dif > 0 % breidt mZ uit
            mZ(:, end+(1:size(cholV,2)), :) = randn(Nscen, dif, simhor+1);
        end
    end
    
    for j=1:Nscen
        for t=1:simhor+1
            Y(j,:,t+1) = vgam + mGam*squeeze(Y(j,:,t))' + cholV*squeeze( mZ(j,:,t)' );
        end
    end
    
    economy.state_variables_term_structure = Y(:,1:length(X0), :);
    Y                       = diff(Y, [], 3);
    
    economy.price_inflation = exp( squeeze( Y(:,1+length(X0)    , :) ) );
    economy.stock_returns   = exp( squeeze( Y(:,2+length(X0)    , :) ) );
    
    economy.wage_inflation  = economy.price_inflation + 0.005;  %DNB: De looninflatie wordt, conform het advies van de Commissie Parameters, gedefinieerd als de  geldende  prijsinflatie  plus een opslag van 0,5% en is daarom niet als aparte grootheid in de scenarioset opgenomen.
    [economy.RTS, economy.RTS_maturities] = constructRTS(economy, 1);
    
    % economy.bond_returns    = exp( squeeze( Y(:,3+length(X0):end, :) ) );  %for various maturities
    vTau1 = find(vTau>=1);
    vTau0 = vTau1 - 1;
    economy.bond_returns            = NaN( size(economy.RTS) - [0 0 1] );
    economy.bond_returns(:,2:end,:) = economy.RTS(:, vTau1, 1:end-1).^vTau(vTau1) ./ economy.RTS(:, vTau0, 2:end).^vTau(vTau0);
    economy.bond_returns(:,1,:)     = (economy.RTS(:, 1, 2:end) + economy.RTS(:, 1, 1:end-1))/2;
    if isequal(stype,'P')
        economy.probability_measure = 'physical_probability_measure';
    else
        economy.probability_measure = 'risk_neutral_probability_measure';
        if bscaleQ % dwing af dat S_t/N_t en B_t/N_t een martingaal zijn als de set de gehele populatie is
           numeraire            = squeeze(economy.bond_returns(:,economy.bond_portf_maturities==0,:) );
           scfactor             = 1/mean(mean((1+economy.stock_return)./(1+numeraire)));
           economy.stock_return = scfactor * economy.stock_return;
           for i=2:size(economy.bond_returns,2)
               scfactor                    = 1/mean(mean((1+squeeze(economy.bond_returns(:,i,:)))./(1+numeraire)));
               economy.bond_returns(:,i,:) = scfactor * economy.bond_returns(:,i,:);
           end
        end
    end
    economy = orderfields(economy);
    if bsave
        save([savepath0 'KNW_' stype],'economy')
    end
end

%%
% gets used in function simulate (which gets used in function generate_KNW_set)
function [vgam, mGam, V, cholV] = getVarPars(del0pi,del1pi,del0R,del1R,driftX,K,sigPi,sigS, vlam, mLam, etaS, B1, vTau, vTaubond1)
    % maak VAR(1)-model van parameters zoals in Draper (2014) vgl (42) (P-set) en (58) (Q-set)
    % https://www.cpb.nl/sites/default/files/publicaties/download/cpb-background-document-financial-market-model-netherlands.pdf
    % B1: -b*tau, dus gevoeligheid van obligatieprijzen, niet van rente
    % vTauBond1: meegenomen bond durations ( +1 zodat index klopt)
    h       = vTau(2) - vTau(1);
    k       = length(K);
    nobs    = 2 + size(B1,2); % aantal waargenomen variabelen
    % (9)
    vTheta0 = [driftX(:); del0pi - .5*sum(sigPi.^2);
        squeeze(del0R + etaS -.5*sum(sigS.^2));
        del0R + B1'*(vlam(:)+driftX(:)) - .5*sum(B1'.^2,2)];

    mTheta1 = [    -K                                       zeros(k,nobs);
                del1pi                                      zeros(1,nobs);
                del1R                                       zeros(1,nobs);
                repmat(del1R,[size(B1,2) 1]) + B1'*mLam     zeros(size(B1,2),nobs)];

    sigYps  = [eye(k) zeros(k,2);
               sigPi    0;
               sigS;
               B1'    zeros(size(B1,2),2)];
    nn      = size(mTheta1,1);
    
    % (11)
    [U,D]  = eig(mTheta1);
    F      = diag(h*alpha(diag(D)*h));
    W      = ((U\sigYps)*(U\sigYps)').*(h*alpha(h*(repmat(diag(D),[1 nn])+repmat(diag(D)',[nn 1])) ));

    % (10)
    vgam   = U*F*(U\vTheta0(:));
    mGam   = U*expm(D*h)/U;

    V      = U*W*U';
    if rank(V)==size(mTheta1,1)
        cholV             = chol(V)';
    else
        [VV,DD] = eig(V);
        % The Hadamard product satisfies the rank inequality: DD(:, 1:k+2)
        cholV = VV*DD.^.5;
        if max(abs(imag(cholV(:)))) <1e-5
            cholV = real(cholV);
        end
        cholV = cholV(:, 1:min(end,rank(V)) );
    end
    index = [1:length(sigS)  length(sigS)+vTaubond1];
    vgam  = vgam(index);
    mGam  = mGam(index, index);
    V     = V(index, index);
    cholV = cholV(index,:);
end

%%
%  gets used in function simulate (which gets used in function generate_KNW_set)
function [mu,Sigma,muY,SigmaY,vgeomean,vdiffStats] = getVarStats(mGam,vgam,V)
    % bepaal lange termijn statistieken met Muns (2015)
    % https://www.cpb.nl/sites/default/files/publicaties/download/cpb-background-document-august2015-financial-market-model-netherlands.pdf
    k                   = size(mGam,1);
    % bereken stats als RW (evt met trend), want anders is variantie oneindig:
    vdiffStats          = all(mGam-eye(length(mGam))==0); % & vgam(:)'~=0;
    mGam(:, vdiffStats) = 0;
    % (22)
    mu     = (eye(k) - mGam)\vgam(:);
    Sigma  = reshape((eye(k^2) - kron(mGam,mGam) )\V(:), k, k);
    
    % (25) en (26) 'annually compounded' jaarrendementen (gecorrigeerde typo in (26))
    muY    = exp(mu(:) + .5*diag(Sigma)) - 1;
    % https://en.wikipedia.org/wiki/Log-normal_distribution#Multivariate_log-normal
    SigmaY = (exp(Sigma) - 1).*exp(mu(:)*mu(:)' + .5*(repmat(diag(Sigma),[1 k]) + repmat(diag(Sigma)',[k 1])));
    
    % geometric means (27)
    vgeomean = exp(mu) - 1;
end

% gets used in function getVarPars (which gets used in function simulate which gets used in function generate_KNW_set)
function y = alpha(x)
    % hulpfunctie (49) in Draper (2014)
    % https://www.cpb.nl/sites/default/files/publicaties/download/cpb-background-document-financial-market-model-netherlands.pdf
    y       = (exp(x)-1)./x;
    y(x==0) = 1;
end

% gets used in function generate_KNW_set
function [mu, Sigma, muY, SigmaY, vgeomeanexp, vdifStats] = getStatsFromPars(economy, stype, vTau, vTauBond1)
    % bepaal statistieken van parameters. tussenstap via VAR(1)-model
    if isfield(economy, 'par')
        par = economy.par;
    else
        par = economy.parameters;
    end
    switch stype % maak VAR(1)-model van parameters
        case 'P' % (42) in Draper (2014)
            [vgam, mGam, V, ~] = getVarPars(par.del0pi, par.del1pi, par.del0R, par.del1R, zeros(size(par.vlam)), par.K, par.sigPi, ...
                par.sigS, par.vlam, par.mLam, par.etaS, -economy.nominal_term_structure.b'.*vTau, vTau, vTauBond1);
        case 'Q' % (58) in Draper (2014)
            [vgam, mGam, V, ~] = getVarPars(par.del0pi, par.del1pi, par.del0R, par.del1R, -par.vlam, par.K+par.mLam, par.sigPi, ...
                par.sigS, par.vlam, zeros(size(par.mLam)), 0, -economy.nominal_term_structure.b'.*vTau, vTau, vTaubond1);
    end
    % bepaal statistieken VAR(1)-systeem
    [mu,Sigma,muY,SigmaY,vgeomeanexp,vdifStats] = getVarStats(mGam,vgam,V);
end